; vga_threaded.asm - threaded generate VGA text, background for frame updates
; author Neil Franklin, last modification 2009.01.07


; ------ hardware usage

; --- controller model

; we are using an ATmega32
  .include "m32def.inc"

; --- fuse settings

; fast external quarz osc, max stable startup
  ; FF: 7:BODLVL=1, 6:BOD=1 (disable), 54:SUT=11, 3210:CKSEL=1111 (ext resonat)
  ; D9: 7:OCDEN=1 (dis), 6:JTAGEN=1 (dis), 5:SPIEN=0 (en), 4:CKOPT=1
  ;     3:EESAVE=1 (not preserve), 21:BOOTSZ=00, 0:BOOTRST=1 


; --- board wiring

; VGA "HD-15" connector pinout (correctly named DE15, D-Shell E-size 15pin)

  ; pins   1/2/3:  R/G/B,    0.7V analog
  ; pin    4:      ID2,      TTL
  ; pin    5:      GND       (for IDs?)
  ; pins   6/7/8:  GND RGB   (for R/G/B coax shields)
  ; pin    9:      n.c./key  (can also be wired as +5V for DDC)
  ; pin   10:      GND HV    (for H-Sync + V-Sync coax shields)
  ; pins  11/12:   ID0/ID1,  TTL
  ; pin   13:      H-Sync,   TTL (can also be wired as C-Sync)
  ; pin   14:      V-Sync,   TTL
  ; pin   15:      n.c.      (can also be wired as ID3)
  ;                     (ID1/ID3 can also be wired as SDA/SCA for DDC I2C bus)

; VGA "R" DACs, 3 times 2bit/4value, gives 4*4*4=64 colours, RRGGBB

  ; each DAC from each bit TTL 5V, to 2bit 0.233+0.466V=0.7V @ 75ohm monitor
  ; needs per DAC resistors:
  ;   bit0  75ohm/0.233V*5V-75ohm = 1534okm  -> resistor 1k5ohm
  ;   bit1  75ohm/0.466V*5V-75ohm = 730ohm -> resistor 680ohm (not 820ohm)

; VGA ports and port bits

  ; video pixel data
  .equ VGADPORT = PORTC  ; 3 2bit colour DACs, write fast, must be same port
                         ;   use full 8bit out, bits 7+6 must accept overwrite
  .equ VGADDDR  = DDRC
  .equ VGABLACK = 0x00   ;   all 3 DACs set to 0b00, clear port
  .equ VGABLUE  = 0x03   ;   pins 1+0 blue  2bit DAC
  .equ VGAGREEN = 0x0C   ;   pins 3+2 green 2bit DAC
  .equ VGARED   = 0x30   ;   pins 5+4 red   2bit DAC
  .equ VGAWHITE = VGABLUE+VGAGREEN+VGARED  ; all 3 DACs set to 0b11, also DDR

  ; sync control lines
  .equ VGAHPORT = PORTC  ; 1 H-Sync TTL out, can accept draw overwrite with 0
  .equ VGAHDDR  = DDRC
  .equ VGAHBIT  = PINC6
  .equ VGAVPORT = PORTC  ; 1 V-Sync TTL out, can accept draw overwrite with 0
  .equ VGAVDDR  = DDRC
  .equ VGAVBIT  = PINC7

; control ports and port bits

  ; size, 25 or 30 rows, VGA text/400@70Hz or graphics/480@60Hz timing
  .equ SIZPORT  = PORTD  ; 1 size switch TTL in with Pull-Up
  .equ SIZPIN   = PIND
  .equ SIZDDR   = DDRD
  .equ SIZBIT   = PIND7  ; 0(jumper) = 30 rows, 1(default) = 25 rows

  ; line 2 mode, 2nd blank or double scan
  .equ LN2PORT  = PORTD  ; 1 mode switch TTL in with Pull-Up
  .equ LN2PIN   = PIND
  .equ LN2DDR   = DDRD
  .equ LN2BIT   = PIND6  ; 0(jumper) = double scan, 1(default) = 2nd blank

  ; 2nd blank colour, blank or background
  .equ CO2PORT  = PORTD  ; 1 colour switch TTL in with Pull-Up
  .equ CO2PIN   = PIND
  .equ CO2DDR   = DDRD
  .equ CO2BIT   = PIND5  ; 0(jumper) = background, 1(default) = blank


; ------ data usage

; --- VGA timing constants

; VGA text mode
  .equ VRBBTLINES = 12   ; vertical blanking lines
  .equ VRPUTLINES = 13
  .equ VRTBTLINES = 25
  .equ CHDRTROWS  = 25   ; draw rows, *segments(8)*lines(2) = lines(25*8*2=400)

; VGA graphics mode
  .equ VRBBGLINES = 12   ; vertical blanking lines
  .equ VRPUGLINES = 13
  .equ VRTBGLINES = 25
  .equ CHDRGROWS  = 30   ; draw rows, *segments(8)*lines(2) = lines(30*8*2=480)

; --- vertical blanking state numbers

  ; VR*LINE loops run VRINSTATE = with these constants, all negative values
  .equ VRINBB     = -1   ; all negative values
  .equ VRINPU     = -2
  .equ VRINTB     = -3
  ; CHDRROW loop runs CHROW = CHDRTROWSorCHDRGROWS..1, all positive values


; ------ register usage

; --- standard ones

; use an systematic naming of AVR registers
  .include "avr_registers.inc"

; here use/(re-)name R15..R2 (fast variables) registers
  ; avra fails to make an .def (or .equ) to an register name defined by .def
  ;   so here use an 2nd .def with the original R* name, and use  -W NoRegDef

; --- project specific ones

; for drawing timer ISR
  ; registers for storing drawing timer ISR continuation address
  .def BGCONTH    = R15
  .def BGCONTL    = R14
  ; registers for storing drawing timer ISR frame buffer address
  .def BGFBUFH    = R13
  .def BGFBUFL    = R12

; for drawing control
  ; registers for vertical blanking line/blanktype loop counting
  .def VRLINE     = R11  ; identical with CHLINE, same use, same register
  .def VRINSTATE  = R10  ; identical with CHROW, track where drawing is
  ; registers for line drawing line/segm/row loop counting
  .def CHLINE     = R11  ; identical with VRLINE, same use, same register
  .def CHSEGM     = R23  ; rename for I1, must be, because subi/andi for step
  .def CHROW      = R10  ; identical with VRINSTATE, track where drawing is

; for colours
  ; registers for colour patterns, are the LUT, use for  out VGADPORT,value-reg
  ;   use non-immediate fast/specials vars, no immediate load and permanent use
  .def BLANK      = R9   ; constant for colour blanking (black) after line

  .def LUTBACK    = R8   ; 2 colours for foreground and background
  .def LUTFORE    = R7

  .def LUTCOL3    = R6   ; 3 alternate foreground colours per row
  .def LUTCOL2    = R5
  .def LUTCOL1    = R4

; R3 to R2 are still free to use as F1 to F0


; ------ SRAM usage

  .dseg
  .org MINRAM

; --- define frame buffer

  ; one row
  .equ FBROWBEG = 5      ; 5 color bytes a begin of row
  .equ FBCOLS   = 30     ; 30 columns of character index bytes in row
  .equ FBROWEND = 1      ; 1 abort pseudo-char "return" index, at end of row
  .equ FBROWLEN = FBROWBEG+FBCOLS+FBROWEND  ; legth of row in frame buffer
  ; entire frame
  .equ FBROWS   = 30     ; rows in entire frame buffer, use 25 or 30
  .equ FBSIZE   = FBROWS*FBROWLEN  ; space used for frame buffer

FBUF:
  .byte FBSIZE           ; uses 30*(5+30+1) = 1080bytes (of ATmega32 2048)

; --- snail animation demo position

DESNPOS:
  .byte 1                ; snail horizontal position, 0..FBCOLS-1

; --- cursor animation demo position

DECUPOS:
  .byte 1                ; cursor horizontal position, 0..FBCOLS-1


; ------ Flash usage

  .cseg
  .org MINROM

; ------ system initialisation

; --- handle reset and interrupts (all identical as all not used)

; set up reset and interrupt vectors

VECRESET:
  jmp RESET
VECINT0:
  jmp UNUINT
VECINT1:
  jmp UNUINT
VECINT2:
  jmp UNUINT
VECT2CMP:
  jmp UNUINT
VECT2OVF:
  jmp UNUINT
VECT1CAP:
  jmp UNUINT
VECT1CMPA:
  jmp UNUINT
VECT1CMPB:
  jmp UNUINT
VECT1OVF:
  jmp UNUINT
VECT0CMP:
  jmp T0CMPINT
VECT0OVF:
  jmp UNUINT
VECSPI:
  jmp UNUINT
VECURXC:
  jmp UNUINT
VECUDRE:
  jmp UNUINT
VECUTXC:
  jmp UNUINT
VECADC:
  jmp UNUINT
VECEERDY:
  jmp UNUINT
VECACMP:
  jmp UNUINT
VECTWI:
  jmp UNUINT
VECSPM:
  jmp UNUINT


; --- handle unused interrupts, despite not used nor triggered

UNUINT:
  ; just return, nothing else makes sense here
  reti


; ------ debugging helpers

; --- debuging LED, using TxD pin + RS232 adapter + RSR23 tester TxD LED

; reset state is DDR-TxD=0=in, TTL=float/5V, RS232=-12V, LED=red
;   allways set port bit and DDR, so no setup is required

DBGLEDGRN:
  cbi PORTD,PINTXD       ; TxD=0, TTL=0V, RS232=12V, LED=green
  sbi DDRD,PINTXD        ; DDR-TxD=1=out
  ret

DBGLEDRED:
  sbi PORTD,PINTXD       ; TxD=1, TTL=5V, RS232=-12V, LED=red
  sbi DDRD,PINTXD        ; DDR-TxD=1=out
  ret

DBGLEDINV:
  sbis PORTD,PINTXD
  rjmp DBGLEDRED         ; TxD is 0, invert to TxD=1
  rjmp DBGLEDGRN         ; TxD is 1, invert to TxD=0


; ------ font unthreading and drawing mechanism

; --- segment drawing routines

  ; each char in font is an set of 8 segments, drawn on 8 line(pair)s

  ; each segment consists of 4 pixels, drawing an segment consists of:
  ;   4 clocks write pixels with  out VGADPORT,xxxx (xxxx = LUTFORE or LUTBACK)
  ;   between these 4*3 clocks for unthreading and indirect jumping next char

  ; the code for unthreading and indirect jumping works like this:
  ;   ijmp requires address in ZH:ZL, no other possible
  ;     ZH is constant, high addr of this table, settable by  ldi ZH,high(SEGM)
  ;     ZL is fetched from font segment index tables, by  lpm ??,Y
  ;   lpm also requires address in ZH:ZL, no other possible
  ;     ZH is per-segment constant, high addr of table, set once for line draw
  ;     ZL is fetched from frame buffer, by  ld ZL,Y+

  ; addressed/executed by indirect computed ijmp, only 8bit indexing for speed
  ;   the labels are only needed for jump table building

  .org 0x0100            ; must be 0x??00, for begin of 256word|512byte page
                         ;   hardcoded 0x0100, as avra knows no .align

SEGM:


  ; all 16 segments for character drawing, repeat this for each segment
  ;                      ;   {4*4=16} and 12instr, 16segm=192instr, 3/4 space

  ; SEGMxxxx:
  ; out VGADPORT,LUT*    ; {1}  current character: draw pixel 1
  ; ld ZL,Y+             ; {2}  next character: get character code/index
  ;                      ;        is basically low(FONT*)*2 (word->byte)
  ; mov ZH,CHSEGM        ; {1}    and select proper font segment index table
  ;                      ;        is basically high(FONT*)*2 (word->byte)
  ; out VGADPORT,LUT*    ; {1}  current character: draw pixel 2
  ; lpm I0,Z             ; {3}  next character: read index, is low(SEGM*)
  ;                      ;        can not be directly loaded into ZH or ZL!
  ; out VGADPORT,LUT*    ; {1}  current character: draw pixel 3
  ; mov ZL,I0            ; {1}  next character: compensate for lpm restrictions
  ; ldi ZH,high(SEGM)    ; {1}    use segment drawing code, const, high(SEGM)
  ; nop                  ; {1}    wait rest of pixel time
  ; out VGADPORT,LUT*    ; {1}  current character: draw pixel 4
  ; nop                  ; {1}  next character: wait beginning of pixel time
  ; ijmp                 ; {2}    go and execute it
  ;                      ;        continuation at next chars pixel 1, as above
  ;                      ;        this results in an endless loop

SEGM0000:
  out VGADPORT,LUTBACK
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTBACK
  lpm I0,Z
  out VGADPORT,LUTBACK
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTBACK
  nop
  ijmp

SEGM0001:
  out VGADPORT,LUTBACK
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTBACK
  lpm I0,Z
  out VGADPORT,LUTBACK
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTFORE
  nop
  ijmp

SEGM0010:
  out VGADPORT,LUTBACK
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTBACK
  lpm I0,Z
  out VGADPORT,LUTFORE
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTBACK
  nop
  ijmp

SEGM0011:
  out VGADPORT,LUTBACK
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTBACK
  lpm I0,Z
  out VGADPORT,LUTFORE
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTFORE
  nop
  ijmp

SEGM0100:
  out VGADPORT,LUTBACK
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTFORE
  lpm I0,Z
  out VGADPORT,LUTBACK
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTBACK
  nop
  ijmp

SEGM0101:
  out VGADPORT,LUTBACK
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTFORE
  lpm I0,Z
  out VGADPORT,LUTBACK
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTFORE
  nop
  ijmp

SEGM0110:
  out VGADPORT,LUTBACK
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTFORE
  lpm I0,Z
  out VGADPORT,LUTFORE
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTBACK
  nop
  ijmp

SEGM0111:
  out VGADPORT,LUTBACK
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTFORE
  lpm I0,Z
  out VGADPORT,LUTFORE
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTFORE
  nop
  ijmp

SEGM1000:
  out VGADPORT,LUTFORE
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTBACK
  lpm I0,Z
  out VGADPORT,LUTBACK
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTBACK
  nop
  ijmp

SEGM1001:
  out VGADPORT,LUTFORE
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTBACK
  lpm I0,Z
  out VGADPORT,LUTBACK
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTFORE
  nop
  ijmp

SEGM1010:
  out VGADPORT,LUTFORE
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTBACK
  lpm I0,Z
  out VGADPORT,LUTFORE
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTBACK
  nop
  ijmp

SEGM1011:
  out VGADPORT,LUTFORE
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTBACK
  lpm I0,Z
  out VGADPORT,LUTFORE
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTFORE
  nop
  ijmp

SEGM1100:
  out VGADPORT,LUTFORE
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTFORE
  lpm I0,Z
  out VGADPORT,LUTBACK
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTBACK
  nop
  ijmp

SEGM1101:
  out VGADPORT,LUTFORE
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTFORE
  lpm I0,Z
  out VGADPORT,LUTBACK
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTFORE
  nop
  ijmp

SEGM1110:
  out VGADPORT,LUTFORE
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTFORE
  lpm I0,Z
  out VGADPORT,LUTFORE
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTBACK
  nop
  ijmp

SEGM1111:
  out VGADPORT,LUTFORE
  ld ZL,Y+
  mov ZH,CHSEGM
  out VGADPORT,LUTFORE
  lpm I0,Z
  out VGADPORT,LUTFORE
  mov ZL,I0
  ldi ZH,high(SEGM)
  nop
  out VGADPORT,LUTFORE
  nop
  ijmp


  ; the pseudo-segment for the row end abort pseudo-character

  ; no drawing after 1st pixel, no timing, no unthreading next character

SEGMABRT:
  out VGADPORT,BLANK     ; {1}  switch off whatever was last background colour
                         ;        this is timing critical, so before jump
  jmp CHEND              ; {3}  go back to common CHDRLINE handler for rest
                         ;        no rjmp because coming from end of Flash


  ; some pseudo-segments for the pseudo-characters "specials"

  ; no drawing after 1st pixel, no timing, but after unthread next character
  ; {1}               current character: draw only pixel 1, as are spaces
  ; {16-1-10=5}       time for special stuff, the desired side effects
  ; {2+1+3+1+1+2=10}  next character: all unthread stuff at once

SEGM0FXB:                ; special pseudo-character "invert/underline on"
  out VGADPORT,LUTBACK   ;   for invert use 8 times, for underline 7space+1time
  eor LUTFORE,LUTBACK    ; {1}  FORE.eor.BACK "mixed"
  eor LUTBACK,LUTFORE    ; {1}  BACK.eor.(FORE.eor.BACK) = FORE "unmix"
  eor LUTFORE,LUTBACK    ; {1}  (FORE.eor.BACK).eor.FORE = BACK "unmix"
  nop                    ; {2}
  nop
  ld ZL,Y+
  mov ZH,CHSEGM
  lpm I0,Z
  mov ZL,I0
  ldi ZH,high(SEGM)
  ijmp

SEGM1BXF:                ; special pseudo-character "invert/underline off"
  out VGADPORT,LUTFORE   ;   needs this else space still inverted/underlined
  eor LUTFORE,LUTBACK    ; {1}  FORE.eor.BACK "mixed"
  eor LUTBACK,LUTFORE    ; {1}  BACK.eor.(FORE.eor.BACK) = FORE "unmix"
  eor LUTFORE,LUTBACK    ; {1}  (FORE.eor.BACK).eor.FORE = BACK "unmix"
  nop                    ; {2}
  nop
  ld ZL,Y+
  mov ZH,CHSEGM
  lpm I0,Z
  mov ZL,I0
  ldi ZH,high(SEGM)
  ijmp

SEGM0FX1:                ; special pseudo-character "alternate colour 1"
  out VGADPORT,LUTBACK
  eor LUTFORE,LUTCOL1    ; {1}  FORE.eor.COL1 "mixed"
  eor LUTCOL1,LUTFORE    ; {1}  COL1.eor.(FORE.eor.COL1) = FORE "unmix"
  eor LUTFORE,LUTCOL1    ; {1}  (FORE.eor.COL1).eor.FORE = COL1 "unmix"
  nop                    ; {2}
  nop
  ld ZL,Y+
  mov ZH,CHSEGM
  lpm I0,Z
  mov ZL,I0
  ldi ZH,high(SEGM)
  ijmp

SEGM0FX2:                ; special pseudo-character "alternate colour 2"
  out VGADPORT,LUTBACK
  eor LUTFORE,LUTCOL2    ; {1}  FORE.eor.COL2 "mixed"
  eor LUTCOL2,LUTFORE    ; {1}  COL2.eor.(FORE.eor.COL2) = FORE "unmix"
  eor LUTFORE,LUTCOL2    ; {1}  (FORE.eor.COL2).eor.FORE = COL2 "unmix"
  nop                    ; {2}
  nop
  ld ZL,Y+
  mov ZH,CHSEGM
  lpm I0,Z
  mov ZL,I0
  ldi ZH,high(SEGM)
  ijmp

SEGM0FX3:                ; special pseudo-character "alternate colour 3"
  out VGADPORT,LUTBACK
  eor LUTFORE,LUTCOL3    ; {1}  FORE.eor.COL3 "mixed"
  eor LUTCOL3,LUTFORE    ; {1}  COL3.eor.(FORE.eor.COL3) = FORE "unmix"
  eor LUTFORE,LUTCOL3    ; {1}  (FORE.eor.COL3).eor.FORE = COL3 "unmix"
  nop                    ; {2}
  nop
  ld ZL,Y+
  mov ZH,CHSEGM
  lpm I0,Z
  mov ZL,I0
  ldi ZH,high(SEGM)
  ijmp


; --- font segment index tables


  ; each char in font is an set of 8 (1 per segment) 8bit indexes into segments
  ;   for each of 8 segments the font has 256 index table, for 256 characters

  ; consists entirely of .db statements, each min 2 bytes/indexes, poss 4|8

  ; addressed/executed by computed lpm, label only for setting font begin
  ;   again only 8bit indexing for speed

  .org 0x0200            ; must be 0x??00, for begin of 8 256byte|128word pages
                         ;   hardcoded 0x0200, as avra knows no .align

FONT:

  ; include font as separate file, so they may be generated automatically
  ;   is done so from file vga_text_font.fon, by generator ./genfont
  .include "vga_text_font.inc"

  ; expected to be now at .org 0x0200+8*0x0080 = 0x0600, but make it sure
  ;   also adds more detailed lines to avra screen output
  .org 0x0600


; ------ frame buffer handling

; --- write colour into frame buffer

; parameters: XH:XL: address where in frame buffer, gets incremented
;                      only use when at begin of row
;             T0:    colour code, gets truncated to black(0x00)..white(0x3F)

FBWCOLOUR:

  andi T0,VGAWHITE       ; prevent false sync signals from user errors

  ; XH:XL set by caller, contains address in frame buffer
  st X+,T0               ; place this colour

  ret


; --- read colour from frame buffer

; parameters: XH:XL: address where in frame buffer, gets incremented
;                      only use when at begin of row
; returns:    T0:    colour code

FBRCOLOUR:

  ; XH:XL set by caller, contains address in frame buffer
  ld T0,X+               ; extract this colour

  ret


; --- write character into frame buffer

; parameters: XH:XL: address where in frame buffer, gets incremented
;                      only use when inside an row
;             T0:    character code

FBWCHAR:

                         ; prevent inserion of false abort from user errors
  cpi T0,ABORT           ; char - ABORT (abort pseudo-char, not in text)
    brne FBNOTABORT      ;   <>, is ok
    ldi T0,0x2E          ;   =, replace with ASCII "." to show fixup
  FBNOTABORT:

  ; XH:XL set by caller, contains address in frame buffer
  st X+,T0               ; place this char

  ret

; --- read character from frame buffer

; parameters: XH:XL: address where in frame buffer, gets incremented
;                      only use when inside an row
; returns:    T0:    character code

FBRCHAR:

  ; XH:XL set by caller, contains address in frame buffer
  ld T0,X+               ; extract this character

  ret


; --- write immediate string into frame buffer

; parameters: TOS (old PC): address of string (in Flash!)
;                           format of constant: .dw length, .db data
;                             data in .db must allways be pairs
;                               if not, add an 0 filler byte
;                             no not include in length count, else gets drawn
;                           these are assembled-in, after  call|rcall FBWSTRI
;             XH:XL: address where in frame buffer, gets incremented
;                    write only while inside current row, only to end of row

FBWSTRI:

                         ; next address in calling program, begin of constant
                         ;   must use ZH:ZL because of  lpm  instr
  pop ZH                 ; call|rcall stores low first/top, big endian(!)
  pop ZL                 ;   so read low as last/top

  lsl ZL                 ; correct address for prog word vs lpm byte addressing
  rol ZH                 ;   this will fail on chips larger 64kByte/32kWord

  push DL                ; source length, not YH:YL, as not an address
  push DH

  lpm DL,Z+              ; length of string constant, from program
  lpm DH,Z+

  ; DH:DL set above      ; for char = constantlen..1
  FBWSLOOP:

    ; XH:XL set by caller, contains address in frame buffer
    ld T0,X              ; check for row end, look at frame buffer index value
    cpi T0,ABORT         ; index - ABORT (abort pseudo-char, not in text)
      breq FBDSREST      ;   =, is end of row, abort copying string, drop rest

    lpm T0,Z+            ; copy string data, from program ...
    rcall FBWCHAR        ;   ... to frame buffer, breaks T0, does X+

    sbiw DH:DL,1         ; next char
    brne FBWSLOOP

  FBDSDONE:

  pop DH
  pop DL

  sbrc ZL,0              ; first unused byte is in 2nd half of an word ?
    adiw ZH:ZL,1         ;   yes, jump over added 2nd halfword filler byte

  lsr ZH                 ; undo prog word vs lpm byte address correction
  ror ZL

                         ; new address in calling program, instr after constant
  ijmp                   ; ijmp  is faster than  push ZL  push ZH  ret


; --- drop immediate string, instead of into frame buffer

  ; DH:DL still set      ; for char = current..1
  FBDSLOOP:
    lpm T0,Z+            ; step over string data, in program, no copy/store

  FBDSREST:
    sbiw DH:DL,1         ; next char
    brne FBDSLOOP
  rjmp FBDSDONE


; --- initialise frame buffer

FBINIT:

  push S0
  push S1

  ldi XL,low(FBUF)       ; begin of frame buffer
  ldi XH,high(FBUF)

  ldi S0,FBROWS          ; for row = FBROWS..1
  FBRLOOP:

    ; set default background colour for this row
    ldi T0,0x03          ; blue
    rcall FBWCOLOUR      ; breaks T0, does X+

    ; set default foreground colour for this row
    ;ldi T0,0x0F          ; cyan
    ;ldi T0,0x1F          ; light blue, slight reddish cyan
    ldi T0,0x1B          ; light blue, slight reddisch + lower green cyan, best
    ;ldi T0,0x2B          ; light blue, low red+green white
    rcall FBWCOLOUR

    ; default alternate colours for this row, progressively "stronger"
    ldi T0,0x0F          ; cyan
    rcall FBWCOLOUR
    ldi T0,0x3F          ; white
    rcall FBWCOLOUR
    ldi T0,0x3C          ; yellow
    rcall FBWCOLOUR

    ; blank out the frame buffer
    ldi S1,FBCOLS        ; for column = FBCOLS..1
    FBCLOOP:

      ldi T0,0x2E        ; fill unused space with ASCII "."
      rcall FBWCHAR      ; breaks T0 (requiring re-load), does X+

      dec S1             ; next column
      brne FBCLOOP

    ldi T0,ABORT         ; abort pseudo-char, not in lines, frame at row end
    st X+,T0             ; can not use FBWCHAR because framing is invalid char

    dec S0               ; next row
    brne FBRLOOP

  pop S1
  pop S0

  ret


; ------ timing delay helper

; --- wait a while doing nothing, spinloop

; parameters: I0: loop count, 1..256 (actually 1..255,0 where 0=256)
;                 gets destroyed (reduced to 0), to avoid push/pop time
;                 OK, as seldom reused, often loaded, so use immediate register
;                 this is only ever called from drawing, is interrupt, so I0

; timing range: (I0-1)*(1+2) + 1*(1+1) + 4 = I0*3+3 clocks
;   together with needed  ldi I0,<n>  and  rcall WAIT  1+3+(I0*3+3) clocks
;   gives min 1+3+(1*3+3)=10 clocks, for I0=1
;   gives max 1+3+(256*3+3)=774 clocks, for I0=0 (=256)

WAIT:
                         ; {n} in comments = clocks used, for time computation
                         ;   this program requires lots of exact clock counting
                         ;   single-clock miscounts produce visual artifacts!

  ; I0 set up by caller  ; {0}  for LOOP = I0..1
  WAILOOP:

    ; nothing in loop, for highest loop resolution, only 3 clocks

    dec I0               ; {1}  next LOOP
    brne WAILOOP         ; {2|1}

  ret                    ; {4}


; ------ handle timer interrupt driven drawing

; backgrounding drawing while blank lines are displayed


; --- interrupt service routine, push background, restore drawing state

; as far as drawing goes, this is an "return" from background

T0CMPINT:

                         ;      TCNT0 now = 0, in 8 clks 1, in 16 clks 2, ...
                         ;        in 600-8 ckls 74, in 600clks again 0
                         ; {0..3?}  last background instr jitter, if not {1}
                         ; {4}  interrupt entry|call|cli processing
                         ; {3}  interrupt vector jmp T0CMPINT

  in I0,SREG             ; {1}  SREG will be randomly overwritten
  push I0                ; {2}

  push XH                ; {2}  these are used for drawing, not ISR dedicated
  push XL                ; {2}
  push YH                ; {2}
  push YL                ; {2}
  push ZH                ; {2}
  push ZL                ; {2}

                         ; compensate for timer interrupt latency jitter
                         ;   basic method is to nop until 0 or 1 clock before
                         ;     timer steps, if not stepped wait 1 clock longer
                         ;   repeat entire game 8 clocks (1 timer step) later
                         ;     3 times so that up to 3clocks correction
                         ;     as minimal instruction 1clk, maximal 4clk

                         ;        now is 22 + 0..3 clocks since TCNT0=0
  nop                    ; {1}
                         ;        now is 23 + 1..3 clocks since TCNT0=0
  in I0,TCNT0            ; {1}  if 23+0=23 still TCNT0=2 else already TCNT0=3
  cpi I0,2               ; {1}  if TCNT0 = 2
    breq TSYNC1          ; {1|2}  not delayed by background 1, delay +1
  TSYNC1:
                         ;        now is 26 + 1..3 clocks since TCNT0=0
  nop                    ; {4}
  nop
  nop
  nop
                         ;        now is 30 + 1..3 clocks since TCNT0=0
  in I0,TCNT0            ; {1}  if 30+1=31 still TCNT0=3 else already TCNT0=4
  cpi I0,3               ; {1}  if TCNT0 = 3
    breq TSYNC12         ; {1|2}  not delayed by background 1..2, delay +1
  TSYNC12:
                         ;        now is 34 + 2..3 clocks since TCNT0=0
  nop                    ; {4}
  nop
  nop
  nop
                         ;        now is 37 + 2..3 clocks since TCNT0=0
  in I0,TCNT0            ; {1}  if 37+2=39 still TCNT0=4 else already TCNT0=5
  cpi I0,4               ; {1}  if TCNT0 = 4
    breq TSYNC123        ; {1|2}  not delayed by background 1..3, delay +1
  TSYNC123:
                         ;        now is 40 + 3 clocks since TCNT0=0

                         ; restore drawing state, as before BACKGROUND
  movw YH:YL,BGFBUFH:BGFBUFL  ; {1]  restore framebuf address for drawing
  movw ZH:ZL,BGCONTH:BGCONTL  ; {1}  restore continuation point for "return"

  ijmp                   ; {3}  and "return" to the line type that called
                         ;        now is 52 clocks since TCNT0=0

                         ; gives minimal for backgrounding = 23 + 4 + 52 clocks
                         ;   so forget backgrounding while horiz retrace

                         ; here is always ijmp to lines *CONT point
                         ;   that is left of screen blanking, then horiz pulse
                         ;   after that count lines and possibly state change
                         ;   next line (or next state) will again BACKGROUND


; --- end of interrupt, store drawing state, pop background

; as far as drawing goes, it "calls" background until timer "returns"

; parameters: ZH:ZL: address for continuation, "return" of the "call"
;             YH:YL: current frame buffer address, to save and restore

BACKGROUND:

                         ; timing not critical here, as enter background
                         ;   only for computing minimal backgrounding time

                         ; drawing state part which has no dedicated variables
  movw BGCONTH:BGCONTL,ZH:ZL  ; {1]  save continuation point for "return"
  movw BGFBUFH:BGFBUFL,YH:YL  ; {1]  save framebuffer address for drawing

  ldi I0,(1<<OCF0)       ; {1}  output compare flag reset, rest leave
  out TIFR,I0            ; {1}

  pop ZL                 ; {2}  undo pushed registers, non dedicated
  pop ZH                 ; {2}
  pop YL                 ; {2}
  pop YH                 ; {2}
  pop XL                 ; {2}
  pop XH                 ; {2}

  pop I0                 ; {2}  undo push SREG
  out SREG,I0            ; {1}

  reti                   ; {4}  enable interrupts and off to the background
                         ; {1..4?}  at least one random background instruction

                         ; used by timer interrupt 52 + n + 23 clocks


; --- setup interrupts, out of an simulated interrupt service routine call

DISPLAY:

                         ; only on startup, not repeated, not time critical
                         ;   so no {} time computation here

                         ; call|rcall to here already pushes PC, like in ISR
  cli                    ; disable interrupts, as if running as interrupt
                         ;   blank lines rjmp to BACKGROUND will (re-)enable

  in I0,SREG             ; same as for ISR
  push I0

  push XH                ; same as for ISR
  push XL
  push YH
  push YL
  push ZH
  push ZL

  ldi I0,0               ; clear timer 0
  out TCNT0,I0
  ldi I0,74              ; timer runs 0..74 for 75*8=600 clocks
  out OCR0,I0
  ldi I0,0x0A            ; no FOC, CTC mode, no COM, prescale 8, start
  out TCCR0,I0
                         ; TIFR left alone, output compare flag not cleared
                         ;   as OCF0 will be cleared anyway in BACKGROUND
  ldi I0,(1<<OCIE0)      ; output compare interrupt enable, rest disable
  out TIMSK,I0           ;   no problem as cli prevents interrupt triggering
                         ;   will be sei in BACKGROUND after OCF0 cleared

  ; can now enter into any of the 4 states, to start drawing
  ;   that will then call BACKGROUND and so do reti, that returns call DISPLAY
  rjmp VRBBSTATE         ; vertical blank, most settling time for monitor


; ------ drawing horizontal and vertical retrace

; --- general drawing sequence

; lines, are the timing critical thing, entire line 600clocks (31.75us/31.5kHz)
;   first horizontal retrace blank right time, 30clocks
;     this is line specific, used for tidying up and next line preparation
;   then horizontal retrace pulse time, 30clocks
;     this is common, is simply wait time
;   then horizontal retrace blank left time, 60clocks, largest spare time
;     this is line specific, used for further preparing for generating line
;     (entire horizontal retrace 30+30+60=120, 1/4 of 480 clocks, 1/5 of all)
;   then visible output pixels, 480 clocks, 160pixels at 3clocks/pixel
;     this is line type specific (blank, vertical retrace, character drawing)
;     this is used in all blank lines for executing background program

; frame, for consistency with lines, entire VGA Text frame 450 lines (70Hz)
;   first vertical retrace blank bottom lines, 1/4 of 50 = 12 lines
;   then vertical retrace pulse lines, 1/4 of 50 = 13 lines
;   then vertical retrace blank top lines, 1/2 of 50 = 25 lines
;     (entire vertical retrace all are blank lines, 12+13+25=50)
;   then visible output lines, 25*8*2 = 25*16 = 400 lines
;     (if 2nd line blank half of these are blank lines, 400/2=200)

; frame, for consistency with lines, entire VGA Graphics frame 525 lines (60Hz)
;   first vertical retrace blank bottom lines, 1/4 of 45 = 11 lines
;   then vertical retrace pulse lines, 1/4 of 45 = 11 lines
;   then vertical retrace blank top lines, 1/2 of 45 = 23 lines
;     (entire vertical retrace all are blank lines, 11+11+23=45)
;   then visible output lines, 30*8*2 = 30*16 = 480 lines
;     (if 2nd line blank half of these are blank lines, 480/2=240)


; --- set up port(s) for sync pulses and display size control switch

SYNCINIT:

  ; get data registers ready
  cbi VGAHPORT,VGAHBIT   ; Port = 0 (= Sync off)
  cbi VGAVPORT,VGAVBIT
  ; get DDR registers ready
  sbi VGAHDDR,VGAHBIT    ; DDR = 1 (= out)
  sbi VGAVDDR,VGAVBIT

  ; control switch

  ; size switch (25 row / 30 row) - get data register ready
  sbi SIZPORT,SIZBIT     ; Port/Pull-Up = 1 (= on)
  ; size switch (25 row / 30 row) - get DDR register ready
  cbi SIZDDR,SIZBIT      ; DDR = 0 (= input)

  ret


; --- horizontal retrace, generate pulse and start of left blank

; timing 1pre, 30 + 1+4 = (30+60=90)-55spare

HPULSE:

  ; {*pre} = "prepulse" time, before timepoint where pulse is generated
  ;          precompensate for this in callers timing

  ; {30}  horizontal retrace pulse, V = 0or1, H = 1, color = blank
  sbi VGAHPORT,VGAHBIT   ; {1pre+1}  start horizontal retrace pulse

  ldi I0,7               ; {1}  wait H pulse time, 30-1-1=28
  rcall WAIT             ; {3+(7*3+3)=27}

  ; {3of60}  horizontal retrace left blank, V = 0or1, H = 0, color = blank
  cbi VGAHPORT,VGAHBIT   ; {1pre+1}  end horizontal retrace pulse

  ; no left blank wait, give rest of time to caller, 55spare

  ret                    ; {4}


; --- vertical retrace bottom blank - state machine setup and draw lines

; timing (4+1)+(1+1)+(3+1)=11pre

VRBBSTATE:

  sbic SIZPIN,SIZBIT     ; {2|1pre}  1(default) = VGA text, 400 line, 25 row
    ldi I0,VRBBTLINES    ;   {1pre}
  sbis SIZPIN,SIZBIT     ; {2|1pre}  0(jumpered) = VGA graph, 480 line, 30 row
    ldi I0,VRBBGLINES    ;   {1pre}
  mov VRLINE,I0          ; {1pre}  for VRLINE=VRBBTLINES|VRBBGLINES..1

  ldi I0,VRINBB          ; {1pre}
  mov VRINSTATE,I0       ; {1pre}

VRBBLINE:

  rcall HPULSE           ; {3pre+(1pre+90-55spare)}

  ; {55of60}  horizontal retrace left blank, V = 0, H = 0, color = blank
  ; +
  ; {480}  "draw"/wait 160 blank pixels, V = 0, H = 0, color = blank
  ldi ZL,low(VRBBCONT)   ; {1}
  ldi ZH,high(VRBBCONT)  ; {1}
  rjmp BACKGROUND        ; {2}

VRBBCONT:

  ; {30}  horizontal retrace right blank, V = 0, H = 0, color = blank
  ldi I0,2               ; {1}  wait rest of its time, 30-2-6-6=14
  rcall WAIT             ; {3+(2*3+3)=12}
  nop                    ; {1}

  dec VRLINE             ; {1}  next VRLINE
  breq VRPUSTATE         ; {1|2+13pre}  or even next state

  nop                    ; {8} compensate difference next state vs next line
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  rjmp VRBBLINE          ; {2+(3+1)=6pre}


; --- vertical retrace pulse - state machine setup and draw lines

; timing (4+1)+2+(1+1)+(3+1)=13pre

VRPUSTATE:

  sbic SIZPIN,SIZBIT     ; {2|1pre}  1(default) = VGA text, 400 line, 25 row
    ldi I0,VRPUTLINES    ;   {1pre}
  sbis SIZPIN,SIZBIT     ; {2|1pre}  0(jumpered) = VGA graph, 480 line, 30 row
    ldi I0,VRPUGLINES    ;   {1pre}
  mov VRLINE,I0          ; {1pre}  for VRLINE=VRPUTLINES|VRPUGLINES..1

  ldi I0,VRINPU          ; {1pre}
  mov VRINSTATE,I0       ; {1pre}

  sbi VGAVPORT,VGAVBIT   ; {2pre}  start vertical retrace pulse

VRPULINE:

  rcall HPULSE           ; {3pre+(1pre+90-55spare)}

  ; {55of60}  horizontal retrace left blank, V = 1, H = 0, color = blank
  ; +
  ; {480}  "draw"/wait 160 blank pixels, V = 1, H = 0, color = blank
  ldi ZL,low(VRPUCONT)   ; {1}
  ldi ZH,high(VRPUCONT)  ; {1}
  rjmp BACKGROUND        ; {2}

VRPUCONT:

  ; {30}  horizontal retrace right blank, V = 1, H = 0, color = blank
  ldi I0,2               ; {1}  wait rest of its time, 30-2-6-6=14
  rcall WAIT             ; {3+(2*3+3)=12}
  nop                    ; {1}

  dec VRLINE             ; {1}  next VRLINE
  breq VRTBSTATE         ; {1|2+13pre}  or even next state

  nop                    ; {8} compensate difference next state vs next line
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  rjmp VRPULINE          ; {2+(3+1)=6pre}


; --- vertical retrace top blank - state machine setup and draw lines

; timing (4+1)+2+(1+1)+(3+1)=13pre

VRTBSTATE:

  sbic SIZPIN,SIZBIT     ; {2|1pre}  1(default) = VGA text, 400 line, 25 row
    ldi I0,VRTBTLINES    ;   {1pre}
  sbis SIZPIN,SIZBIT     ; {2|1pre}  0(jumpered) = VGA graph, 480 line, 30 row
    ldi I0,VRTBGLINES    ;   {1pre}
  mov VRLINE,I0          ; {1pre}  for VRLINE=VRTBTLINES|VRTBGLINES..1

  ldi I0,VRINTB          ; {1pre}
  mov VRINSTATE,I0       ; {1pre}

  cbi VGAVPORT,VGAVBIT   ; {2pre}  end vertical retrace pulse

VRTBLINE:

  rcall HPULSE           ; {3pre+(1pre+90-55spare)}

  ; {55of60}  horizontal retrace left blank, V = 0, H = 0, color = blank
  ; +
  ; {480}  "draw"/wait 160 blank pixels, V = 0, H = 0, color = blank
  ldi ZL,low(VRTBCONT)   ; {1}
  ldi ZH,high(VRTBCONT)  ; {1}
  rjmp BACKGROUND        ; {2}

VRTBCONT:

  ; {30}  horizontal retrace right blank, V = 0, H = 0, color = blank
  ldi I0,2               ; {1}  wait rest of its time, 30-2-9-6=13
  rcall WAIT             ; {3+(2*3+3)=12}
  nop                    ; {1}

  dec VRLINE             ; {1}  next VRLINE
  breq CHDRSTATE         ; {1|2+13pre}  or even next state

  nop                    ; {8} compensate difference next state vs next line
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  rjmp VRTBLINE          ; {2+(3+1)=6pre}


; --- background wait for (and sync with) begin of vertical retrace

VRWAIT:

  ldi T0,VRINBB

  VRSTILLPREVBB:
  cp VRINSTATE,T0        ; still in ver retr bottom blank of prev frame
    breq VRSTILLPREVBB   ;   wait while there, until out of it, to see next one

  VRNOTNEXTBB:
  cp VRINSTATE,T0        ; not yet in ver retr bottom blank of next frame
    brne VRNOTNEXTBB     ;   wait until there, then return to drawing

  ret


; ------ drawing visible output

; --- set up registers and ports for DACs and control switches

CHINIT:

  ; do not set LUT colour registers, they will be set while drawing every row
  ;   but set register for blanking after pixels drawn
  ldi T0,VGABLACK        ; black is all DAC bits set to 0, no signal
  mov BLANK,T0           ;   use this for retrace colour blanking

  ; VGA DACs - get data register ready
  out VGADPORT,BLANK
  ; VGA DACs - get DDR register ready
  in T0,VGADDDR
  ori T0,VGAWHITE        ; DDR = 1 = out, white is all DAC bits set to 1
  out VGADDDR,T0

  ; control switches

  ; line 2 mode switch (2nd blank / double scan) - get data register ready
  sbi LN2PORT,LN2BIT     ; Port/Pull-Up = 1 (= on)
  ; line 2 mode switch (2nd blank / double scan) - get DDR register ready
  cbi LN2DDR,LN2BIT      ; DDR = 0 (= input)

  ; 2nd blank colour switch (blank / background) - get data register ready
  sbi CO2PORT,CO2BIT     ; Port/Pull-Up = 1 (= on)
  ; 2nd blank colour switch (blank / background) - get DDR register ready
  cbi CO2DDR,CO2BIT      ; DDR = 0 (= input)

  ret


; --- draw character based visible output line

; timing (1+1)+(4+1)+(1+1)+(3+1)=13pre

CHDRSTATE:

  ldi YL,low(FBUF)       ; {1pre}  get frame buffer first row begin
  ldi YH,high(FBUF)      ; {1pre}

  sbic SIZPIN,SIZBIT     ; {2|1pre}  1(default) = VGA text, 400 line, 25 row
    ldi I0,CHDRTROWS     ;   {1pre}
  sbis SIZPIN,SIZBIT     ; {2|1pre}  0(jumpered) = VGA graph, 480 line, 30 row
    ldi I0,CHDRGROWS     ;   {1pre}
  mov CHROW,I0           ; {1pre}  for CHROW = CHDRTROWS|CHDRGROWS..1

CHDRROW:

  ldi CHSEGM,high(FONT)*2; {1pre}  for CHSEGM = FONT..FONT+7*256

CHDRSEGM:

  eor CHLINE,CHLINE      ; {1pre}  for CHLINE = 0x00|0xFF

CHDRLINE:

  rcall HPULSE           ; {3pre+(1pre+90-55spare)}

  ; {55of60}  horizontal retrace left blank, V = 0, H = 0, color = blank
  sbrs CHLINE,0          ; {2|1}  1 = second line, double scan or 2nd blank
    rjmp CHFIRST         ;   {2}  0 = first line, always draw

  sbis LN2PIN,LN2BIT     ; {2|1}  1(default) = 2nd line blank, blank line
    rjmp CHDOUBLE        ;   {2}  0(jumpered) = double scan, drawn line


  ; {remaining51of60}
  ld LUTBACK,Y           ; {2}  first byte in row is background colour
  sbic CO2PIN,CO2BIT     ; {2|1}  0(jumpered) = use background colour
    mov LUTBACK,BLANK    ; {1}   1(default) = set to blank colour

  adiw YH:YL,FBROWLEN    ; {2}  simulate increments, as if we had drawn

  ldi I0,12              ; {1}  wait rest of its time, 55-(2+2)-(2+2)-2=45
  rcall WAIT             ; {3+(12*3+3)=42}
  nop                    ; {2}
  nop

  ; {480}  "draw"/wait 160 blank pixels, V = 0, H = 0, color = blank
  out VGADPORT,LUTBACK   ; {1}  draw blank or background colour

  ldi ZL,low(CHDRCONT)   ; {1}  only now go to background
  ldi ZH,high(CHDRCONT)  ; {1}
  rjmp BACKGROUND        ; {1}

CHDRCONT:

  ; {30}  horizontal retrace right blank, V = 0, H = 0, color = blank
  out VGADPORT,BLANK     ; {1}  switch off whatever was last background colour
                         ;        this is timing critical, so before jump
  jmp CHEND              ; {3}  go to common CHDRLINE end handler for rest
                         ;        no rjmp for consistency with line abbort


  ; {remaining52of60}
  CHFIRST:
  nop                    ; {2}  compensate for only one skip, (2+1+2)-(1+2)=2
  nop

  ; {remaining50of60}
  CHDOUBLE:
  ldi I0,7               ; {1}  wait rest of its time, 50-10-10=30
  rcall WAIT             ; {3+(7*3+3)=27}
  nop                    ; {2}
  nop


  ; {10}  load LUT colour registers for this row
  ld LUTBACK,Y+          ; {2}  first 5 bytes in row are colours
  ld LUTFORE,Y+          ; {2}
  ld LUTCOL1,Y+          ; {2}
  ld LUTCOL2,Y+          ; {2}
  ld LUTCOL3,Y+          ; {2}


  ; {10}  get first character index and unthread it
  ;   use standard font unthreading technique, see there for generic case
  ;     but compressed, without 4 clocks of intermixed pixel OUT statements
  ;   CHSEGM segment index table selection has already been set up by above
  ld ZL,Y+               ; {2}  next (=first) char: get character code/index
                         ;        is basically low(FONT*)*2 (word->byte)
  mov ZH,CHSEGM          ; {1}    and select proper font segment index table
                         ;        is basically high(FONT*)*2 (word->byte)
  lpm I0,Z               ; {3}  next (=first) char: read index, is low(SEGM*)
                         ;        can not be directly loaded into ZH or ZL!
  mov ZL,I0              ; {1}  next (=first) char: compensate for lpm restrict
  ldi ZH,high(SEGM)      ; {1}    use segment drawing code, const, high(SEGM)
  ijmp                   ; {2}  next (=first) char: go and execute it
                         ;        continuation at next chars pixel 1, as above
                         ;        this results in an endless loop


  ; {30*4*4=480}  draw 30chars*4pixels=120 pixels, V = 0, H = 0, color = pixel

  ; for each X+ fetched char index: index with it and exec one segment
  ;   drawing done by  out VGADPORT,colour, one for each pixel, every 4 clocks
  ;   {so per segment time 4 pixels a 4 clocks = 16 clocks/segment}


  ; {30}  horizontal retrace right blank, V = 0, H = 0, color = blank

  ; above continues endless until an  abort pseudo-char  index in frame buffer
  ;   that pseudo-char exits the unthreading instead of drawing, breaks loop
  ;     it must be done this way, because no time for end of char loop counting
  ;   the pseudo-char also runs one  out VGADPORT,BLANK  to end line drawing

  ; SEGMABRT:
  ; out VGADPORT,BLANK   ; {1}  switch off whatever was last background colour
                         ;        this is timing critical, so before jump
  ; jmp CHEND            ; {3}  go back to common CHDRLINE handler for rest
                         ;        no rjmp because coming from end of Flash

CHEND:
                         ; {1+3}  above in abort pseudo-char, or blank line

  com CHLINE             ; {1}  next/other CHLINE, switch/flip 0x00 <-> 0xFF
    breq CHNEXSEG        ; {1|2}  CHLINE now 0x00, was 0xFF, do next segment

    sbiw YH:YL,FBROWLEN  ; {2}  undo reads, repeat same segment for 2nd line

    ldi I0,3             ; {1}  wait rest of its time, 30-(1+3)-1-1-2-7=15
    rcall WAIT           ; {3+(3*3+3)=15}

    rjmp CHDRLINE        ; {2+(3+1)=6pre}
  CHNEXSEG:

  inc CHSEGM             ; {1}  next CHSEGM, + 256byte|128word per segment
  cpi CHSEGM,high(FONT)*2+8  ; {1}  done all 8 segments for this row?
    breq CHNEXROW        ; {1|2}  yes, do next row

    sbiw YH:YL,FBROWLEN  ; {2}  undo reads, repeat same row for other segments

    ldi I0,1             ; {1}  wait rest time, 30-(1+3)-(1+2)-(1+1+1)-2-7=11
    rcall WAIT           ; {3+(1*3+3)=9}
    nop                  ; {1}

    rjmp CHDRSEGM        ; {2+(1+3+1)=7pre}
  CHNEXROW:

  dec CHROW              ; {1}  next CHROW
    brne CHDOROW         ; {1|2}  reversed logic, as too far for breq VRBBSTATE
    nop                  ; {4}  wait rest, 30-(1+3)-(1+2)-(1+1+2)-(1-1)-13=4
    nop
    nop
    nop
    rjmp VRBBSTATE       ; {2+11=13pre}  on to next state
  CHDOROW:

  nop                    ; {8}  wait rest, 30-(1+3)-(1+2)-(1+1+2)-(1+2)-8=8
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  rjmp CHDRROW           ; {2+(1+1+3+1)=8pre}


; ------ X/Y coordinate based frame buffer handling

; --- X/Y write colour into frame buffer

; parameters: S0: row in frame buffer to place colour
;             T0: colour code for FBWCOLOUR, gets truncated
;             T1: LUT reg to set, 0..4 = back/fore/alt1/alt2/alt3

XYWCOLOUR:

  cpi S0,FBROWS          ; row in 0..(FBROWS-1) ?
    brsh XYOUTRANGE      ;   >=, dont' colour non-existant row
  cpi T1,FBROWBEG
    brsh XYOUTRANGE      ;   >=, dont' set to non-existant LUT colour

  ldi XL,low(FBUF)       ; begin of frame buffer colours
  ldi XH,high(FBUF)

  ldi T2,FBROWLEN        ; skip rows before this one
  mul S0,T2
  add XL,ML              ; this row of frame buffer colours
  adc XH,MH

  add XL,T1              ; add index for wanted colour byte
  ldi T2,0
  adc XH,T2

  rjmp FBWCOLOUR


  XYOUTRANGE:
  ret


; --- X/Y read colour from frame buffer

; parameters: S0: row in frame buffer to extract colour
;             T1: LUT reg to set, 0..4 = back/fore/alt1/alt2/alt3
; returns:    T0: colour code from FBRCOLOUR

XYRCOLOUR:

  cpi S0,FBROWS          ; row in 0..(FBROWS-1) ?
    brsh XYOUTRANGE      ;   >=, dont' colour non-existant row
  cpi T1,FBROWBEG
    brsh XYOUTRANGE      ;   >=, dont' set to non-existant LUT colour

  ldi XL,low(FBUF)       ; begin of frame buffer colours
  ldi XH,high(FBUF)

  ldi T2,FBROWLEN        ; skip rows before this one
  mul S0,T2
  add XL,ML              ; this row of frame buffer colours
  adc XH,MH

  add XL,T1              ; add index for wanted colour byte
  ldi T2,0
  adc XH,T2

  rjmp FBRCOLOUR


; --- X/Y write character into frame buffer

; parameters: S0: row in frame buffer to place character
;             S1: column in frame buffer to place character
;             T0: ASCII character for FBWCHAR, left converted

XYWCHAR:

  cpi S0,FBROWS          ; row in 0..(FBROWS-1) ?
    brsh XYOUTRANGE      ;   >=, dont' draw in non-existant row

  cpi S1,FBCOLS          ; column in 0..(FBCOLS-1) ?
    brsh XYOUTRANGE      ;   >=, dont' draw in non-existant column

  ldi XL,low(FBUF+FBROWBEG) ; begin of frame buffer chars
  ldi XH,high(FBUF+FBROWBEG)

  ldi T2,FBROWLEN        ; skip rows before this one
  mul S0,T2
  add XL,ML              ; this row of frame buffer chars
  adc XH,MH

  add XL,S1              ; this column in frame buffer
  ldi T2,0
  adc XH,T2

  rjmp FBWCHAR


; --- X/Y read character from frame buffer, not needed up to now


; --- X/Y write immediate string into frame buffer

; parameters: S0: row in frame buffer to place first character of string
;             S1: column in frame buffer to place first character of string

XYWSTRI:

  cpi S0,FBROWS          ; row in 0..(FBROWS-1) ?
    brsh XYOUTRANGE      ;   >=, dont' draw in non-existant row

  cpi S1,FBCOLS          ; column in 0..(FBCOLS-1) ?
    brsh XYOUTRANGE      ;   >=, dont' draw in non-existant column

  ldi XL,low(FBUF+FBROWBEG) ; begin of frame buffer chars
  ldi XH,high(FBUF+FBROWBEG)

  ldi T2,FBROWLEN        ; skip rows before this one
  mul S0,T2
  add XL,ML              ; this row of frame buffer chars
  adc XH,MH

  add XL,S1              ; this column in frame buffer
  ldi T2,0
  adc XH,T2

  rjmp FBWSTRI


; ------ demo program

; --- draw an blank line

; parameters: S0: row in frame buffer to blank out

DEBLANK:

  push S1

  ldi S1,0

  rcall XYWSTRI
  .dw 40
  .db "                                        "  ; empty string
  inc S0

  pop S1

  ret


; --- draw entire ASCII char set

; parameters: S0: row in frame buffer to begin at
;             S1: column in frame buffer to begin at

DEASC_3X32:

  ; show 95 ASCII chars 32..126 (0x20..0x7E) in 3 rows of 32, 127 as blank

  rcall XYWSTRI
  .dw 32
  .db " !", 0x22, "#$%&", 0x27, "()*+,-./0123456789:;<=>?"
  inc S0

  rcall XYWSTRI
  .dw 32
  .db "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_"
  inc S0

  rcall XYWSTRI
  .dw 32
  .db "`abcdefghijklmnopqrstuvwxyz{|}~ "
  inc S0

  ret


DEASC_3X40:

  push S1

  rcall DEBLANK          ; get rid of left over dots at both ends
  rcall DEBLANK
  rcall DEBLANK

  subi S0,3              ; back up the 3 rows, for drawing actual stuff

  ldi S1,4               ; indent (40-32)/2=4 unused characters

  rcall DEASC_3X32

  pop S1

  ret


; --- draw black line to separate sections

; parameters: S0: row in frame buffer to black out

DEBLACK:

  ldi T0,0x00            ; black
  ldi T1,0               ; set background to black, leave unused foreground
  rcall XYWCOLOUR

  rjmp DEBLANK


; --- setup 6 black text on rainbow coloured background lines

; parameters: S0: row in frame buffer to begin at
;             T0: background colour

DERAIN_1X40:

  ldi T1,0               ; set background to colour T0
  rcall XYWCOLOUR

  ldi T0,0x00            ; black
  ldi T1,1               ; set foreground to that
  rcall XYWCOLOUR

  inc S0

  ret


; parameters: S0: row in frame buffer to begin at

DERAIN_6X40:

  ldi T0,0x30            ; red
  rcall DERAIN_1X40
  ldi T0,0x3C            ; yellow
  rcall DERAIN_1X40
  ldi T0,0x0C            ; green
  rcall DERAIN_1X40
  ldi T0,0x0F            ; cyan
  rcall DERAIN_1X40
  ldi T0,0x03            ; blue
  rcall DERAIN_1X40
  ldi T0,0x33            ; magenta
  rcall DERAIN_1X40

  subi S0,6              ; back up the 6 rows, for drawing stuff

  ret


; --- draw ASCII art wave(s)

; parameters: S0: row in frame buffer to begin at
;             S1: column in frame buffer to begin at

DEWAVE_1X23:

  rcall XYWSTRI
  .dw 23
  .db "-", 0x27, "__`-.__.-", 0x27, "__`-.__.-", 0x27, "_", 0
  inc S0

  ret


DEWAVE_6X23:

  rcall DEWAVE_1X23      ; make 6 lines with this (and all other graphics)
  rcall DEWAVE_1X23
  rcall DEWAVE_1X23
  rcall DEWAVE_1X23
  rcall DEWAVE_1X23
  rcall DEWAVE_1X23

  ret


; --- draw separator(s) of ASCII wave(s) from blockgraphic logo

; parameters: S0: row in frame buffer to begin at
;             S1: column in frame buffer to begin at

DESEPAR_1X1:

  ldi T0,BG1111          ; block graphics all 4 2x2pixel set to foreground
  rcall XYWCHAR
  inc S0

  ret


DESEPAR_6X1:

  rcall DESEPAR_1X1      ; make 6 lines with this (and all other graphics)
  rcall DESEPAR_1X1
  rcall DESEPAR_1X1
  rcall DESEPAR_1X1
  rcall DESEPAR_1X1
  rcall DESEPAR_1X1

  ret


; --- draw logo put together from 2x2pixel blockgraphics

; parameters: S0: row in frame buffer to begin at
;             S1: column in frame buffer to begin at

DELOGO_6x16:

  ; print this text using 1/2+7+1/2 * 4x12pixels, of 2x2pixel blockgraphics
  ;   = space used 1/2+7+1/2 * 2x6 chars = 16 chars width, always 6 rows

  ; .... ........ ........ ........ ....  .... ........ ........ ........ ....
  ; .... ()()()() ()()()() ()()()() ()()  ()() ()()()() ()()()() ()()()() ....

  ; .... ........ ........ ........ ....  .... ........ ........ ........ ....
  ; .... ..()().. ........ ....().. ..()  .... ..()..() ....()() ....().. ....

  ; .... ()...... ........ ..().... ..()  .... ..()..() ..().... ..()..() ....
  ; .... ..().... ..().... ()()().. ()()  ().. ..()..() ..().... ..()..() ....

  ; .... ....().. ()..().. ..().... ..()  .... ..()..() ..()..() ..()()() ....
  ; .... ....().. ()..().. ..().... ..()  .... ....().. ..()..() ..()..() ....

  ; .... ()().... ..().... ..().... ....  ().. ....().. ....()() ..()..() ....
  ; .... ........ ........ ........ ....  .... ........ ........ ........ ....

  ; .... ()()()() ()()()() ()()()() ()()  ()() ()()()() ()()()() ()()()() ....
  ; .... ........ ........ ........ ....  .... ........ ........ ........ ....

  rcall XYWSTRI
  .dw 16
  .db BG0000,  BG0011, BG0011,  BG0011, BG0011,  BG0011, BG0011,  BG0011
  .db BG0011,  BG0011, BG0011,  BG0011, BG0011,  BG0011, BG0011,  BG0000
  inc S0

  rcall XYWSTRI
  .dw 16
  .db BG0000,  BG0001, BG0010,  BG0000, BG0000,  BG0000, BG0010,  BG0001
  .db BG0000,  BG0001, BG0001,  BG0000, BG0011,  BG0000, BG0010,  BG0000
  inc S0

  rcall XYWSTRI
  .dw 16
  .db BG0000,  BG1001, BG0000,  BG0001, BG0000,  BG0111, BG0010,  BG0111
  .db BG0010,  BG0101, BG0101,  BG0101, BG0000,  BG0101, BG0101,  BG0000
  inc S0

  rcall XYWSTRI
  .dw 16
  .db BG0000,  BG0000, BG1010,  BG1010, BG1010,  BG0101, BG0000,  BG0101
  .db BG0000,  BG0100, BG0110,  BG0101, BG0101,  BG0101, BG1101,  BG0000
  inc S0

  rcall XYWSTRI
  .dw 16
  .db BG0000,  BG1100, BG0000,  BG0100, BG0000,  BG0100, BG0000,  BG0000
  .db BG1000,  BG0000, BG1000,  BG0000, BG1100,  BG0100, BG0100,  BG0000
  inc S0

  rcall XYWSTRI
  .dw 16
  .db BG0000,  BG1100, BG1100,  BG1100, BG1100,  BG1100, BG1100,  BG1100
  .db BG1100,  BG1100, BG1100,  BG1100, BG1100,  BG1100, BG1100,  BG0000
  inc S0

  ret


; --- draw graphics, wave(s) + seprator(s) + logo

; parameters: S0: row in frame buffer to begin at
;             S1: column in frame buffer to begin at

DEGRAPH_6X40:

  push S1

  rcall DEWAVE_6X23

  subi S0,6              ; back up the 6 rows, for separator

  subi S1,-23            ; indent by 23 columns, jump over ASCII wave(s)
  rcall DESEPAR_6X1

  subi S0,6              ; back up the 6 rows, for logo

  subi S1,-1             ; indent by 1 column, jump over separator(s)
  rcall DELOGO_6X16

  pop S1

  ret


; --- draw announce message

; parameters: S0: row in frame buffer to begin at
;             S1: column in frame buffer to begin at

DEANN_3X40:

  rcall XYWSTRI
  .dw 40
  .db "ATmega32 AVR uC + SoftVGA = 400line 70Hz"
  inc S0

  rcall XYWSTRI
  .dw 40
  .db "18.432MHz, 3clk/pixl, 4pixl/char, 40char"
  inc S0

  rcall XYWSTRI
  .dw 40
  .db "4x8pixel font, 95 ASCII, 16 blockgraphic"
  inc S0

  ret


; --- setup 7 RBG+derivatives coloured text on black background lines

; parameters: S0: row in frame buffer to begin at
;             T0: main foreground colour, rest of colours derived from this

DERGB_1X40:

  push T0                ; save for multiple reuse

  ldi T0,0x00            ; black
  ldi T1,0               ; set background
  rcall XYWCOLOUR

  pop T0
  push T0

  ldi T1,1               ; set foreground to colur T0
  rcall XYWCOLOUR

  pop T0
  push T0

  andi T0,0x2A           ; 2/3 bright, kill bit0 of each DAC
  ldi T1,2               ; set alternate colour 1
  rcall XYWCOLOUR

  pop T0
  push T0

  andi T0,0x15           ; 1/3 bright, kill bit1 of each DAC
  ldi T1,3               ; set alternate colour 2
  rcall XYWCOLOUR

  pop T0

  ori T0,0x2A            ; 2/3 pastel, force bit1 of all DACs
  ldi T1,4               ; set alternate colour 3
  rcall XYWCOLOUR

  inc S0

  ret


; parameters: S0: row in frame buffer to begin at

DERGB_7X40:

  ldi T0,0x03            ; blue
  rcall DERGB_1X40
  ldi T0,0x0C            ; green
  rcall DERGB_1X40
  ldi T0,0x0F            ; cyan
  rcall DERGB_1X40
  ldi T0,0x30            ; red
  rcall DERGB_1X40
  ldi T0,0x33            ; magenta
  rcall DERGB_1X40
  ldi T0,0x3C            ; yellow
  rcall DERGB_1X40
  ldi T0,0x3F            ; white
  rcall DERGB_1X40

  subi S0,7              ; back up the 7 rows, for drawing stuff

  ret


; --- draw multistyle/-coloured text

; parameters: S0: row in frame buffer to begin at
;             S1: column in frame buffer to begin at

DEMULT_1X40:

  rcall XYWSTRI
  .dw 40
  .db "nor", IVON, "inv", IVOFF, "nor", ULON, "ulin", ULOFF, "nor"
  .db COL1, "2/3", COL1, "nor", COL2, "1/3", COL2, "nor", COL3, "pas"
  inc S0

  ret


DEMULT_7X40:

  rcall DEMULT_1X40      ; make 7 lines with this
  rcall DEMULT_1X40
  rcall DEMULT_1X40
  rcall DEMULT_1X40
  rcall DEMULT_1X40
  rcall DEMULT_1X40
  rcall DEMULT_1X40

  ret


; --- draw website message

; parameters: S0: row in frame buffer to begin at
;             S1: column in frame buffer to begin at

DEWEB_2X40:

  rcall XYWSTRI
  .dw 40
  .db "design and code open source available at"
  inc S0

  rcall XYWSTRI
  .dw 40
  .db "http://neil.franklin.ch/Projects/SoftVGA"
  inc S0

  ret


; --- draw entire screen

DESCREEN:

  push S0
  push S1

  ldi S0,0               ; row = 0, start at top of screen
  ldi S1,0               ; column = 0, start at left of screen


  rcall DEASC_3X40       ; lines 0..2

  rcall DEBLACK          ; line 3

  rcall DERAIN_6X40      ; lines 4..9
  rcall DEGRAPH_6X40

  rcall DEBLACK          ; line 10

  rcall DEANN_3X40       ; lines 11..13

  rcall DEBLACK          ; line 14

  rcall DERGB_7X40       ; lines 15..21
  rcall DEMULT_7X40

  rcall DEBLACK          ; line 22

  rcall DEWEB_2X40       ; lines 23..24

  rcall DEBLACK          ; line 25

                         ; leave remaining 4 lines 26..29 of dots unused


  pop S1
  pop S0

  ret


; --- rotate ASCII char set background colour

; parameters: S0: row in frame buffer to rotate colour

DEROTBG_1X40:

  ldi T1,0               ; background
  rcall XYRCOLOUR        ; read existing colour
  subi T0,-0x10          ; only modify red, 4 steps
  rcall XYWCOLOUR        ; write new colour

  ret


; parameters: S2: frame count 0..255

DEROTBG_3X40:

  sbi PORTD,PIND4
  cbi DDRD,PIND4
  sbic PIND,PIND4        ; only if switch on PortD4 activated
    ret                  ;   no switch, no rotate, abort

  mov T0,S2              ; rotate every 64th frame, ~1s
  andi T0,0x3F           ; frame 0|64|128|192|0|...?
    brne DERONOT         ;   no, do nothing

    push S0              ;   yes, rotate background colour

    ldi S0,0             ;   first row of ASCII char set
    rcall DEROTBG_1X40
    inc S0
    rcall DEROTBG_1X40
    inc S0
    rcall DEROTBG_1X40

    pop S0

  DERONOT:

  ret


; --- move andalusian video snail across blank row

; parameters: S2: frame count 0..255

DESNAIL_1X40:

  push S0
  push S1

  ldi S0,22              ; black row just above web address message

  lds S1,0
  rcall DEBLANK          ; restore blank in case switch off, remove old snail
  subi S0,1              ; back up the row, for drawing new snail

  sbi PORTD,PIND3
  cbi DDRD,PIND3
  sbic PIND,PIND3        ; only if switch on PortD3 activated
    rjmp DESNNOT         ;   no switch, no rotate, abort

  lds S1,DESNPOS         ; wherever snail was drawn last time

  mov T0,S2              ; move every 16th frame, ~1/4s
  andi T0,0x0F           ; frame 0|16|32|48|..|240|0|...?
    brne DESNSTAY        ;   no, leave column unchanged
    inc S1               ;   yes, move snail one column to the right
    cpi S1,FBCOLS        ;   after last column?
      brlo DESNCOLOK     ;     <, column OK
      ldi S1,0           ;     >=, back to the left column
    DESNCOLOK:
    sts DESNPOS,S1       ;   store any changes
  DESNSTAY:

  rcall XYWSTRI
  .dw 4
  .db "_@_/"             ; the traditional 1-liner ASCII art
  inc S0

  DESNNOT:

  pop S1
  pop S0

  ret


; --- bounce ball in middle of graohic

; parameters: S0: row in frame buffer to begin at
;             S1: column in frame buffer to begin at

DEBOFIELD_1X10:

  rcall XYWSTRI
  .dw 10
  .db FS0000, FS1001, FS0000, FS0000, FS0000, FS0000
  .db FS0000, FS0000, FS1001, FS0000
  inc S0

  ret


DEBOFIELD_8X10:

  rcall XYWSTRI
  .dw 10
  .db FS0000, FS0011, FS0110, FS0110, FS0110, FS0110
  .db FS0110, FS0110, FS0101, FS0000
  inc S0

  rcall DEBOFIELD_1X10
  rcall DEBOFIELD_1X10
  rcall DEBOFIELD_1X10
  rcall DEBOFIELD_1X10
  rcall DEBOFIELD_1X10
  rcall DEBOFIELD_1X10

  rcall XYWSTRI
  .dw 10
  .db FS0000, FS1010, FS0110, FS0110, FS0110, FS0110
  .db FS0110, FS0110, FS1100, FS0000
  inc S0

  ret


DEBOBALL_3X3:

  rcall XYWSTRI
  .dw 3
  .db FR0011, FR0110, FR0101, 0
  inc S0

  rcall XYWSTRI
  .dw 3
  .db FR1001, FR0000, FR1001, 0
  inc S0

  rcall XYWSTRI
  .dw 3
  .db FR1010, FR0110, FR1100, 0
  inc S0

  ret


; parameters: S2: frame count 0..255

DEBOUNCE_8X40:

  push S0
  push S1

  ldi S0,3               ; blank row before graphics

  ldi S1,0
  rcall DEBLACK          ; restore blank in case switch off, remove old
  rcall DEGRAPH_6X40     ; restore graphics in case switch off, remove old
  rcall DEBLACK          ; restore blank in case switch off, remove old
  subi S0,8              ; back up the 1+6+1=8 rows, for box and ball

  sbi PORTD,PIND2
  cbi DDRD,PIND2
  sbic PIND,PIND2        ; only if switch on PortD2 activated
    rjmp DEBONOT         ;   no switch, no new field and ball, abort

  ldi S1,15              ; cut back ASCII waves to 15 wide
  rcall DEBOFIELD_8X10   ;   and add empty bounce field, 1+1+6+1+1 space
  subi S0,8              ; back up the 8 rows, for drawing ball

  inc S0                 ; enter field vertically
  inc S1                 ;   and also horizontally
  inc S1

  mov T0,S2              ; move every 8th frame, ~1/8s
  lsr T0                 ; so divide by 8
  lsr T0
  lsr T0
  mov T1,T0              ; save for column direction

  andi T0,0x07           ; extract 3 bits for 8 (2*4) ball positions
  ldi T2,7
  sbrc T0,2              ; range 4..7 or 0..3
    sub T2,T0            ;   4..7 -> 3..0
  sbrc T0,2              ; range 4..7 or 0..3
    mov T0,T2
  add S0,T0              ; add to row direction

  lsr T1                 ; every 16th frame, so divide by 16
  inc T1                 ; phase shift by 1/8th, 45 degrees

  andi T1,0x07           ; extract 3 bits for 8 (2*4) ball positions
  ldi T2,7
  sbrc T1,2              ; range 4..7 or 0..3
    sub T2,T1            ;   4..7 -> 3..0
  sbrc T1,2              ; range 4..7 or 0..3
    mov T1,T2
  add S1,T1              ; add to column direction

  rcall DEBOBALL_3X3

  DEBONOT:

  pop S1
  pop S0

  ret


; --- blink cursor in first announce line

; parameters: S2: frame count 0..255

DECURSOR_3X40:

  push S0
  push S1

  ldi S0,11              ; first row of announce message

  ldi S1,0
  rcall DEANN_3X40       ; restore announce in case switch off, remove old
  subi S0,3              ; back up the 3 rows, for drawing separator

  sbi PORTD,PIND1
  cbi DDRD,PIND1
  sbic PIND,PIND1        ; only if switch on PortD1 activated
    rjmp DECUNOT         ;   no switch, no new cursor, abort

  sbrc S2,0              ; alternating 2 frames draw cursor or character
    rjmp DECUNOT         ;   2nd, do nothing, draws character

  sbrc S2,4              ; alternating 2*16 frames blink on or off, ~1/2s
    rjmp DECUNOT         ;   16..31, do nothing, cursor off

  lds S1,DECUPOS         ; wherever cursor was drawn last time

  mov T0,S2              ; move every 2nd blink, every 64th frame, ~1s
  andi T0,0x3F           ; frame 0|64|128|192|0|...?
    brne DECUSTAY        ;   no, leave column unchanged
    inc S1               ;   yes, move cursor one column to the right
    cpi S1,FBCOLS        ;   after last column?
      brlo DECUCOLOK     ;     <, column OK
      ldi S1,0           ;     >=, back to the left column
    DECUCOLOK:
    sts DECUPOS,S1       ; store any changes
  DECUSTAY:

  ;ldi T0,0x5F            ; draw "_" character as cursor
  ldi T0,BG1111           ; draw "_" character as cursor
  rcall XYWCHAR

  DECUNOT:

  pop S1
  pop S0

  ret


; --- show something in frame buffer

DEMO:

  rcall DESCREEN

  ldi S2,0               ; frame counter/timer, no conflict with S0 or S1
  sts DESNPOS,S2
  sts DECUPOS,S2

  ENDLESS:

    rcall VRWAIT         ; wait for next frame vertical blank
    inc S2               ; update frame counter

    rcall DEROTBG_3X40

    rcall DESNAIL_1X40

    rcall DEBOUNCE_8X40

    rcall DECURSOR_3X40

    rjmp ENDLESS


; ------ main program

; --- handle reset, initialise system, start display, main loop

RESET:

  ; get stack ready, for calls and interrupts
  ldi ZL,low(MAXRAM)
  ldi ZH,high(MAXRAM)
  out SPL,ZL
  out SPH,ZH

  ; and then enable interrupts
  sei

  ; set up frame buffer
  rcall FBINIT

  ; set up port(s) for generating sync pulses and control switch
  rcall SYNCINIT

  ; set up registers and ports for DACs and control switches
  rcall CHINIT

  ; start display timer and ISR
  rcall DISPLAY

  ; enter demo, with endless loop, use up background time
  ; fill frame buffer to show something
  rjmp DEMO
