; vga_text_static.asm - generate an static VGA text
; author Neil Franklin, last modification 2008.09.29


; ------ hardware specifics

; need fuses, fast external quarz osc, max stable startup
; FF: 7:BODLVL=1, 6:BOD=1 (disable), 54:SUT=11, 3210:CKSEL=1111 (ext resonator)
; D9: 7:OCDEN=1 (dis), 6:JTAGEN=1 (dis), 5:SPIEN=0 (en), 4:CKOPT=1
;     3:EESAVE=1 (not preserve), 21:BOOTSZ=00, 0:BOOTRST=1 

; PORTC is connected to an VGA HD15 adapter, 3*2bit DACs and 2 TTLs
;   bit0+1 blue DAC, bit2+3 green DAC, bit4+5 red DAC
;   bit6 hsync TTL, bit7 vsync TTL


; we are using an ATmega32
  .include "m32def.inc"


; ------ handle reset and interrupts (all identical as not used)

; --- set up vectors

  .cseg
  .org $0000

  jmp RESET
  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR

  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR

  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR

  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR

  jmp INTR


; --- handle interrupts, despite not used

INTR:

  ; just return, nothing else makes sense here
  reti


; ------ auxillary routines

; --- wait a while doing nothing, spinloop

; parameters: TL loop count, 1..256, gets destroyed (reduced to 0)

; timing (TL-1)*3+2+4 = TL*3-1+4 = TL*3+3 clocks, for TL=[1..255,0(=256)]
;   min 1*3-1+4=11 clocks for TL=1, max 256*3-1+4=776 clocks for TL=0(=256)

WAIT:

                         ; {n} = clocks used, for timing computations

  ; TL set up by caller  ; {0}  for count = TL..1
  WAILOOP:

    ; nothing in loop, for highest loop resolution

    dec TL               ; {1}  next count
    brne WAILOOP         ; {2|1}

  ret                    ; {4}


; ------ vertical retrace stuff

; general drawing sequence:
;   frame
;     start with vertical retrace pulse lines, first deviation from all zero
;     then vertical retrace blank top lines
;     then visible output lines
;     then vertical retrace blank bottom lines
;   lines
;     start with horizontal retrace pulse time, first deviation from all zero
;     then horizontal retrace blank left time
;     then visible output pixels
;     then horizontal retrace blank right time


; --- set up registers for generating sync pulses

; constants for sync pulse patterns
  .equ NOSYNC = 0x00
  .equ HSYNC  = 0x40
  .equ VSYNC  = 0x80
  .equ HVSYNC = 0xC0

; registers for sync pulse patterns, needed for  out PORTC,value-reg
  ; use R2..R15 non-immediate range, as one load and permanent use
  .def NONE = R15
  .def HOR = R14
  .def VER = R13
  .def H_V = R12

RETRINIT:

  ; set registers for sync patterns
  ldi TL,NOSYNC
  mov NONE,TL

  ldi TL,HSYNC
  mov HOR,TL

  ldi TL,VSYNC
  mov VER,TL

  ldi TL,HVSYNC
  mov H_V,TL

  ; get PortC data ready
  out PORTC,NONE

  ret


; --- draw vertical retrace pulse line

; start with horizontal retrace pulse, as first deviation from all zero
; end with 1st cycle of blank before next pulse, rest for callers control

; timing 30+60 + 480 + 1+4 = 575(of600) clocks in line, 25spare

VPLINE:

  ; {30}  horizontal retrace pulse, V = 1, H = 1, color = BLK
  out PORTC,H_V   ; {1}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}

  ; {60+480=540}  horiz retrace blank, V = 1, H = 0, color = BLK
  out PORTC,VER   ; {1}
  ldi TL,177      ; {1}
  call WAIT       ; {4+(177*3+3)=538}

  ; horiz end blank, after last pixel, no wait, give rest of time to caller
  out PORTC,VER   ; {1}  no signal effect, just for regular timing of all lines
  ret             ; {4}


; --- draw vertical retrace blank or end line (or any other blank line)

; function, start, end identical to VPLINE, but without VSYNC pin active

; timing identical to VPLINE, 25spare

VBLINE:

  ; {30}  horizontal retrace pulse, V = 0, H = 1, color = BLK
  out PORTC,HOR   ; {1}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}

  ; {60+480=540}  horiz retrace blank, V = 0, H = 0, color = BLK
  out PORTC,NONE  ; {1}
  ldi TL,177      ; {1}
  call WAIT       ; {4+(177*3+3)=538}

  ; horiz end blank, after last pixel, no wait, give rest of time to caller
  out PORTC,NONE  ; {1}  no signal effect, just for regular timing of all lines
  ret             ; {4}


; ------ visible output drawing stuff

; --- set up registers for generating colour signals

; constants for colour patterns

  .equ BLACK   = 0x00
  .equ BLUE    = 0x03
  .equ GREEN   = 0x0C
  .equ CYAN    = 0x0F
  .equ RED     = 0x30
  .equ MAGENTA = 0x33
  .equ YELLOW  = 0x3C
  .equ WHITE   = 0x3F

; registers for colour patterns, needed for  out PORTC,value-reg
  ; use R2..R15 non-immediate range, as one load and permanent use
  .def BACK = R11
  .def FORE = R10

COLINIT:

  ; set registers for colour patterns
  ldi TL,BLUE
  mov BACK,TL

  ;ldi TL,YELLOW
  ;ldi TL,WHITE
  ;ldi TL,CYAN
  ;ldi TL,0x1F     ; 3 attempts at light blue
  ldi TL,0x1B
  ;ldi TL,0x2B
  mov FORE,TL

  ret


; --- initialise frame buffer

; constants for frame buffer size

  .equ FBCOLS  = 40   ; columns of chars in one row
  .equ FBCHARS = 41   ; colums + 1 (abort pseudo-char)

  .equ FBROWS  = 25   ; rows of characters

; define frame buffer

  ; starts from begin of SRAM, as nothing else in SRAM, exept stack at end
  ;   uses FBROWS*FBCHARS, 25*41=1025, of 2048bytes
  .equ FBRAM = MINRAM

; contents of frame buffer

  .equ ABORT = 0x07   ; leave 0..6=7 times 128instr/256bytes space before font
  .equ FIRST = 0x08   ; support font with chars from 08 on
  .equ RANGE = 0x03   ; how many characters do we have currently, 0..3

FBINIT:

  ldi XL,low(FBRAM)   ; get memory for video frame buffer
  ldi XH,high(FBRAM)

  ldi TH,FBROWS       ; for row-in-frame = FBROWS..1
  FBRLOOP:

                      ;   start with char 0..3, gives lines 1 char shifted
    ldi AL,FBROWS     ;   TH=FBROWS..1, we need 0..(FBROWS-1)
    sub AL,TH

    andi AL,RANGE     ;   cut down to range 0..3
    ori AL,FIRST      ;     and add base char code

    ldi TL,FBCOLS     ;   for column-in-row = FBCOLS..1
    FBCLOOP:

      st X+,AL        ;     place char

      inc AL          ;     next char
      andi AL,RANGE   ;       cut down to range 0..3
      ori AL,FIRST    ;       add base char code

      dec TL          ;   next column-in-row
      brne FBCLOOP

    ldi AL,ABORT      ;   place pseudo-char for abort at end of row
    st X+,AL

    dec TH            ; next row-in-frame
    brne FBRLOOP

  ret


; --- draw character based visible output line

; parameters: XH:XL first char SRAM address
;             TH font line segment offset, 0x00, 0x10, .. 0x70

; function, start, end identical to VPLINE, but without VSYNC pin active
;   and with pixels been drawn in the last 480 of the 540 clocks

; timing identical to VPLINE and BLLINE, 25spare

CHLINE:

  ; {30}  horizontal retrace pulse, V = 0, H = 1, color = BLK
  out PORTC,HOR   ; {1}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}

  ; {60}  horizontal retrace blank, V = 0, H = 0, color = BLK
  ;   {60-8=52} partial wait, give rest of time to first char setup
  out PORTC,NONE  ; {1}
  ldi TL,14       ; {1}
  call WAIT       ; {4+(14*3+3=49}
  nop             ; {1}

  ; {8} while horizontal retrace
  ; get first character code and unthread
  ;   uses standard font unthreading technique, see there for generic case
  ;   but compressed, without 4 clocks of intermixed pixel OUT statements
  ;   and adding variable segment offset
  ; XH:XL address has already been set up by caller
  ld ZL,X+        ; {2}  next (=first) character: get code
  mov ZH,ZL       ; {1}  next (=first) character: extract top part
  andi ZH,0x3F    ; {1}    for ATmega32: limit 2*64 chars of each 128instr
  andi ZL,0x80    ; {1}  next (=first) character: extract bottom part
  ; TH pattern has already been set up by caller, because variable
  or ZL,TH        ; {1}    add in segment offset for this line in row
  ijmp            ; {2}  next (=first) character: go and draw it
                  ;        continuation at that chars pixel 1, including above

  ; {40char*4pixel*3=480}  draw character segments, V = 0, H = 0, color = pixel
  ; for each X+ addressed SRAM byte index one segment, until abort charcode
  ;   so no code here, all in font, until abort handler gets called
  ; the abort handler does all "end of line"  out PORTC,NONE  and  ret  stuff
  ;   see pseudo-character code 07 (at bottom of this file) for that


; ------ generic line triggering

; --- draw line, select type it shall be and jump to that

; parameters: TH:TL line number in frame, 0..(LinesPerFrame-1)
;   for   0..(  0+ 12-1=) 11 jump to VPLINE, vertical retrace pulse line
;   for  12..( 12+ 25-1=) 36 jump to VBLINE, vertical retrace blank line
;   for  37..( 37+400-1=)436 jump to CHLINE, character line
;     for these also compute the rows SRAM address and segment offset
;   for 437..(437+ 13-1=)449 jump to VBLINE, vertical retrace end line

; all called line routines:
;  start with horizontal retrace pulse, as first deviation from all zero
;   end with 1st cycle of blank before next pulse, rest for callers control
;   timing 30+60 + 480 + 1+4 = 575(of600) clocks in line, 25spare

DRLINE:


; ------ frame timing control stuff

; at present one big monolithic block, to be split up


; --- draw an entire frame

; start with ver (+ hor) retrace pulse, as first deviation from all zero
; end with 1st cycle of last blank before next pulse, rest for callers control

; timing 12+25 + (25*16=)400 + 12(of13) = 449(of450) lines, each of 600 clocks
;   + 4+575 + 4 = 583(of600) remaining clocks of partial last line, 17spare

; registers for loop counting
  .def FRLINE = R16
  .def FRROW  = R17

FRAME:

  ; {1setup+12*600-1brne}  vertical retrace pulse, must be 12 lines
  ldi FRLINE,12        ; {1}  for line-in-pulse = 12..1
  FRPLOOP:             ; {12*600-1}

    call VPLINE        ; {4+575=579, 21spare}
    ldi TL,3           ; {1}
    call WAIT          ; {4+(3*3+3)=16}
    nop                ; {1}

    dec FRLINE         ; {1}  next line-in-pulse
    brne FRPLOOP       ; {2|1}

  ; {1setup+24*600}  vertical retrace blank, must be 24(of25) lines
  ldi FRLINE,25        ; {1, compensates brne 2|1}  for line-in-blank = 24..1
  FRBLOOP:             ; {24*600-1}

    call VBLINE        ; {4+575=579, 21spare}
    ldi TL,3           ; {1}
    call WAIT          ; {4+(3*3+3)=16}
    nop                ; {1}

    dec FRLINE         ; {1}  next line-in-blank
    brne FRBLOOP       ; {2|1}

  nop                  ; {1, compensates brne 2|1, as no ldi}

  ; {600}  25th vertical retrace blank, last one
  ;   {600-6setup=594} partial wait, give rest of time to char loop setup
  call VBLINE          ; {4+575=579, 21-6setup=15spare}
  ldi TL,2             ; {1}
  call WAIT            ; {4+(2*3+3)=13}
  nop                  ; {1}

  ; {6setup+25*16*600-1brne}  draw character lines
  ;   must be 24(of25)rows * 16lines-per-row = 384(of400) lines
  ldi YL,low(FBRAM)    ; {1}  get frame buffer first row begin
  ldi YH,high(FBRAM)   ; {1}

  ldi FRROW,FBROWS     ; {1}  for row-in-frame = FBROWS..1
  FRRLOOP:             ; {24*16*600-1brne}

    ldi TH,0x00        ; {1}  start row with first line segment offset

    ldi FRLINE,7       ; {1}  for linepair-in-row = 7(of8)..1
    FRLLOOP:           ; {16*600-1brne}

      movw XH:XL,YH:YL ; {1}  XH:XL for this line, YH:YL survive for next

      ; {24*16*600}  call must be synchronous with previous VPLINE and VBLINE
      call CHLINE      ; {4+575=579, 21spare, -1movw=20}
      ldi TL,4         ; {1}
      call WAIT        ; {4+(4*3+3)=19}

      movw XH:XL,YH:YL ; {1}  XH:XL for this line, YH:YL survive for next

      ;call CHLINE      ; {4+575=579, 21spare, -6seg+loop+movw=15}
      call VBLINE      ; {4+575=579, 21spare, -6seg+loop+movw=15}
      ldi TL,2         ; {1}
      call WAIT        ; {4+(2*3+3)=13}
      nop              ; {1}

      subi TH,-0x10    ; {1}  move line segment offset by 16instr/32bytes
      andi TH, 0x70    ; {1}  keep it in range, also after exiting loop

      dec FRLINE       ; {1}
      brne FRLLOOP     ; {2|1}  next linepair-in-row

    nop                ; {1}  compensates brne 2|1, as no ldi}

    ; {600} 8th line, last one in row, no wait, give rest of time to row stepping
    movw XH:XL,YH:YL   ; {1, precompensated in 6setup}

    call CHLINE        ; {4+575=579, 21spare, -1movw=20}
    ldi TL,4           ; {1}
    call WAIT          ; {4+(4*3+3)=19}

    movw XH:XL,YH:YL   ; {1}

    ;call CHLINE        ; {4+575=579, 21spare, -8seg+frame+loops+setup=13}
    call VBLINE        ; {4+575=579, 21spare, -8seg+frame+loops+setup=13}
    ldi TL,1           ; {1}
    call WAIT          ; {4+(1*3+3)=10}
    nop                ; {2}
    nop

    adiw YH:YL,FBCHARS ; {2}  jump video frame buffer to next row

    dec FRROW          ; {1}  next row-in-frame
    brne FRRLOOP       ; {2|1}

  ; {1setup+12*600}  vertical retrace end, must be 12(of13) lines
  ldi FRLINE,12        ; {1, compensates brne 2|1}  for line-in-end = 12..1
  FRELOOP:

    call VBLINE        ; {4+575=579, 21spare}
    ldi TL,3           ; {1}
    call WAIT          ; {4+(3*3+3)=16}
    nop                ; {1}

    dec FRLINE         ; {1}  next line-in-end
    brne FRELOOP       ; {2|1}

  nop                  ; {1, compensates brne 2|1, as no ldi}

  ; 13th line, last one in frame, no wait, give rest of time to caller
  call VBLINE          ; {4+575=579, 21spare}
  ret                  ; {4}


; ------ main program

; --- main loop, infinite loop, endlessly draw frames

LOOP:

    call FRAME    ; {4+(449*600+583), 13 to spare}
    nop           ; {10, ldi TL,1 + call WAIT only from 11 on upwards}
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    jmp LOOP      ; {3}


; --- handle reset, initialise system, enter main loop

RESET:

  ; get stack ready, for calls and interrupts (later not used)
  ldi TL,low(MAXRAM)
  ldi TH,high(MAXRAM)
  out SPL,TL
  out SPH,TH

  ; get interrupts ready, despite not used
  sei


  ; set up retrace sync pulse register patterns, and get port C ready
  call RETRINIT

  ; get port C DDR register ready
  ldi TL,0xFF   ; all out
  out DDRC,TL


  ; set up colour register patterns
  call COLINIT

  ; set up frame buffer
  call FBINIT

  ; enter endless loop, no return back to here
  jmp LOOP


; ------ character processing/drawing routines

; no labels as this code is only ever reached by computed ijmp
;   same as the entire character drawing code in the font file

; per character 8 line segments
;   per segment size 16 instructions a 2 bytes, = 128instr-or-256bytes/char
;     on ATmega32 with 32k this allows 32k/256=128 chars
;       indexable as 2*64, with 00 to 3F and 80 to BF
;   per segment time 4 pixels a 3 clocks = 12 clocks/line


; per character definition code:

; 1 per char setup:
; .org $cco0      ; start character code cc first segment offset o at $cco0
;                       cc = 08..3x or 80..BF
;                         .org and ijmp word addressed, 8*16instr, not 8*32byte
;                         characters interleaved:  08 88 09 89 0A 8A ... 3F BF
;                       o = cc00 for 0x..3x and cc80 for 8x..Bx

; 8 per line sections:
;   size 16 instructions, time 12 clocks
;     4 clocks write pixels with  out PORTC,xxxx (xxxx = FORE or BACK)
;     between these 4*2 clocks for unthreading next character jump

; out PORTC,xxxx  ; {1}  current character: pixel 1
; ld ZL,X+        ; {2}  next character: get code

; out PORTC,xxxx  ; {1}  current character: pixel 2
; mov ZH,ZL       ; {1}  next character: extract top part
; andi ZH,0x3F    ; {1}    for ATmega32: limit to 2*64 chars of each 128instr

; out PORTC,xxxx  ; {1}  current character: pixel 3
; andi ZL,0x80    ; {1}  next character: extract bottom part
; ori ZL,0xo0     ; {1}    add in segment offset for this line in row
                  ;          o = 0x00 0x10 .. 0x70 or o = 0x80 0x90 .. 0xF0

; out PORTC,xxxx  ; {1}  current character: pixel 4, = between characters
; ijmp            ; {2}  next character: go and draw it
                  ;        continuation at that chars pixel 1, as above

; nop             ; overjump rest of unused space, 16-10=6 instructions
; nop
; nop
; nop
; nop
; nop


; overjump program code space, results a few in unusable character codes
;   we currently use character  07  as first one, the abort character
;     this allows 0..6=7 * 128instr/256bytes of code, safely more than needed
;   then use characters  08..3F  and  80..BF  as drawing characters


; pseudo-character code 07, abort line by pseudo-char "drawing" routines

  .org $0700

  ; horiz end blank, after last pixel, no wait, give rest of time to caller
  out PORTC,NONE  ; {1}  switch off whatever was last background colour
  ret             ; {4}
  nop             ; overjump rest of unused space, 16-2=14 instructions
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  ret
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  ret
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  ret
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  ret
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  ret
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  ret
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  ret
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

; space 0780 to 07FF not used presently, would be character code 87

; as last include font, may contain character codes 08 to 3F and 80 to BF
  .include "vga_text_font.inc"
