; vga_text_static.asm - generate an static VGA text
; author Neil Franklin, last modification 2008.10.03


; ------ intro

; general drawing sequence:
;   lines, are the time critical thing
;     first visible output pixels
;     then horizontal retrace blank right time
;     then horizontal retrace pulse time
;     then horizontal retrace blank left time, largest spare time
;   frame, for consistency with lines
;     first visible output lines
;     then vertical retrace blank bottom lines
;     then vertical retrace pulse lines
;     then vertical retrace blank top lines, largest spare time


; ------ hardware specifics

; need fuses, fast external quarz osc, max stable startup
; FF: 7:BODLVL=1, 6:BOD=1 (disable), 54:SUT=11, 3210:CKSEL=1111 (ext resonator)
; D9: 7:OCDEN=1 (dis), 6:JTAGEN=1 (dis), 5:SPIEN=0 (en), 4:CKOPT=1
;     3:EESAVE=1 (not preserve), 21:BOOTSZ=00, 0:BOOTRST=1 

; Port C is connected to an VGA HD15 adapter, 3*2bit DACs and 2 TTLs
;   PC0+PC1 blue DAC, PC2+PC3 green DAC, PC4+PC5 red DAC
;   PC6 hsync TTL, PC7 vsync TTL
; Port D is connected to an mode switch
;   PD6 (next to GND on 10pin header), 0(jumper) = double, 1(default) = blank


; we are using an ATmega32
  .include "m32def.inc"


; ------ handle reset and interrupts (all identical as not used)

; --- set up vectors

  .cseg
  .org $0000

  jmp RESET
  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR

  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR

  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR

  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR

  jmp INTR


; --- handle interrupts, despite not used

INTR:

  ; just return, nothing else makes sense here
  reti


; ------ auxillary routines

; --- wait a while doing nothing, spinloop

; parameters: TL loop count, 1..256
;             gets destroyed (reduced to 0), avoid push/pop time, seldom reuse

; timing (TL-1)*3+2+4 = TL*3-1+4 = TL*3+3 clocks, for TL=[1..255,0(=256)]
;   min 1*3-1+4=11 clocks for TL=1, max 256*3-1+4=776 clocks for TL=0(=256)

WAIT:

                         ; {n} = clocks used, for timing computations
                         ;   this program requires exact clock counting
                         ;   single-clock miscounts produce visual artifacts

  ; TL set up by caller  ; for count = TL..1
  WAILOOP:

    ; nothing in loop, for highest loop resolution, only 3 clocks

    dec TL               ; {1}  next count
    brne WAILOOP         ; {2|1}

  ret                    ; {4}


; ------ vertical retrace stuff

; --- set up registers for generating sync pulses

; constants for sync pulse patterns
  .equ NOSYNC = 0x00
  .equ HSYNC  = 0x40
  .equ VSYNC  = 0x80
  .equ HVSYNC = 0xC0

; registers for sync pulse patterns, needed for  out PORTC,value-reg
  ; use R2..R15 non-immediate range, as one load and permanent use
  .def NONE = R15
  .def HOR  = R14
  .def VER  = R13
  .def H_V  = R12

RETRINIT:

  ; set registers for sync patterns
  ldi TL,NOSYNC
  mov NONE,TL

  ldi TL,HSYNC
  mov HOR,TL

  ldi TL,VSYNC
  mov VER,TL

  ldi TL,HVSYNC
  mov H_V,TL

  ; get PortC data ready
  out PORTC,NONE

  ret


; --- draw vertical retrace top or bottom blank line (or any other blank line)

; timing 480 + 30 + 30 + 5 = 600-55spare

BLLINE:

  ; {480}  "draw" blank pixels, V = 0, H = 0, color = blank
  out PORTC,NONE  ; {1}
  ldi TL,157      ; {1}
  call WAIT       ; {4+(157*3+3)=478}

  ; {30}  horizontal retrace left blank, V = 0, H = 0, color = blank
  out PORTC,NONE  ; {1}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}

  ; {30}  horizontal retrace pulse, V = 0, H = 1, color = blank
  out PORTC,HOR   ; {1}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}

  ; {5of60} horizontal retrace right blank, V = 0, H = 0, color = blank
  out PORTC,NONE  ; {1}
  ; no wait, give rest of time to caller

  ret             ; {4}


; --- draw vertical retrace pulse line

; timing 480 + 30 + 30 + 5 = 600-55spare

VPLINE:

  ; {480}  "draw" blank pixels, V = 1, H = 0, color = blank
  out PORTC,VER   ; {1}
  ldi TL,157      ; {1}
  call WAIT       ; {4+(157*3+3)=478}

  ; {30}  horizontal retrace left blank, V = 1, H = 0, color = blank
  out PORTC,VER   ; {1}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}

  ; {30}  horizontal retrace pulse, V = 1, H = 1, color = blank
  out PORTC,H_V   ; {1}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}

  ; {5of60}  horizontal retrace right blank, V = 1, H = 0, color = blank
  out PORTC,VER   ; {1}
  ; no wait, give rest of time to caller

  ret             ; {4}


; ------ visible output drawing stuff

; --- define frame buffer

; define beginning address

  ; starts from begin of SRAM, as nothing else in SRAM, exept stack at end
  ;   uses FBROWS*FBROWLEN, 25*43=1075, of 2048bytes
  .equ FBRAM = MINRAM

; constants for frame buffer size

  .equ FBCOLS   = 40  ; columns of characters in one row
  .equ FBROWLEN = 43  ; legth of row in frame buffer
                      ;   2(colours) + 40(colums) + 1(abort)
  .equ FBROWS   = 25  ; rows of characters in one frame


; --- write char into frame buffer

; parameters: XH:XL where to place char
;             AL ASCII character

FBWCHAR:

  push AL

  cpi AL,0x20         ; below " "?
  brsh FBBOTTOMOK     ;   >=, is ok
  ldi AL,0x2E         ;   <, replace with ASCII "."
FBBOTTOMOK:
  
  cpi AL,0x7F         ; below DEL?
  brlo FBTOPOK        ;   <, is ok
  ldi AL,0x2E         ;   >=, replace with ASCII "."
FBTOPOK:

  bst AL,0            ; convert ASCII code to char code, 8bit rotate right
  lsr AL              ;   .654 3210 -> ..65 4321 (0)
  bld AL,7            ;             -> 0.65 4321

  st X+,AL            ; place this char

  pop AL
  ret


; --- initialise frame buffer

FBINIT:

  ldi XL,low(FBRAM)   ; get memory for video frame buffer
  ldi XH,high(FBRAM)

  ldi TH,FBROWS       ; for row-in-frame = FBROWS..1
  FBRLOOP:

                      ; background colour for this row
    ldi TL,0x03       ;   blue
    st X+,TL          ;   place this colour

                      ; foreground colour for this row
    ;ldi AH,0x3C       ; yellow
    ;ldi AH,0x3F       ; white
    ;ldi AH,0x0F       ; cyan
    ;ldi AH,0x1F       ; light blue, slight reddish cyan
    ;ldi AH,0x1B       ; light blue, sligh reddisch and lower green cyan, best
    ;ldi AH,0x2B       ; light blue, low red+green white
    ;st X+,AH          ; place this colour

                      ; each row with colour = row number
                      ;   gives each row different colour
    ldi AH,FBROWS     ;   TH=FBROWS..1, we need 0..(FBROWS-1)
    sub AH,TH
    andi AH,0x03      ;   cut down to range, we have DAC values 0..3
    lsl AH            ;   mult 16 for modifying the red component
    lsl AH
    lsl AH
    lsl AH
    ori AH,0x0F       ;   and add base colour, cyan
    st X+,AH          ; place this colour


                      ; start row with char ASCII derived from row number
                      ;   gives each row shifted 1 to left
    ldi AL,FBROWS     ;   TH=FBROWS..1, we need 0..(FBROWS-1)
    sub AL,TH

                      ; would show 40 + (25-1) = 64 chars, we want full 95, +31
    cpi AL,12         ;   below middle (12 lines, 3 4-line colour bands)?
    brsh FBBOTTOM     ;     >= middle, aim for touching end
      subi AL,-0x20   ;     < middle, add base char code, from ASCII 0x20
      jmp FBTOP
    FBBOTTOM:
      subi AL,-0x3F   ;     >=, add base, shifted for last chars, 0x20+31=0x3F
    FBTOP:


    ldi TL,FBCOLS     ; for column-in-row = FBCOLS..1
    FBCLOOP:

      call FBWCHAR    ; write char
      inc AL          ; next char

      dec TL          ; next column-in-row
      brne FBCLOOP


    ; after row abort pseudo-char
    ldi AL,0xBF       ; use ASCII DEL, aa=0x7F, hh=0x3F, ll=0x80, cc=0xBF
    st X+,AL          ;   place char

    dec TH            ; next row-in-frame
    brne FBRLOOP

  ret


; --- set up colour registers for drawing, are the LUT

; use R2..R15 non-immediate range, as no immediate load
  .def BACK = R11
  .def FORE = R10


; --- draw character based visible output line

; parameters: XH:XL first char in line SRAM address
;             TH font line segment offset, 0x00, 0x10, .. 0x70

; timing 4+8=12pre, 480 + 30 + 30 + 5 = 600-55spare

CHLINE:

  ; {*pre} "predraw" time, before drawing begins, precompensate this in caller

  ; {4pre} load LUT colour registers for this row
  ld BACK,X+      ; {2}  first 2 bytes in row in frame buffer are colours
  ld FORE,X+      ; {2}

  ; {8pre} get first character code and unthread it
  ;   use standard font unthreading technique, see there for generic case
  ;     but compressed, without 4 clocks of intermixed pixel OUT statements
  ;     and adding variable segment offset, for entering proper segment group
  ; XH:XL address has already been set up by caller
  ld ZL,X+        ; {2}  next (=first) character: get code
  mov ZH,ZL       ; {1}  next (=first) character: extract top part
  andi ZH,0x3F    ; {1}    for ATmega32: 0x3F limits address to 64(*2) chars
  andi ZL,0x80    ; {1}  next (=first) character: extract bottom part
  ; TH pattern has already been set up by caller, because variable
  or ZL,TH        ; {1}    add in segment offset for this line in row
  ijmp            ; {2}  next (=first) character: go and draw it
                  ;        continuation at that chars pixel 1, including above


  ; {(40*)12}  draw stuff, 40 char segments, V = 0, H = 0, color = pixel

  ; for each X+ fetched code: index and exec one line segment in font space
  ;   drawing done by  out PORTC,colour, one for each pixel, one every 3 clocks
  ;   {so per segment time 4 pixels a 3 clocks = 12 clocks/line}

  ; per character 8 such segments, one for each line in row
  ;   so per segment size 16 instructions * 2 bytes = 128instr/char
  ;     on ATmega32 32k Flash this allows 32k/2/128=128chars, minus used space
  ;   note that with max 6 drawable lines this could reduce to 6*10(+4==64instr
  ;     on ATmega32 this would allow 32k/2/64/=256chars, minus used space


  ; in font, once per character:
  ; .org $hhll    ; ASCII ".", aa=0xaa, hh=0xhh, ll=0xll, cc=0xcc
                  ;   because 8*16=128instr blocks use up lowest 7 address bits
                  ;     and we need fast 8bit indexing in unthreading code
                  ;   index as (aa/2)*2*128, gives char codes in font
                  ;     interleaved as  00 80 01 81 02 82 ... 3F BF
                  ;   compensate this by 8bit rotate right ASCII to char code
                  ;     aa=.654 3210, hh=..65 4321, ll=0... ...., cc=0.65 4321
                  ;   hh=0x00..3F for aa=0x00,02..7E and repeats aa=0x01,03..7F
                  ;   ll=0x00 for aa=0x00..3F and 0x80 for aa=0x8x..BF
  ;   no labels after .org, as font code is only ever reached by computed ijmp

  ; {12}  in font, 8 times per character, each line segment:
  ;         4 clocks write pixels with  out PORTC,xxxx (xxxx = FORE or BACK)
  ;         between these 4*2 clocks for unthreading next character jump

  ; out PORTC,xxxx ; {1}  current character: draw pixel 1
  ; ld ZL,X+      ; {2}  next character: get character code

  ; out PORTC,xxxx ; {1}  current character: draw pixel 2
  ; mov ZH,ZL     ; {1}  next character: copy and extract top part of index
  ; andi ZH,0x3F  ; {1}    for ATmega32: 0x3F limits address to 64(*2) chars

  ; out PORTC,xxxx ; {1}  current character: draw pixel 3
  ; andi ZL,0x80  ; {1}  next character: extract bottom part for (64*)2 chars
  ; ori ZL,0xo0   ; {1}    add in line segment offset for this line in row
                  ;          o = 0x00 0x10 .. 0x70 or o = 0x80 0x90 .. 0xF0

  ; out PORTC,xxxx ; {1}  current character: draw pixel 4
  ; ijmp          ; {2}  next character: go and execute it
                  ;        continuation at that chars pixel 1, as above

  ; nop           ; {0}  fill rest of unused space, 16-10=6 instruction words
  ; nop           ;        no execution time as never run!
  ; nop
  ; nop
  ; nop
  ; nop

  ; above continues until an  abort pseudo-char  code is hit in frame buffer
  ;   that pseudo-char exits the unthreadinging instead of drawing
  ;   must be done this way, because no time for 40 char loop counting


  ; {30}  horizontal retrace left blank, V = 0, H = 0, color = blank
  ;   pseudo-char directly runs time critical   out PORTC,NONE  then jumps back

  ; .org $hhll    ; ASCII ".", aa=0xaa, hh=0xhh, ll=0xll, cc=0xcc
  ;   no labels after .org, as font code is only ever reached by computed ijmp

  ; this code is in abort pseudo-character
  ; out PORTC,NONE ; {1}  switch off whatever was last background colour
  ; jmp CHEND     ; {3}  go back to CHLINE to make horizontal retrace pulse
  ; nop           ; {0}  fill rest of unused space, 16-1-2=13 instruction words
  ; nop           ;        no execution time as never run!
  ; nop           ;        followd by annother 10 nop instructions!
  ; like for every other char this is repeated for 8 line pseudo-segments

CHEND:

  ldi TL,6        ; {1}
  call WAIT       ; {4+(6*3+3)=25}

  ; {30}  horizontal retrace pulse, V = 0, H = 1, color = blank
  out PORTC,HOR   ; {1}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}

  ; {5of60}  horizontal retrace right blank, V = 0, H = 0, color = blank
  out PORTC,NONE  ; {1}
  ; no wait, give rest of time to caller

  ret             ; {4}


; --- draw pair of character based visible output lines
;   belonging to one font line, doublescan or 2nd line blank

; parameters: XH:XL first char in line SRAM address
;             TH font line segment offset, 0x00, 0x10, .. 0x70

; timing 4+12=16pre, 600 + 480 + 30 + 30 + 5 = 1200-55spare

CHPAIR:

  ; first line of linepair
  call CHLINE     ; {4+(12pre+600-55spare)}
  ldi TL,9        ; {1}
  call WAIT       ; {4+(9*3+3)=34}
  nop             ; {2}
  nop

  ; second line of linepair, same segment, same TH
  sbiw XH:XL,FBROWLEN ; {2}  return to frame buffer row begin, undo all X+

  ; switch for doublescan or 2nd blank
  sbis PIND,PIND6 ; {1|3}  1(default) = 2nd blank, 0(jumpered) = doublescan

  jmp CHLINE      ; {3+(12pre+600-55spare)}  doublescan: draw chars second time

  nop             ; {10}  correction, as BLLINE has no 12pre, unlike CHLINE
  nop             ;        give {12} together with {3-1=2} from the sbis
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  jmp BLLINE      ; {3+(600-55spare)}  2nd blank: draw an blank
  ; no wait, give rest of time to caller

  ; no ret here, as jmp = call+ret faster {3 instead of 4+4}


; ------ frame timing control loops stuff

; registers for loop counting
  .def FRLINE = R16
  .def FRROW  = R17


; --- draw all output lines of one row

; parameters: YH:YL first char in row SRAM address

; timing 1+1+1+4+16=23pre, (7*2+1)*600 + (600-55=)545 + 4 = 16*600-51spare

ROWDRAW:

  ldi TH,0x00          ; {1}  start row with first line segment offset

  ; loop through 7(of8) linepairs, 8th linepair without wait after it
  ldi FRLINE,7         ; {1}  for line-in-row = 7..1
  ROWLOOP:

    movw XH:XL,YH:YL   ; {1}  XH:XL for this line, YH:YL survive for next

    call CHPAIR        ; {4+(16pre+600-55spare)}
    ldi TL,7           ; {1}
    call WAIT          ; {4+(7*3+3)=28}

    ; segment switching for next linepair
    subi TH,-0x10      ; {1}  move line segment offset by 16instr/32bytes
    andi TH, 0x70      ; {1}  keep it in range, also after exiting loop

    dec FRLINE         ; {1}  next line-in-row
    brne ROWLOOP       ; {2|1}
                       ; {1+4+16 precompensate movw + call CHLINE + predraw}

  nop                  ; {1, compensates brne 2|1, as no ldi}

  ; 8th linepair, last one in row, no wait, give rest of time to caller
  movw XH:XL,YH:YL     ; {1}
  call CHPAIR          ; {4+(16pre+600-55spare)}

  ret                  ; {4}


; --- draw all output rows of one frame

; timing 2+2+4+23=31pre, 24*8*600 + 7*600 + 545 + 4 + 4 = 400*600-47spare

FRDRAW:

  ldi YL,low(FBRAM)    ; {1}  get frame buffer first row begin
  ldi YH,high(FBRAM)   ; {1}

  ; loop through 24(of25) rows, 25th row without wait after it

  ldi FRROW,FBROWS     ; {1}  for row-in-frame = (FBROWS-1)..1
  dec FRROW            ; {1}
  FRDLOOP:

    call ROWDRAW       ; {4+(23pre+16*600-51spare)}
    ldi TL,3           ; {1}
    call WAIT          ; {4+(3*3+3)=16}
    nop                ; {2}
    nop

    adiw YH:YL,FBROWLEN ; {2}  go to video frame buffer next row begin

    dec FRROW          ; {1}  next row-in-frame
    brne FRDLOOP       ; {2|1}
                       ; {4+23 precompensate call ROWDRAW + predraw}

  nop                  ; {1, compensates brne 2|1, as no ldi}

  ; 25th row, last one in frame, no wait, give rest of time to caller
  call ROWDRAW         ; {4+(23pre+16*600-51spare)}

  ret                  ; {4}


; --- draw all vertical retrace bottom blank lines of one frame

; timing 1+4=5pre, 11*600 + (600-55=)545 + 4 = 12*600-51spare

FRBOTTOM:

  ; loop through 11(of12) lines, 12th line without wait after it
  ldi FRLINE,11        ; {1}  for line-in-bottom = (12-1)..1
  FRBLOOP:

    call BLLINE        ; {4+(600-55spare)}
    ldi TL,13          ; {1}
    call WAIT          ; {4+(13*3+3)=46}
    nop                ; {1}

    dec FRLINE         ; {1}  next line-in-bottom
    brne FRBLOOP       ; {2|1}
                       ; {4 precompensate call BLLINE}

  nop                  ; {1, compensates brne 2|1, as no ldi}

  ; 12th line, last one in bottom, no wait, give rest of time to caller
  call BLLINE          ; {4+(600-55spare)}

  ret                  ; {4}


; --- draw all vertical retrace pulse lines of one frame

; timing 1+4=5pre, 12*600 + (600-55=)545 + 4 = 13*600-51spare

FRVPULSE:

  ; loop through 12(of13) lines, 13th line without wait after it
  ldi FRLINE,12        ; {1}  for line-in-vpulse = (13-1)..1
  FRVLOOP:

    call VPLINE        ; {4+(600-55spare)}
    ldi TL,13          ; {1}
    call WAIT          ; {4+(13*3+3)=46}
    nop                ; {1}

    dec FRLINE         ; {1}  next line-in-vpulse
    brne FRVLOOP       ; {2|1}
                       ; {4 precompensate call VPLINE}

  nop                  ; {1, compensates brne 2|1, as no ldi}

  ; 13th line, last one in vpulse, no wait, give rest of time to caller
  call VPLINE          ; {4+(600-55spare)}

  ret                  ; {4}


; --- draw all vertical retrace top blank lines of one frame

; timing 1+4=5pre, 24*600 + (600-55=)545 + 4 = 25*600-51spare

FRTOP:

  ; loop through 24(of25) lines, 25th line without wait after it
  ldi FRLINE,24        ; {1}  for line-in-top = (25-1)..1
  FRTLOOP:

    call BLLINE        ; {4+(600-55spare)}
    ldi TL,13          ; {1}
    call WAIT          ; {4+(13*3+3)=46}
    nop                ; {1}

    dec FRLINE         ; {1}  next line-in-bottom
    brne FRTLOOP       ; {2|1}
                       ; {4 precompensate call BLLINE}

  nop                  ; {1, compensates brne 2|1, as no ldi}

  ; 25th line, last one in top, no wait, give rest of time to caller
  call BLLINE          ; {4+(600-55spare)}

  ret                  ; {4}


; --- draw all sections of one frame

; timing 4+31=35pre, (25*8+12+13+24)*600 + 545+4 + 4 = 450*600-47spare

FRAME:

  call FRDRAW     ; {4+(31pre+400*600-47spare)}
  ldi TL,10       ; {1}
  call WAIT       ; {4+(10*3+3)=37}

  call FRBOTTOM   ; {4+(5pre+12*600-51spare)}
  ldi TL,11       ; {1}
  call WAIT       ; {4+(11*3+3)=40}
  nop             ; {1}

  call FRVPULSE   ; {4+(5pre+13*600-51spare)}
  ldi TL,11       ; {1}
  call WAIT       ; {4+(11*3+3)=40}
  nop             ; {1}

  call FRTOP      ; {4+(5pre+25*600-51spare)}
  ; no wait, give rest of time to caller

  ret             ; {4}


; ------ main program

; --- main loop, infinite loop, endlessly draw frames

LOOP:

  call FRAME      ; {4+35pre+(450*600-47)}
  nop             ; {5}
  nop
  nop
  nop
  nop
  jmp LOOP        ; {3}
                  ; {4+31 precompensate call FRAME + predraw}


; --- handle reset, initialise system, enter main loop

RESET:

  ; get stack ready, for calls and interrupts (later not used)
  ldi TL,low(MAXRAM)
  ldi TH,high(MAXRAM)
  out SPL,TL
  out SPH,TH

  ; get interrupts ready, despite not used
  sei


  ; set up retrace sync pulse register patterns, and get port C ready
  call RETRINIT

  ; get port C (VGA output) DDR register ready
  ldi TL,0xFF     ; DDR = 1 = output
  out DDRC,TL

  ; get port D6 (double/2ndblank mode sel) pullup and DDR ready
  sbi PORTD,PIND6 ; pullup = 1 = on
  cbi DDRD,PIND6  ; DDR = 0 = input


  ; set up frame buffer
  call FBINIT

  ; enter endless loop, no return back to here
  jmp LOOP


; ------ font, drawing characters and abort pseudo-character

; --- drawing characters

; overjump program code space, results in a few unusable character codes
;   expected to contain 2*(64-8)-1 chars, 0x08 0x88 0x09 0x89 .. 0x3F, no 0xBF
;     enough for 95 printable ASCII chars aa=0x20..0x7E (no non-draw DEL/0x7F)
;     and 16 more non-ASCII chars aa=0x10..1F before them
;   this allows for 8*2*128=2k instr of code, safely more than is needed

  ; with 0x10 pre-ASCII chars, from 0x10 on
  ;.org $0800     ; pre-ASCII, aa=0x10, hh=0x08, ll=0x00, cc=0x08
  ; with only standard ASCII chars, from 0x20 on
  .org $1000     ; pre-ASCII, aa=0x20, hh=0x10, ll=0x00, cc=0x10

  ; include font as an separate file, so that it may be generated automatically
  .include "vga_text_font.inc"


; --- abort pseudo-character "drawing" routines

  .org $3F80     ; use ASCII non-draw DEL, aa=0x7F, hh=0x3F, ll=0x80, cc=0xBF

  out PORTC,NONE ; 8 line segments repeated  out  jmp  13*nop
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
