; vga_text_static.asm - generate an static VGA text
; author Neil Franklin, last modification 2008.10.10


; ------ intro

; general drawing sequence:
;   lines, are the time critical thing
;     first visible output pixels
;     then horizontal retrace blank right time
;     then horizontal retrace pulse time
;     then horizontal retrace blank left time, largest spare time
;   frame, for consistency with lines
;     first visible output lines
;     then vertical retrace blank bottom lines
;     then vertical retrace pulse lines
;     then vertical retrace blank top lines, largest spare time


; ------ hardware specifics

; need fuses, fast external quarz osc, max stable startup
; FF: 7:BODLVL=1, 6:BOD=1 (disable), 54:SUT=11, 3210:CKSEL=1111 (ext resonator)
; D9: 7:OCDEN=1 (dis), 6:JTAGEN=1 (dis), 5:SPIEN=0 (en), 4:CKOPT=1
;     3:EESAVE=1 (not preserve), 21:BOOTSZ=00, 0:BOOTRST=1 

; Port C is connected to an VGA HD15 adapter, 3*2bit DACs and 2 TTLs
;   PC0+PC1 blue DAC, PC2+PC3 green DAC, PC4+PC5 red DAC
;   PC6 hsync TTL, PC7 vsync TTL
; Port D is connected to an mode switch
;   PD6 (next to GND on 10pin header), 0(jumper) = double, 1(default) = blank


; we are using an ATmega32
  .include "m32def.inc"


; ------ handle reset and interrupts (all identical as not used)

; --- set up vectors

  .cseg
  .org $0000

  jmp RESET
  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR

  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR

  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR

  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR
  jmp INTR

  jmp INTR


; --- handle interrupts, despite not used

INTR:

  ; just return, nothing else makes sense here
  reti


; ------ auxillary routines

; --- wait a while doing nothing, spinloop

; parameters: TL loop count, 1..256
;             gets destroyed (reduced to 0), avoid push/pop time, seldom reuse

; timing (TL-1)*3+2+4 = TL*3-1+4 = TL*3+3 clocks, for TL=[1..255,0(=256)]
;   min 1*3-1+4=11 clocks for TL=1, max 256*3-1+4=776 clocks for TL=0(=256)

WAIT:

                         ; {n} = clocks used, for timing computations
                         ;   this program requires exact clock counting
                         ;   single-clock miscounts produce visual artifacts

  ; TL set up by caller  ; for count = TL..1
  WAILOOP:

    ; nothing in loop, for highest loop resolution, only 3 clocks

    dec TL               ; {1}  next count
    brne WAILOOP         ; {2|1}

  ret                    ; {4}


; ------ string handling stuff


; --- define an string buffer, "string accumulator", no dynamic alloc yet

  .equ STMAXLEN = 80

  .dseg
STBUFLEN:
  .byte 2 ; buffer length, self documenting, must be set to STMAXLEN while init
STCONTLEN:
  .byte 2 ; content length, actually used amount of buffer
STDATA:
  .byte STMAXLEN ; data comes into here
  .cseg


; --- initialise string subsystem

STINIT:
  ldi TL,low(STMAXLEN)  ; mark string buffer with length, never changed
  ldi TH,high(STMAXLEN)
  sts STBUFLEN+0,TL
  sts STBUFLEN+1,TH
  ldi TL,0              ; mark string buffer as empty, changed with data
  ldi TH,0
  sts STCONTLEN+0,TL
  sts STCONTLEN+1,TH
  ret


; --- load an string constant into string buffer

; parameters: TOS (old PC) pointer to string constant (in Flash!)
;             format of constant: .dw length, .db data
;             these is assembled in, directly after the  call STLDI

STLDI:
  pop ZH                ; next calling program address, begin of constant
  pop ZL                ;   call/push low first, ret/pop high first
  lsl ZL                ;   and correct for word vs byte addressing
  rol ZH

  lpm TL,Z+             ; length of string constant, from program
  lpm TH,Z+
  lds AL,STBUFLEN+0     ; length of buffer, for fitting
  lds AH,STBUFLEN+1
  cp TH,AH              ; high(constlen) < high(buflen)?
    brlo STFITS         ;   <, is OK, AH:AL is length of new content
    breq STMAYFIT       ;   =, test also the low byte
    movw TH:TL,AH:AL    ;   >, too long, shorten to constlen = buflen
    rjmp STFITS
  STMAYFIT:
    cp TL,AL            ; low(constlen) < low(buflen)?
      brlo STFITS       ;   <, is OK
      breq STFITS       ;   =, also OK, just fits
      movw TH:TL,AH:AL  ;   >, too long, shorten to constlen = buflen
  STFITS:

  sts STCONTLEN+0,TL    ; TH:TL record as new length of content of buffer
  sts STCONTLEN+1,TH

  ldi YL,low(STDATA)    ; stuff will be written to the buffer
  ldi YH,high(STDATA)

  ; TH:TL set above     ; for char = STCONTLEN..1
  STLLOOP:

    lpm AL,Z+           ; copy string data, from program
    st Y+,AL

    subi TL,1           ; next char
    sbci TH,0
    brne STLLOOP

  lsr ZH                ; new calling program address, continuation of code
  ror ZL                ;   undo correction
  push ZL               ;   undo pop
  push ZH
  ret


; ------ vertical retrace stuff

; --- set up registers for generating sync pulses

; constants for sync pulse patterns
  .equ NOSYNC = 0x00  ; no sync
  .equ HSYNC  = 0x40  ; only H sync
  .equ VSYNC  = 0x80  ; only V sync
  .equ HVSYNC = 0xC0  ; both syncs
  .equ DDRVGA = 0xFF  ; DDR all 1 (= output)

; registers for sync pulse patterns, needed for  out PORTC,value-reg
  ; use R2..R15 non-immediate range, as one load and permanent use
  .def NONE = R15
  .def HOR  = R14
  .def VER  = R13
  .def H_V  = R12

RETRINIT:

  ; set registers for sync patterns
  ldi TL,NOSYNC
  mov NONE,TL

  ldi TL,HSYNC
  mov HOR,TL

  ldi TL,VSYNC
  mov VER,TL

  ldi TL,HVSYNC
  mov H_V,TL

  ; get PortC data register ready
  out PORTC,NONE

  ; get PortC DDR register ready
  ldi TL,DDRVGA
  out DDRC,TL

  ret


; --- draw vertical retrace top or bottom blank line (or any other blank line)

; timing 480 + 30 + 30 + 5 = 600-55spare

BLLINE:

  ; {480}  "draw" blank pixels, V = 0, H = 0, color = blank
  out PORTC,NONE  ; {1}
  ldi TL,157      ; {1}
  call WAIT       ; {4+(157*3+3)=478}

  ; {30}  horizontal retrace left blank, V = 0, H = 0, color = blank
  out PORTC,NONE  ; {1}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}

  ; {30}  horizontal retrace pulse, V = 0, H = 1, color = blank
  out PORTC,HOR   ; {1}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}

  ; {5of60} horizontal retrace right blank, V = 0, H = 0, color = blank
  out PORTC,NONE  ; {1}
  ; no wait, give rest of time to caller

  ret             ; {4}


; --- draw vertical retrace pulse line

; timing 480 + 30 + 30 + 5 = 600-55spare

VPLINE:

  ; {480}  "draw" blank pixels, V = 1, H = 0, color = blank
  out PORTC,VER   ; {1}
  ldi TL,157      ; {1}
  call WAIT       ; {4+(157*3+3)=478}

  ; {30}  horizontal retrace left blank, V = 1, H = 0, color = blank
  out PORTC,VER   ; {1}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}

  ; {30}  horizontal retrace pulse, V = 1, H = 1, color = blank
  out PORTC,H_V   ; {1}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}

  ; {5of60}  horizontal retrace right blank, V = 1, H = 0, color = blank
  out PORTC,VER   ; {1}
  ; no wait, give rest of time to caller

  ret             ; {4}


; ------ frame buffer stuff

; --- define frame buffer

  .equ FBCOLOURS = 5  ; color bytes a begin of row
  .equ FBCOLUMNS = 40 ; columns of character code bytes in row
  .equ FBROWEND  = 1  ; abort pseudo character at end of row
                      ;   use ASCII DEL, aa=0x7F, hh=0x3F, ll=0x80, cc=0xBF
  .equ FBROWLEN  = FBCOLOURS+FBCOLUMNS+FBROWEND ; legth of row in frame buffer
  .equ FBROWS    = 25 ; rows of characters in one frame
  .equ FBEND     = 1  ; no-more-lines pseudo colour byte at end of frame
                      ;   use invalid (not 0x00..0x1F) code 0xFF
  .equ FBSIZE    = FBROWS*FBROWLEN+FBEND ; space used for frame buffer
  ; uses FBROWS*FBROWLEN = 25*(5+40+1)+1=1151bytes (of ATmega32 2048)

  .dseg
FBUF:
  .byte FBSIZE
  .cseg


; --- write colours into frame buffer

; parameters: XH:XL pointer to where in frame buffer to place char
;             AL colour code

FBWCOLOUR:

  push AL

  andi AL,0x3F   ; prevent false sync signals from broken colour values
  st X+,AL       ; place this colour

  pop AL
  ret


; --- write character into frame buffer

; parameters: XH:XL pointer to where in frame buffer to place char
;             AL ASCII character

FBWCHAR:

  push AL

  cpi AL,0x08         ; char < IVON ? lowest defined, specials pseudo-char
    brsh FBBOTTOMOK   ;   >=, is ok
    ldi AL,0x2E       ;   <, replace with ASCII "."
  FBBOTTOMOK:
  
  cpi AL,0x7F         ; char < DEL ? lowest not defined, abort pseudo-char
    brlo FBTOPOK      ;   <, is ok
    ldi AL,0x2E       ;   >=, replace with ASCII "."
  FBTOPOK:

  bst AL,0            ; convert ASCII code to char code, 8bit rotate right
  lsr AL              ;   .654 3210 -> ..65 4321 (0)
  bld AL,7            ;             -> 0.65 4321

  ; XH:XL set by caller
  st X+,AL            ; place this char

  pop AL
  ret


; --- write string buffer into frame buffer

; parameters: XH:XL pointer to where in frame buffer to place string

FBWSTR:

  push YL
  push YH
  push AL

  ldi YL,low(STDATA)    ; from the string buffer
  ldi YH,high(STDATA)

  ; XH:XL set by caller, contains address in frame buffer

  lds TL,STCONTLEN+0    ; for char = STCONTLEN..1
  lds TH,STCONTLEN+1
  FBWLOOP:

    ld AL,X             ; check for line wrap
    cpi AL,0xBF         ;   end of row abort pseudo-char
      brne FBWCOPY
      ld AL,X+          ; correct address, skip row end abort pseudo-char
      ld AL,X           ; check for page wrap
      cpi AL,0xFF       ;   end of frame pseudo-colour
        brne FBWCOL
        ldi XL,low(FBUF)  ; correct address, wrap to begin of frame buffer
        ldi XH,high(FBUF)
      FBWCOL:
      ld AL,X+          ; correct address, skip row begin colour codes
      ld AL,X+
    FBWCOPY:

    ld AL,Y+            ; write string data
    call FBWCHAR        ;   uses  st X+,AL

    subi TL,1           ; next char
    sbci TH,0
    brne FBWLOOP

  pop AL
  pop YH
  pop YL
  ret


; --- initialise frame buffer

FBINIT:

  push AL

                      ; get port D6 (double/2ndblank mode select) ready
  sbi PORTD,PIND6     ; pullup 1 (= on)
  cbi DDRD,PIND6      ; DDR 0 (= input)

  ldi XL,low(FBUF)    ; begin of frame buffer
  ldi XH,high(FBUF)

  ldi TH,FBROWS       ; for row-in-frame = FBROWS..1
  FBRLOOP:

    ; default background colour for this row
    ldi AL,0x03       ; blue
    call FBWCOLOUR    ; uses  st X+,AL

    ; default foreground colour for this row
    ;ldi AL,0x0F       ; cyan
    ;ldi AL,0x1F       ; light blue, slight reddish cyan
    ldi AL,0x1B       ; light blue, sligh reddisch and lower green cyan, best
    ;ldi AL,0x2B       ; light blue, low red+green white
    call FBWCOLOUR    ; uses  st X+,AL

    ; default alternate colours for this row, progressively "stronger"
    ldi AL,0x0F       ; cyan
    call FBWCOLOUR    ; uses  st X+,AL
    ldi AL,0x3F       ; white
    call FBWCOLOUR    ; uses  st X+,AL
    ldi AL,0x3C       ; yellow
    call FBWCOLOUR    ; uses  st X+,AL

    ; blank out the frame buffer
    ldi TL,FBCOLUMNS  ; for column-in-row = FBCOLUMNS..1
    FBCLOOP:

      ldi AL,0x2E     ; fill unused space with ASCII "."
      call FBWCHAR    ; uses  st X+,AL

      dec TL          ; next column-in-row
      brne FBCLOOP

    ldi AL,0xBF       ; end of row abort pseudo-char
    st X+,AL          ;   can not use FBWCHAR because framing

    dec TH            ; next row-in-frame
    brne FBRLOOP

  ldi AL,0xFF         ; end of frame no-more-lines pseudo-colour
  st X+,AL            ;   can not use FBWCHAR or FBWCOLOUR because framing

  pop AL
  ret


; ------ X/Y coordinate based frame buffer stuff

; registers for row/coloumn loop counting and parameter passing
  .def XYCOL = R16
  .def XYROW  = R17


; --- X/Y write colours into frame buffer

; parameters: XYROW where in frame buffer to place colour
;             AL colour code
;             AH colour to set, 0..4 BACK/FORE/ALT1/ALT2/ALT3

XYWCOLOUR:

  cpi XYROW,FBROWS      ; row in 0..(FBROWS-1) ?
    brsh XYOUTRANGE     ;   >=, dont' colour non-existant row
  cpi AH,FBCOLOURS
    brsh XYOUTRANGE     ;   >=, dont' set to non-existant colour

  ldi XL,low(FBUF)      ; begin of frame buffer colours
  ldi XH,high(FBUF)

  ldi TL,FBROWLEN       ; + FBROWLEN * ROW
  mul XYROW,TL
  add XL,ML
  adc XH,MH

  add XL,AH             ; + index for proper colour byte
  ldi AH,0
  adc XH,AH

  jmp FBWCOLOUR

XYOUTRANGE:
  ret


; --- X/Y write character into frame buffer

; parameters: XYROW and XYCOL where in frame buffer to place string
;             AL ASCII character

XYWCHAR:

  cpi XYROW,FBROWS      ; row in 0..(FBROWS-1) ?
  brsh XYOUTRANGE       ;   >=, dont' draw in non-existant row

  cpi XYCOL,FBCOLUMNS   ; column in 0..(FBCOLUMNS-1) ?
  brsh XYOUTRANGE       ;   >=, dont' draw in non-existant column

  ldi XL,low(FBUF+FBCOLOURS) ; begin of frame buffer chars
  ldi XH,high(FBUF+FBCOLOURS)

  ldi TL,FBROWLEN       ; + FBROWLEN * ROW
  mul XYROW,TL
  add XL,ML
  adc XH,MH

  add XL,XYCOL          ; + COL
  ldi TH,0
  adc XH,TH

  jmp FBWCHAR


; --- X/Y write string buffer into frame buffer

; parameters: XYROW and XYCOL where in frame buffer to place string

XYWSTR:

  cpi XYROW,FBROWS      ; row in 0..(FBROWS-1) ?
  brsh XYOUTRANGE       ;   >=, dont' draw in non-existant row

  cpi XYCOL,FBCOLUMNS   ; column in 0..(FBCOLUMNS-1) ?
  brsh XYOUTRANGE       ;   >=, dont' draw in non-existant column

  ldi XL,low(FBUF+FBCOLOURS) ; begin of frame buffer chars
  ldi XH,high(FBUF+FBCOLOURS)

  ldi TL,FBROWLEN       ; + FBROWLEN * ROW
  mul XYROW,TL
  add XL,ML
  adc XH,MH

  add XL,XYCOL          ; + COL
  ldi TH,0
  adc XH,TH

  jmp FBWSTR


; ------ visible output drawing stuff

; colour registers for drawing, are the LUT
  .def BACK = R11 ; use R2..R15 non-immediate range, as no immediate load
  .def FORE = R10
  .def ALT1 = R09 ; 3 alternate foreground colours per row
  .def ALT2 = R08
  .def ALT3 = R07
  .def TEMP = R06 ; temporary for FORE-TEMP-<target> rotate/swap/exchange


; --- draw character based visible output line

; parameters: XH:XL pointer to first char in row
;             TH font line segment offset, 0x00, 0x10, .. 0x70

; timing 5*2+8=18pre, 480 + 30 + 30 + 5 = 600-55spare

CHLINE:

  ; {*pre} "predraw" time, before drawing begins, precompensate this in caller

  ; {10pre} load LUT colour registers for this row
  ld BACK,X+      ; {2}  first 5 bytes in row in frame buffer are colours
  ld FORE,X+      ; {2}
  ld ALT1,X+      ; {2}
  ld ALT2,X+      ; {2}
  ld ALT3,X+      ; {2}

  ; {8pre} get first character code and unthread it
  ;   use standard font unthreading technique, see there for generic case
  ;     but compressed, without 4 clocks of intermixed pixel OUT statements
  ;     and adding variable segment offset, for entering proper segment group
  ; XH:XL address has already been set up by caller
  ld ZL,X+        ; {2}  next (=first) character: get code
  mov ZH,ZL       ; {1}  next (=first) character: extract top part
  andi ZH,0x3F    ; {1}    for ATmega32: 0x3F limits address to 64(*2) chars
  andi ZL,0x80    ; {1}  next (=first) character: extract bottom part
  ; TH pattern has already been set up by caller, because variable
  or ZL,TH        ; {1}    add in segment offset for this line in row
  ijmp            ; {2}  next (=first) character: go and draw it
                  ;        continuation at that chars pixel 1, including above


  ; {(40*)12}  draw stuff, 40 char segments, V = 0, H = 0, color = pixel

  ; for each X+ fetched code: index and exec one line segment in font space
  ;   drawing done by  out PORTC,colour, one for each pixel, one every 3 clocks
  ;   {so per segment time 4 pixels a 3 clocks = 12 clocks/line}

  ; per character 8 such segments, one for each line in row
  ;   so per segment size 16 instructions * 2 bytes = 128instr/char
  ;     on ATmega32 32k Flash this allows 32k/2/128=128chars, minus used space
  ;   note that with max 6 drawable lines this could reduce to 6*10(+4==64instr
  ;     on ATmega32 this would allow 32k/2/64/=256chars, minus used space


  ; in font, once per character:
  ; .org $hhll    ; ASCII ".", aa=0xaa, hh=0xhh, ll=0xll, cc=0xcc
                  ;   because 8*16=128instr blocks use up lowest 7 address bits
                  ;     and we need fast 8bit indexing in unthreading code
                  ;   index as (aa/2)*2*128, gives char codes in font
                  ;     interleaved as  00 80 01 81 02 82 ... 3F BF
                  ;   compensate this by 8bit rotate right ASCII to char code
                  ;     aa=.654 3210, hh=..65 4321, ll=0... ...., cc=0.65 4321
                  ;   hh=0x00..3F for aa=0x00,02..7E and repeats aa=0x01,03..7F
                  ;   ll=0x00 for aa=0x00..3F and 0x80 for aa=0x8x..BF
  ;   no labels after .org, as font code is only ever reached by computed ijmp

  ; {12}  in font, 8 times per character, each line segment:
  ;         4 clocks write pixels with  out PORTC,xxxx (xxxx = FORE or BACK)
  ;         between these 4*2 clocks for unthreading next character jump

  ; out PORTC,xxxx ; {1}  current character: draw pixel 1
  ; ld ZL,X+      ; {2}  next character: get character code

  ; out PORTC,xxxx ; {1}  current character: draw pixel 2
  ; mov ZH,ZL     ; {1}  next character: copy and extract top part of index
  ; andi ZH,0x3F  ; {1}    for ATmega32: 0x3F limits address to 64(*2) chars

  ; out PORTC,xxxx ; {1}  current character: draw pixel 3
  ; andi ZL,0x80  ; {1}  next character: extract bottom part for (64*)2 chars
  ; ori ZL,0xo0   ; {1}    add in line segment offset for this line in row
                  ;          o = 0x00 0x10 .. 0x70 or o = 0x80 0x90 .. 0xF0

  ; out PORTC,xxxx ; {1}  current character: draw pixel 4
  ; ijmp          ; {2}  next character: go and execute it
                  ;        continuation at that chars pixel 1, as above

  ; nop           ; {0}  fill rest of unused space, 16-10=6 instruction words
  ; nop           ;        no execution time as never run!
  ; nop
  ; nop
  ; nop
  ; nop

  ; above continues until an  abort pseudo-char  code is hit in frame buffer
  ;   that pseudo-char exits the unthreadinging instead of drawing
  ;   must be done this way, because no time for 40 char loop counting


  ; {30}  horizontal retrace left blank, V = 0, H = 0, color = blank
  ;   pseudo-char directly runs time critical   out PORTC,NONE  then jumps back

  ; .org $hhll    ; ASCII ".", aa=0xaa, hh=0xhh, ll=0xll, cc=0xcc
  ;   no labels after .org, as font code is only ever reached by computed ijmp

  ; this code is in abort pseudo-character
  ; out PORTC,NONE ; {1}  switch off whatever was last background colour
  ; jmp CHEND     ; {3}  go back to CHLINE to make horizontal retrace pulse
  ; nop           ; {0}  fill rest of unused space, 16-1-2=13 instruction words
  ; nop           ;        no execution time as never run!
  ; nop           ;        followd by annother 10 nop instructions!
  ; like for every other char this is repeated for 8 line pseudo-segments

CHEND:

  ldi TL,6        ; {1}
  call WAIT       ; {4+(6*3+3)=25}

  ; {30}  horizontal retrace pulse, V = 0, H = 1, color = blank
  out PORTC,HOR   ; {1}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}

  ; {5of60}  horizontal retrace right blank, V = 0, H = 0, color = blank
  out PORTC,NONE  ; {1}
  ; no wait, give rest of time to caller

  ret             ; {4}


; --- draw pair of character based visible output lines
;   belonging to one font line, doublescan or 2nd line blank

; parameters: XH:XL pointer to first char in row
;             TH font line segment offset, 0x00, 0x10, .. 0x70

; timing 4+18=22pre, 600 + 480 + 30 + 30 + 5 = 1200-55spare
;   no call+ret at end, as jmp does same and is faster {3 instead of 4+4}

CHPAIR:

  ; first line of linepair
  call CHLINE     ; {4+(18pre+600-55spare)}
  ldi TL,7        ; {1}
  call WAIT       ; {4+(7*3+3)=28}
  nop             ; {2}
  nop

  ; second line of linepair, same segment, same TH
  sbiw XH:XL,FBROWLEN ; {2}  return to frame buffer row begin, undo all X+

  ; switch for doublescan or 2nd blank
  sbis PIND,PIND6 ; {1|3}  1(default) = 2nd blank, 0(jumpered) = doublescan

  jmp CHLINE      ; {3+(18pre+600-55spare)}  doublescan: draw chars second time

  ; {16}  correction, as BLLINE has no 18pre, unlike CHLINE
  ;         {16} + {3-1=2} from the sbis = {18}
  ldi TL,2        ; {1}
  call WAIT       ; {4+(2*3+3)=13}
  nop             ; {2}
  nop
  jmp BLLINE      ; {3+(600-55spare)}  2nd blank: draw an blank
  ; last line, no wait, give rest of time to caller


; ------ frame timing control loops stuff

; registers for loop counting
  .def FRLINE = R18
  .def FRROW  = R19


; --- draw all output lines of one row

; parameters: YH:YL pointer to first char in row

; timing 1+1+1+4+22=29pre, (7*2+1)*600 + (600-55=)545 = 16*600-55spare
;   no call+ret at end, as jmp does same and is faster {3 instead of 4+4}

ROWDRAW:

  ldi TH,0x00          ; {1}  start row with first line segment offset

  ; loop through 7(of8) linepairs, 8th linepair without wait after it
  ldi FRLINE,7         ; {1}  for line-in-row = 7..1
  ROWLOOP:

    movw XH:XL,YH:YL   ; {1}  XH:XL for this line, YH:YL survive for next

    call CHPAIR        ; {4+(22pre+600-55spare)}
    ldi TL,5           ; {1}
    call WAIT          ; {4+(5*3+3)=22}

    ; segment switching for next linepair
    subi TH,-0x10      ; {1}  move line segment offset by 16instr/32bytes
    andi TH, 0x70      ; {1}  keep it in range, also after exiting loop

    dec FRLINE         ; {1}  next line-in-row
    brne ROWLOOP       ; {2|1}
                       ; {1+4+22 precompensate movw + call CHLINE + predraw}

  nop                  ; {1, compensates brne 2|1, as no ldi}

  ; 8th linepair, last one in row, no wait, give rest of time to caller
  movw XH:XL,YH:YL     ; {1}
  nop                  ; {1, precompensate jmp 3 instead of call 4}
  jmp CHPAIR           ; {3+(22pre+600-55spare)}


; --- draw all output rows of one frame

; timing 2+2+4+29=37pre, 24*8*600 + 7*600 + 545 = 400*600-55spare
;   no call+ret at end, as jmp does same and is faster {3 instead of 4+4}

FRDRAW:

  ldi YL,low(FBUF)     ; {1}  get frame buffer first row begin
  ldi YH,high(FBUF)    ; {1}

  ; loop through 24(of25) rows, 25th row without wait after it

  ldi FRROW,FBROWS     ; {1}  for row-in-frame = (FBROWS-1)..1
  dec FRROW            ; {1}
  FRDLOOP:

    call ROWDRAW       ; {4+(29pre+16*600-55spare)}
    ldi TL,3           ; {1}
    call WAIT          ; {4+(3*3+3)=16}

    adiw YH:YL,FBROWLEN ; {2}  go to frame buffer next row begin

    dec FRROW          ; {1}  next row-in-frame
    brne FRDLOOP       ; {2|1}
                       ; {4+29 precompensate call ROWDRAW + predraw}

  nop                  ; {1, compensates brne 2|1, as no ldi}

  ; 25th row, last one in frame, no wait, give rest of time to caller
  nop                  ; {1, precompensate jmp 3 instead of call 4}
  jmp ROWDRAW          ; {3+(29pre+16*600-55spare)}


; --- draw all vertical retrace bottom blank lines of one frame

; timing 1+4=5pre, 11*600 + (600-55=)545 = 12*600-55spare
;   no call+ret at end, as jmp does same and is faster {3 instead of 4+4}

FRBOTTOM:

  ; loop through 11(of12) lines, 12th line without wait after it
  ldi FRLINE,11        ; {1}  for line-in-bottom = (12-1)..1
  FRBLOOP:

    call BLLINE        ; {4+(600-55spare)}
    ldi TL,13          ; {1}
    call WAIT          ; {4+(13*3+3)=46}
    nop                ; {1}

    dec FRLINE         ; {1}  next line-in-bottom
    brne FRBLOOP       ; {2|1}
                       ; {4 precompensate call BLLINE}

  nop                  ; {1, compensates brne 2|1, as no ldi}

  ; 12th line, last one in bottom, no wait, give rest of time to caller
  nop                  ; {1, precompensate jmp 3 instead of call 4}
  jmp BLLINE           ; {3+(600-55spare)}


; --- draw all vertical retrace pulse lines of one frame

; timing 1+4=5pre, 12*600 + (600-55=)545 = 13*600-55spare
;   no call+ret at end, as jmp does same and is faster {3 instead of 4+4}

FRVPULSE:

  ; loop through 12(of13) lines, 13th line without wait after it
  ldi FRLINE,12        ; {1}  for line-in-vpulse = (13-1)..1
  FRVLOOP:

    call VPLINE        ; {4+(600-55spare)}
    ldi TL,13          ; {1}
    call WAIT          ; {4+(13*3+3)=46}
    nop                ; {1}

    dec FRLINE         ; {1}  next line-in-vpulse
    brne FRVLOOP       ; {2|1}
                       ; {4 precompensate call VPLINE}

  nop                  ; {1, compensates brne 2|1, as no ldi}

  ; 13th line, last one in vpulse, no wait, give rest of time to caller
  nop                  ; {1, precompensate jmp 3 instead of call 4}
  jmp VPLINE           ; {3+(600-55spare)}


; --- draw all vertical retrace top blank lines of one frame

; timing 1+4=5pre, 24*600 + (600-55=)545 = 25*600-55spare
;   no call+ret at end, as jmp does same and is faster {3 instead of 4+4}

FRTOP:

  ; loop through 24(of25) lines, 25th line without wait after it
  ldi FRLINE,24        ; {1}  for line-in-top = (25-1)..1
  FRTLOOP:

    call BLLINE        ; {4+(600-55spare)}
    ldi TL,13          ; {1}
    call WAIT          ; {4+(13*3+3)=46}
    nop                ; {1}

    dec FRLINE         ; {1}  next line-in-bottom
    brne FRTLOOP       ; {2|1}
                       ; {4 precompensate call BLLINE}

  nop                  ; {1, compensates brne 2|1, as no ldi}

  ; 25th line, last one in top, no wait, give rest of time to caller
  nop                  ; {1, precompensate jmp 3 instead of call 4}
  jmp BLLINE           ; {3+(600-55spare)}


; --- draw all sections of one frame

; timing 4+37=41pre, (25*8+12+13+24)*600 + 545 = 450*600-55spare

FRAME:

  call FRDRAW     ; {4+(37pre+400*600-55spare)}
  ldi TL,12       ; {1}
  call WAIT       ; {4+(12*3+3)=43}
  nop             ; {2}
  nop

  call FRBOTTOM   ; {4+(5pre+12*600-55spare)}
  ldi TL,12       ; {1}
  call WAIT       ; {4+(12*3+3)=43}
  nop             ; {2}
  nop

  call FRVPULSE   ; {4+(5pre+13*600-55spare)}
  ldi TL,12       ; {1}
  call WAIT       ; {4+(12*3+3)=43}
  nop             ; {2}
  nop

  ; last section, no wait, give rest of time to caller
  nop             ; {1, precompensate jmp 3 instead of call 4}
  jmp FRTOP       ; {3+(5pre+25*600-55spare)}


; --- main display loop, infinite loop, endlessly draw frames

DISPLAY:

  call FRAME      ; {4+41pre+(450*600-55spare)}
  nop             ; {8}
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  rjmp DISPLAY    ; {2}
                  ; {4+41 precompensate call FRAME + predraw}


; ------ main program

; --- blank line, to separate sections

DEBLANK:

  ldi AL,0x00         ; black
  ldi AH,0            ; bg
  call XYWCOLOUR

  call STLDI          ; empty string
  .dw 40
  .db "                                        "
  call XYWSTR

  inc XYROW

  ret


; --- show something in frame buffer

DEMO:

  ldi XYROW,0         ; start at top left of screen
  ldi XYCOL,0


  ; show all 95 ASCII chars, 3+1 rows

  call STLDI
  .dw 40
  .db "     !", 0x22, "#$%&", 0x27, "()*+,-./0123456789:;<=>?    "
  call XYWSTR
  inc XYROW

  call STLDI
  .dw 40
  .db "    @ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_    "
  call XYWSTR
  inc XYROW

  call STLDI
  .dw 40
  .db "    `abcdefghijklmnopqrstuvwxyz{|}~     "
  call XYWSTR
  inc XYROW


  call DEBLANK


  ; show rainbow backgrounds, 6+1 rows

  call STLDI          ; common text pattern
  .dw 40
  .db "----._.-'_`----< { [] } >----._.-'_`----"

  ldi AL,0x30         ; red
  ldi AH,0            ; bg
  call XYWCOLOUR
  ldi AL,0x00         ; black
  ldi AH,1            ; fg
  call XYWCOLOUR
  call XYWSTR
  inc XYROW

  ldi AL,0x3C         ; yellow
  ldi AH,0            ; bg
  call XYWCOLOUR
  ldi AL,0x00         ; black
  ldi AH,1            ; fg
  call XYWCOLOUR
  call XYWSTR
  inc XYROW

  ldi AL,0x0C         ; green
  ldi AH,0            ; bg
  call XYWCOLOUR
  ldi AL,0x00         ; black
  ldi AH,1            ; fg
  call XYWCOLOUR
  call XYWSTR
  inc XYROW

  ldi AL,0x0F         ; cyan
  ldi AH,0            ; bg
  call XYWCOLOUR
  ldi AL,0x00         ; black
  ldi AH,1            ; fg
  call XYWCOLOUR
  call XYWSTR
  inc XYROW

  ldi AL,0x03         ; blue
  ldi AH,0            ; bg
  call XYWCOLOUR
  ldi AL,0x00         ; black
  ldi AH,1            ; fg
  call XYWCOLOUR
  call XYWSTR
  inc XYROW

  ldi AL,0x33         ; magenta
  ldi AH,0            ; bg
  call XYWCOLOUR
  ldi AL,0x00         ; fg black
  ldi AH,1            ; fg
  call XYWCOLOUR
  call XYWSTR
  inc XYROW


  call DEBLANK


  ; announce text, middle 3 rows, of the 11+3+11

  call STLDI
  .dw 40
  .db "AVR ATmega32 uC + SoftVGA = 400line 70Hz"
  call XYWSTR
  inc XYROW

  call STLDI
  .dw 40
  .db "18.342MHz, 3clk/pixl, 4pixl/char, 40char"
  call XYWSTR
  inc XYROW

  call STLDI
  .dw 40
  .db "full ASCII, 4x8pixel font, 2 lines/pixel"
  call XYWSTR
  inc XYROW


  ; show RGB foregrounds, 1+7 rows

  call DEBLANK


  call STLDI          ; common text pattern
  .dw 40
  .db "nor", IVON, "inv", IVOFF
  .db "nor", ULON, "ulin"
  .db ULOFF, "nor", COL1, "2/3"
  .db COL1, "nor", COL2, "1/3"
  .db COL2, "nor", COL3, "pas"

  ldi AL,0x00         ; black
  ldi AH,0            ; bg
  call XYWCOLOUR
  ldi AL,0x03         ; blue
  ldi AH,1            ; fg
  call XYWCOLOUR
  ldi AL,0x02         ; blue
  ldi AH,2            ; co1
  call XYWCOLOUR
  ldi AL,0x01         ; blue
  ldi AH,3            ; co2
  call XYWCOLOUR
  ldi AL,0x2B         ; blue
  ldi AH,4            ; co3
  call XYWCOLOUR
  call XYWSTR
  inc XYROW

  ldi AL,0x00         ; black
  ldi AH,0            ; bg
  call XYWCOLOUR
  ldi AL,0x0C         ; green
  ldi AH,1            ; fg
  call XYWCOLOUR
  ldi AL,0x08         ; green
  ldi AH,2            ; co1
  call XYWCOLOUR
  ldi AL,0x04         ; green
  ldi AH,3            ; co2
  call XYWCOLOUR
  ldi AL,0x2E         ; green
  ldi AH,4            ; co3
  call XYWCOLOUR
  call XYWSTR
  inc XYROW

  ldi AL,0x00         ; black
  ldi AH,0            ; bg
  call XYWCOLOUR
  ldi AL,0x0F         ; cyan
  ldi AH,1            ; fg
  call XYWCOLOUR
  ldi AL,0x0A         ; cyan
  ldi AH,2            ; co1
  call XYWCOLOUR
  ldi AL,0x05         ; cyan
  ldi AH,3            ; co2
  call XYWCOLOUR
  ldi AL,0x2F         ; cyan
  ldi AH,4            ; co3
  call XYWCOLOUR
  call XYWSTR
  inc XYROW

  ldi AL,0x00         ; black
  ldi AH,0            ; bg
  call XYWCOLOUR
  ldi AL,0x30         ; red
  ldi AH,1            ; fg
  call XYWCOLOUR
  ldi AL,0x20         ; red
  ldi AH,2            ; co1
  call XYWCOLOUR
  ldi AL,0x10         ; red
  ldi AH,3            ; co2
  call XYWCOLOUR
  ldi AL,0x3A         ; red
  ldi AH,4            ; co3
  call XYWCOLOUR
  call XYWSTR
  inc XYROW

  ldi AL,0x00         ; black
  ldi AH,0            ; bg
  call XYWCOLOUR
  ldi AL,0x33         ; magenta
  ldi AH,1            ; fg
  call XYWCOLOUR
  ldi AL,0x22         ; magenta
  ldi AH,2            ; co3
  call XYWCOLOUR
  ldi AL,0x11         ; magenta
  ldi AH,3            ; co2
  call XYWCOLOUR
  ldi AL,0x3B         ; magenta
  ldi AH,4            ; co3
  call XYWCOLOUR
  call XYWSTR
  inc XYROW

  ldi AL,0x00         ; black
  ldi AH,0            ; bg
  call XYWCOLOUR
  ldi AL,0x3C         ; yellow
  ldi AH,1            ; fg
  call XYWCOLOUR
  ldi AL,0x28         ; yellow
  ldi AH,2            ; co1
  call XYWCOLOUR
  ldi AL,0x14         ; yellow
  ldi AH,3            ; co2
  call XYWCOLOUR
  ldi AL,0x3E         ; yellow
  ldi AH,4            ; co3
  call XYWCOLOUR
  call XYWSTR
  inc XYROW

  ldi AL,0x00         ; black
  ldi AH,0            ; bg
  call XYWCOLOUR
  ldi AL,0x3F         ; white
  ldi AH,1            ; fg
  call XYWCOLOUR
  ldi AL,0x2A         ; white
  ldi AH,2            ; co1
  call XYWCOLOUR
  ldi AL,0x15         ; white
  ldi AH,3            ; co2
  call XYWCOLOUR
  ldi AL,0x3F         ; white
  ldi AH,4            ; co3
  call XYWCOLOUR
  call XYWSTR
  inc XYROW


  ; show website info, 1+2 rows

  call DEBLANK


  call STLDI
  .dw 40
  .db "design and code open source available at"
  call XYWSTR
  inc XYROW

  call STLDI
  .dw 40
  .db "http://neil.franklin.ch/Projects/SoftVGA"
  call XYWSTR
  inc XYROW


  ret


; --- handle reset, initialise system, enter main loop

RESET:

  ; get stack ready, for calls and interrupts (later not used)
  ldi TL,low(MAXRAM)
  ldi TH,high(MAXRAM)
  out SPL,TL
  out SPH,TH

  ; get interrupts ready, despite not used
  sei

  ; set up string handling
  call STINIT

  ; set up retrace sync pulse register patterns, and get port C ready
  call RETRINIT

  ; set up frame buffer, and get port D6 ready
  call FBINIT

  ; fill frame buffer
  call DEMO

  ; enter endless loop, no return back to here
  rjmp DISPLAY


; ------ fixed calculated address stuff, for font, executed by ijmp, no labels

; overjump program code space, which results in a few unusable character codes
;   expected to contain 2*(64-4)-1 chars, 0x04 0x48 0x05 0x59 .. 0x3F, no 0xBF
;     enough for 95 printable ASCII chars aa=0x20..0x7E (no non-draw DEL/0x7F)
;     and 16 more non-ASCII drawable chars aa=0x10..1F before them
;     and  8 more non-ASCII pseudo   chars aa=0x08..0F before them
;   this allows for 4*2*128=1k instr of code, safely more than is needed


; --- specials pseudo-characters "drawing" routines

  ; with pre-ASCII specials pseudo-chars from 0x08-0x0F
  .org $0400     ; use ASCII non-draw 08, aa=0x08, hh=0x04, ll=0x00, cc=0x04

  ; include specials as separate file, so they may be generated automatically
  .include "vga_text_specials.inc"


; --- drawable characters drawing routines

  ; with pre-ASCII drawable chars 0x10 to 0x1F and ASCII chars 0x20 to 0x7E
  .org $0800     ; pre-ASCII, aa=0x10, hh=0x08, ll=0x00, cc=0x08

  ; include font as separate file, so they may be generated automatically
  .include "vga_text_font.inc"


; --- abort pseudo-characters "drawing" routines

  .org $3F80     ; use ASCII non-draw DEL, aa=0x7F, hh=0x3F, ll=0x80, cc=0xBF

                 ; 8 line segments repeated identically

  out PORTC,NONE ; {1}  switch off whatever was last background colour
  jmp CHEND      ; {3}  go back to CHLINE to make horizontal retrace pulse
  nop            ; {0}  fill rest of unused space, 16-1-2=13 instruction words
  nop            ;        no execution time as never run!
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  out PORTC,NONE
  jmp CHEND
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
