; vga_threaded.asm - threaded generate VGA text, background for frame updates
; author Neil Franklin, last modification 2008.11.15


; ------ startup

; --- MCU hardware specifics

; we are using an ATmega32
  .include "m32def.inc"

; needed fuses, fast external quarz osc, max stable startup
  ; FF: 7:BODLVL=1, 6:BOD=1 (disable), 54:SUT=11, 3210:CKSEL=1111 (ext resonat)
  ; D9: 7:OCDEN=1 (dis), 6:JTAGEN=1 (dis), 5:SPIEN=0 (en), 4:CKOPT=1
  ;     3:EESAVE=1 (not preserve), 21:BOOTSZ=00, 0:BOOTRST=1 


; --- prepare coding conventions

; we use an systematic naming of AVR registers
  .include "avr_registers.inc"


; --- board hardware specifics

; VGA "HD-15" connector pinout (correctly named DE15, D-Shell E-size 15pin)

  ; pins   1/2/3:  R/G/B,    0.7V analog
  ; pin    4:      ID2,      TTL
  ; pin    5:      GND       (for IDs?)
  ; pins   6/7/8:  GND RGB   (for R/G/B coax shields)
  ; pin    9:      n.c./key  (can also be wired as +5V for DDC)
  ; pin   10:      GND HV    (for H-Sync + V-Sync coax shields)
  ; pins  11/12:   ID0/ID1,  TTL
  ; pin   13:      H-Sync,   TTL (can also be wired as C-Sync)
  ; pin   14:      V-Sync,   TTL
  ; pin   15:      n.c.      (can also be wired as ID3)
  ;                     (ID1/ID3 can also be wired as SDA/SCA for DDC I2C bus)

; VGA "R" DACs, 3 times 2bit/4value, gives 4*4*4=64 colours, RRGGBB

  ; each DAC from each bit TTL 5V, to 2bit 0.233+0.466V=0.7V @ 75ohm monitor
  ; needs per DAC resistors:
  ;   bit0  75ohm/0.233V*5V-75ohm = 1534okm  -> resistor 1k5ohm
  ;   bit1  75ohm/0.466V*5V-75ohm = 730ohm -> resistor 680ohm (not 820ohm)

; VGA ports and port bits

  ; video pixel data
  .equ VGADPORT = PORTC  ; 3 2bit colour DACs, write fast, must be same port
                         ;   use full 8bit out, bits 7+6 must accept overwrite
  .equ VGADDDR  = DDRC
  .equ VGABLACK = 0x00   ;   all 3 DACs set to 0b00, clear port
  .equ VGABLUE  = 0x03   ;   pins 1+0 blue  2bit DAC
  .equ VGAGREEN = 0x0C   ;   pins 3+2 green 2bit DAC
  .equ VGARED   = 0x30   ;   pins 5+4 red   2bit DAC
  .equ VGAWHITE = VGABLUE+VGAGREEN+VGARED  ; all 3 DACs set to 0b11, also DDR

  ; sync control lines
  .equ VGAHPORT = PORTC  ; 1 H-Sync TTL out, can accept draw overwrite with 0
  .equ VGAHDDR  = DDRC
  .equ VGAHBIT  = PINC6
  .equ VGAVPORT = PORTC  ; 1 V-Sync TTL out, can accept draw overwrite with 0
  .equ VGAVDDR  = DDRC
  .equ VGAVBIT  = PINC7

; mode ports and port bits

  ; size, 25 or 30 rows, VGA text/400@70Hz or graphics/480@60Hz timing
  .equ SIZPORT  = PORTD  ; 1 mode switch TTL in with Pull-Up
  .equ SIZPIN   = PIND
  .equ SIZDDR   = DDRD
  .equ SIZBIT   = PIND7  ; 0(jumper) = 30 rows, 1(default) = 25 rows

  ; line 2, 2nd blank or double scan
  .equ LN2PORT  = PORTD  ; 1 mode switch TTL in with Pull-Up
  .equ LN2PIN   = PIND
  .equ LN2DDR   = DDRD
  .equ LN2BIT   = PIND6  ; 0(jumper) = double scan, 1(default) = 2nd blank


; --- handle reset and interrupts (all identical as all not used)

  .cseg
  .org MINROM

  ; set up vectors
  jmp RESET
  jmp UNUINT
  jmp UNUINT
  jmp UNUINT
  jmp UNUINT

  jmp UNUINT
  jmp UNUINT
  jmp UNUINT
  jmp UNUINT
  jmp UNUINT

  jmp UNUINT
  jmp UNUINT
  jmp UNUINT
  jmp UNUINT
  jmp UNUINT

  jmp UNUINT
  jmp UNUINT
  jmp UNUINT
  jmp UNUINT
  jmp UNUINT

  jmp UNUINT


; --- handle unused interrupts, despite not used

UNUINT:
  ; just return, nothing else makes sense here
  reti


; ------ debugging helpers

; --- debuging LED, using TxD pin + RS232 adapter + RSR23 tester TxD LED

; reset state is DDR-TxD=0=in, TTL=float/5V, RS232=-12V, LED=red
;   allways set port bit and DDR, so no setup is required

DBGLEDGRN:
  cbi PORTD,PINTXD   ; TxD=0, TTL=0V, RS232=12V, LED=green
  sbi DDRD,PINTXD    ; DDR-TxD=1=out
  ret

DBGLEDRED:
  sbi PORTD,PINTXD   ; TxD=1, TTL=5V, RS232=-12V, LED=red
  sbi DDRD,PINTXD    ; DDR-TxD=1=out
  ret

DBGLEDINV:
  sbis PORTD,PINTXD
  rjmp DBGLEDRED     ; TxD is 0, invert to TxD=1
  rjmp DBGLEDGRN     ; TxD is 1, invert to TxD=0


; ------ string handling

; --- define an string buffer, "string accumulator", no dynamic alloc yet

  .equ STRMAXLEN = 80

  .dseg
STRBUF:
  STRBUFLEN:
  .byte 2               ; buffer length, self documenting
                        ;   must be set to STMAXLEN while init
  STRCONTLEN:
  .byte 2               ; used length, actually used amount of buffer
                        ;   must be cleared to 0 while init
  STRDATA:
  .byte STRMAXLEN       ; data comes into here
  .cseg


; --- initialise string subsystem

STRINIT:
  ldi XL,low(STRBUF)    ; at present only one string buffer
  ldi XH,low(STRBUF)

  ldi ZL,low(STRMAXLEN) ; mark string buffer with length, never changed
  ldi ZH,high(STRMAXLEN)
  st X+,ZL
  st X+,ZH

  ldi ZL,0              ; mark string buffer used as empty, changes with data
  ldi ZH,0
  st X+,ZL
  st X+,ZH
  ret


; --- load an string constant into string buffer

; parameters: TOS (old PC) address of string constant (in Flash!)
;             format of constant: .dw length, .db data
;             these is assembled in, directly after the  call STRLDI

STRLDI:

                        ; next address in calling program, begin of constant
                        ;   must use ZH:ZL because of  lpm  instr, are temp
  pop ZH                ;   call stores low first/top, big endian(!)
  pop ZL                ;     so read low as last/top

  lsl ZL                ; correct address for word vs byte addressing
  rol ZH

  push DL               ; no YH:YL for this, as not used as address
  push DH

  lpm DL,Z+             ; length of string constant, from program
  lpm DH,Z+

  ldi XL,low(STRBUF)    ; data will be copied to the string buffer
  ldi XH,high(STRBUF)

  ld MH,X+              ; check if string will fit in buffer
  ld ML,X+
  cp DH,MH              ; high(constlen) - high(buflen)
    brlo STRFITS        ;   <, is OK
    breq STRMAYFIT      ;   =, may fit, test also the low byte
    movw DH:DL,MH:ML    ;   >, too long, shorten to constlen = buflen
    rjmp STRFITS
  STRMAYFIT:
    cp DL,ML            ; low(constlen) - low(buflen)
      brlo STRFITS      ;   <, is OK
      breq STRFITS      ;   =, is also OK, just fits
      movw DH:DL,MH:ML  ;   >, too long, shorten to constlen = buflen
  STRFITS:

  st X+,DL              ; record used length of new constant in buffer
  st X+,DH

  ; DH:DL set above     ; for CHAR = STRCONTLEN..1
  STRLOOP:

    lpm T0,Z+           ; copy string data, from program
    st X+,T0

    sbiw DH:DL,1        ; next CHAR
    brne STRLOOP

  pop DH
  pop DL

  bst ZL,0              ; amount of characters dividable by 2 ?
    brtc STRNOFIL       ;   yes, OK
    adiw ZH:ZL,1        ;   no, jump over added 0x00 filler character
  STRNOFIL:

  lsr ZH                ; undo word vs byte address correction
  ror ZL

                        ; new address in calling program, instr after constant
  ijmp                  ; ijmp  is faster than  push push  back and then  ret


; ------ frame buffer handling

; --- define frame buffer

  ; flag ends of row and screen
  .equ FBENDTOKEN = 0xBF  ; ABORTPSEUDO (= 0x7F) shifted to bits 0.654321
  .equ FBENDVALUE = 0xFF  ; invalid (not 0x00..0x1F) colour value 0xFF

  ; row
  .equ FBROWBEG = 5  ; 5 color bytes a begin of row
  .equ FBCOLS   = 40 ; 40 columns of character token bytes in row
  .equ FBROWEND = 1  ; 1 abort pseudo-character "return" token at end of row
  .equ FBROWLEN = FBROWBEG+FBCOLS+FBROWEND ; legth of row in frame buffer
  ; frame
  .equ FBROWS   = 30 ; rows of characters in entire frame buffer, 25 or 30 use
  .equ FBFREND  = 1  ; no-more-lines pseudo colour value byte at end of frame
  .equ FBSIZE   = FBROWS*FBROWLEN+FBFREND ; space used for frame buffer
  ; uses 25*(5+40+1)+1 = 1151bytes (of ATmega32 2048)

  .dseg
FBUF:
  .byte FBSIZE
  .cseg


; --- write colours into frame buffer

; parameters: XH:XL: address where in frame buffer, gets incremented
;             T0:    colour code, gets truncated to black(0x00)..white(0x3F)

FBWCOLOUR:

  andi T0,VGAWHITE    ; prevent false frame end from broken colour values

  ; XH:XL set by caller, contains address in frame buffer
  st X+,T0            ; place this colour

  ret


; --- write character into frame buffer

; parameters: XH:XL: address where in frame buffer, gets incremented
;             T0:    character code, left converted to token

FBWCHAR:

  cpi T0,FIRSTPSEUDO  ; char - FIRSTPSEUDO (lowest defined, pseudo-char)
    brsh FBBOTTOMOK   ;   >=, is ok
    ldi T0,0x2E       ;   <, replace with ASCII "." to show fixup
  FBBOTTOMOK:
  
  cpi T0,ABORTPSEUDO  ; char - ABORTPSEUDO (lowest not defined, abort ps-char)
    brlo FBTOPOK      ;   <, is ok
    ldi T0,0x2E       ;   >=, replace with ASCII "." to show fixup
  FBTOPOK:

                      ; convert char code to token, 8bit rotate right
                      ;   see CHLINE for explanation of logic used here
  bst T0,0            ; low address:    grab  bit  0
  lsr T0              ; high address:   shift bits .654 3210 -> ..65 4321 (0)
  bld T0,7            ; merge to token                       -> 0.65 4321

  ; XH:XL set by caller, contains address in frame buffer
  st X+,T0            ; place this char

  ret


; --- write string buffer into frame buffer

; parameters: XH:XL: address where in frame buffer, gets incremented

FBWSTR:

  push DL               ; no YH:YL for this, analog to STRLDI
  push DH

  ; XH:XL set by caller, contains address in frame buffer

  ldi ZL,low(STRBUF)    ; from the string buffer, must be ZH:ZL as XH:HL gone
  ldi ZH,high(STRBUF)

  adiw ZH:ZL,2          ; skip buffer length

  ld DL,Z+              ; get used length, for char = LEN..1
  ld DH,Z+
  FBWLOOP:

    ld T0,X             ; check for line wrap
    cpi T0,FBENDTOKEN   ; token - end-of-row-abort-pseudo-char
      brne FBWCOPY      ;   <>, is OK
      adiw XH:XL,FBROWEND ; =, correct address, skip row end abort pseudo-char
      ld T0,X           ;   check for page wrap
      cpi T0,FBENDVALUE ;   value - end-of-frame-pseudo-colour
        brne FBWBEGIN   ;     <>, is OK
        ldi XL,low(FBUF)  ;   =, correct address, wrap to begin of frame buffer
        ldi XH,high(FBUF)
      FBWBEGIN:
      adiw XH:XL,FBROWBEG ; correct address, skip row begin colour codes
    FBWCOPY:

    ld T0,Z+            ; write string data
    call FBWCHAR        ; breaks T0, does X+

    subi DL,1           ; next char
    sbci DH,0
    brne FBWLOOP

  pop DH
  pop DL
  ret


; --- initialise frame buffer

  ; avra fails to make an .equ (or .def) to an register name defined by .def
  .def FBCOL = R19  ; S0
  .def FBROW = R20  ; S1

FBINIT:

  ldi XL,low(FBUF)    ; begin of frame buffer
  ldi XH,high(FBUF)

  ldi FBROW,FBROWS    ; for FBROW = FBROWS..1
  FBRLOOP:

    ; default background colour for this row
    ldi T0,0x03       ; blue
    call FBWCOLOUR    ; breaks T0, does X+

    ; default foreground colour for this row
    ;ldi T0,0x0F       ; cyan
    ;ldi T0,0x1F       ; light blue, slight reddish cyan
    ldi T0,0x1B       ; light blue, sligh reddisch and lower green cyan, best
    ;ldi T0,0x2B       ; light blue, low red+green white
    call FBWCOLOUR    ; breaks T0, does X+

    ; default alternate colours for this row, progressively "stronger"
    ldi T0,0x0F       ; cyan
    call FBWCOLOUR    ; breaks T0, does X+
    ldi T0,0x3F       ; white
    call FBWCOLOUR    ; breaks T0, does X+
    ldi T0,0x3C       ; yellow
    call FBWCOLOUR    ; breaks T0, does X+

    ; blank out the frame buffer
    ldi FBCOL,FBCOLS  ; for FBCOL = FBCOLS..1
    FBCLOOP:

      ldi T0,0x2E     ; fill unused space with ASCII "."
      call FBWCHAR    ; breaks T0 (requiring re-load), does X+

      dec FBCOL       ; next FBCOL
      brne FBCLOOP

    ldi T0,FBENDTOKEN ; shifted end of row abort pseudo-char
    st X+,T0          ; can not use FBWCHAR because framing is invalid char

    dec FBROW         ; next FBROW
    brne FBRLOOP

  ldi T0,FBENDVALUE   ; end of frame no-more-lines pseudo-colour
  st X+,T0            ; can not use FBWCOLOUR because framing is invalid colour

  ret


; ------ X/Y coordinate based frame buffer

; registers for row/coloumn loop counting and parameter passing
  ; avra fails to make an .equ (or .def) to an register name defined by .def
  .def XYCOL = R19  ; S0
  .def XYROW = R20  ; S1


; --- X/Y write colours into frame buffer

; parameters: XYROW: where in frame buffer to place colour
;             T0:    colour code for FBWCOLOUR, gets truncated
;             T1:    LUT reg to set, 0..4 = back/fore/alt1/alt2/alt3

XYWCOLOUR:

  cpi XYROW,FBROWS      ; XYROW in 0..(FBROWS-1) ?
    brsh XYOUTRANGE     ;   >=, dont' colour non-existant row
  cpi T1,FBROWBEG
    brsh XYOUTRANGE     ;   >=, dont' set to non-existant colour

  ldi XL,low(FBUF)      ; begin of frame buffer colours
  ldi XH,high(FBUF)

  ldi T2,FBROWLEN       ; skip rows before this one
  mul XYROW,T2
  add XL,ML             ; this row of frame buffer colours
  adc XH,MH

  add XL,T1             ; add index for wanted colour byte
  ldi T2,0
  adc XH,T2

  jmp FBWCOLOUR

XYOUTRANGE:
  ret


; --- X/Y write character into frame buffer

; parameters: XYROW and XYCOL: where in frame buffer to place character
;             T0:              ASCII character for FBWCHAR, left converted

XYWCHAR:

  cpi XYROW,FBROWS      ; XYROW in 0..(FBROWS-1) ?
  brsh XYOUTRANGE       ;   >=, dont' draw in non-existant row

  cpi XYCOL,FBCOLS      ; XYCOL in 0..(FBCOLS-1) ?
  brsh XYOUTRANGE       ;   >=, dont' draw in non-existant column

  ldi XL,low(FBUF+FBROWBEG) ; begin of frame buffer chars
  ldi XH,high(FBUF+FBROWBEG)

  ldi T2,FBROWLEN       ; skip rows before this one
  mul XYROW,T2
  add XL,ML             ; this row of frame buffer chars
  adc XH,MH

  add XL,XYCOL          ; this character in frame buffer
  ldi T2,0
  adc XH,T2

  jmp FBWCHAR


; --- X/Y write string buffer into frame buffer

; parameters: XYROW and XYCOL: where in frame buffer to place string

XYWSTR:

  cpi XYROW,FBROWS      ; XYROW in 0..(FBROWS-1) ?
  brsh XYOUTRANGE       ;   >=, dont' draw in non-existant row

  cpi XYCOL,FBCOLS      ; XYCOL in 0..(FBCOLS-1) ?
  brsh XYOUTRANGE       ;   >=, dont' draw in non-existant column

  ldi XL,low(FBUF+FBROWBEG) ; begin of frame buffer chars
  ldi XH,high(FBUF+FBROWBEG)

  ldi T2,FBROWLEN       ; skip rows before this one
  mul XYROW,T2
  add XL,ML             ; this row of frame buffer chars
  adc XH,MH

  add XL,XYCOL          ; this character in frame buffer
  ldi T2,0
  adc XH,T2

  jmp FBWSTR


; ------ drawing, timing helper

; --- wait a while doing nothing, spinloop

; parameters: T0: loop count, 1..256 (actually 1..255,0 where 0=256)
;                 gets destroyed (reduced to 0), avoid push/pop time
;                 OK, as seldom reused, so use immediate temporary register

; timing range: (T0-1)*(1+2) + 1*(1+1) + 4 = T0*3+3 clocks
;   together with needed  ldi T0,<n>  and  call WAIT 1+4+(T0*3+3) clocks
;   gives min 1+4+(1*3+3)=11 clocks, for T0=1
;   gives max 1+4+(256*3+3)=776 clocks, for T0=0 (=256)

WAIT:
                         ; {n} in comments = clocks used, for time computation
                         ;   this program requires lots of exact clock counting
                         ;   single-clock miscounts produce visual artifacts!

  ; T0 set up by caller  ; {0}  for LOOP = T0..1
  WAILOOP:

    ; nothing in loop, for highest loop resolution, only 3 clocks

    dec T0               ; {1}  next LOOP
    brne WAILOOP         ; {2|1}

  ret                    ; {4}


; ------ drawing, horizontal and vertical retrace

; --- general drawing sequence

; lines, are the timing critical thing, entire line 600clocks (31.75us/31.5kHz)
;   first visible output pixels, 480 clocks, 160pixels at 3clocks/pixel
;     this is line type specific (blank, vertical retrace, character drawing)
;   then horizontal retrace blank right time, 30clocks
;     this is line specific, used for tidying up and next line preparation
;   then horizontal retrace pulse time, 30clocks
;     this is common, is simply wait time
;   then horizontal retrace blank left time, 60clocks, largest spare time
;     this is line specific, used for further preparing for generating line
;   (entire horizontal retrace is 30+30+60=120, 1/4 of 480 clocks, 1/5 of all)

; frame, for consistency with lines, entire VGA Text frame 450 lines (70Hz)
;   first visible output lines, 25*8*2 = 25*16 = 400 lines
;   then vertical retrace blank bottom lines, 1/4 of 50 = 12 lines
;   then vertical retrace pulse lines, 1/4 of 50 = 13 lines
;   then vertical retrace blank top lines, 1/2 of 50 = 25 lines

; frame, for consistency with lines, entire VGA Graphics frame 525 lines (60Hz)
;   first visible output lines, 30*8*2 = 30*16 = 480 lines
;   then vertical retrace blank bottom lines, 1/4 of 45 = 11 lines
;   then vertical retrace pulse lines, 1/4 of 45 = 11 lines
;   then vertical retrace blank top lines, 1/2 of 45 = 23 lines


; loop counters

; registers for line/segm/row loop counting
  ; avra fails to make an .equ (or .def) to an register name defined by .def
  .def DRLINE = R19      ; S0
  .def DRROW  = R20      ; S1
  .def DRSEGM = R21      ; S2

; vertical blanking linecount/timing
  ; VGA text mode
  .equ VRBBTLINES = 12
  .equ VRPUTLINES = 13
  .equ VRTBTLINES = 25
  ; VGA graphics mode
  .equ VRBBGLINES = 12
  .equ VRPUGLINES = 13
  .equ VRTBGLINES = 25


; --- set up port(s) for sync pulses

SYNCINIT:

  ; get data registers ready
  cbi VGAHPORT,VGAHBIT   ; Port = 0 (= Sync off)
  cbi VGAVPORT,VGAVBIT

  ; get DDR registers ready
  sbi VGAHDDR,VGAHBIT    ; DDR = 1 (= out)
  sbi VGAVDDR,VGAVBIT


  ; size switch (25 row / 30 row) - get data register ready
  sbi SIZPORT,SIZBIT ; Port/Pull-Up = 1 (= on)

  ; size switch (25 row / 30 row) - get DDR register ready
  cbi SIZDDR,SIZBIT  ; DDR = 0 (= input)

  ret


; --- horizontal retrace - generate pulse and start of left blank

; timing 1pre, 30 + 1+4 = (30+60=90)-55spare

HPULSE:

  ; {*pre} = "prepulse" time, before timepoint where pulse is generated
  ;          precompensate for this in callers timing

  ; {30}  horizontal retrace pulse, V = 0or1, H = 1, color = blank
  sbi VGAHPORT,VGAHBIT   ; {1pre+1}  start horizontal retrace pulse

  ldi T0,6               ; {1}  wait its time, 30-1-1=28
  call WAIT              ; {4+(6*3+3)=25}
  nop                    ; {2}
  nop

  ; {3of60}  horizontal retrace left blank, V = 0or1, H = 0, color = blank
  cbi VGAHPORT,VGAHBIT   ; {1pre+1}  end horizontal retrace pulse

  ; no left blank wait, give rest of time to caller, 55spare

  ret                   ; {4}


; --- vertical retrace bottom blank - state machine setup and draw lines

; timing 4+(4+1)=9pre

DISPLAY:
  ; can enter into any of the 4 states, simply take the first one here

VRBBSTATE:

  sbic SIZPIN,SIZBIT     ; {2|1pre}  1(default) = draw 25 rows
  ldi DRLINE,VRBBTLINES  ; {1pre}  for DRLINE=VRBBTLINES..1
  sbis SIZPIN,SIZBIT     ; {2|1pre}  0(jumpered) = draw 30 rows
  ldi DRLINE,VRBBGLINES  ; {1pre}  for DRLINE=VRBBGLINES..1

VRBBLINE:

  call HPULSE            ; {4pre+(1pre+90-55spare)}

  ; {55of60}  horizontal retrace left blank, V = 0, H = 0, color = blank
  ldi T0,15              ; {1}  wait its time
  call WAIT              ; {4+(15*3+3)=52}
  nop                    ; {2}
  nop

  ; {480}  "draw"/wait 160 blank pixels, V = 0, H = 0, color = blank
  ldi T0,157             ; {1}  wait its time
  call WAIT              ; {4+(157*3+3)=478}
  nop                    ; {1}

  ; {30}  horizontal retrace right blank, V = 0, H = 0, color = blank
  ldi T0,2               ; {1}  wait rest of its time, 30-1-1-2-7=19
  call WAIT              ; {4+(2*3+3)=13}
  nop                    ; {2}
  nop

  dec DRLINE             ; {1}  next DRLINE
  breq VRPUSTATE         ; {1|2+11pre}  or even next state

  nop                    ; {5} compensate difference next state vs next line
  nop
  nop
  nop
  nop

  rjmp VRBBLINE          ; {2+(4+1)=7pre}


; --- vertical retrace pulse - state machine setup and draw lines

; timing 4+2+(4+1)=11pre

VRPUSTATE:

  sbic SIZPIN,SIZBIT     ; {2|1pre}  1(default) = draw 25 rows
  ldi DRLINE,VRPUTLINES  ; {1pre}  for DRLINE=VRPUTLINES..1
  sbis SIZPIN,SIZBIT     ; {2|1pre}  0(jumpered) = draw 30 rows
  ldi DRLINE,VRPUGLINES  ; {1pre}  for DRLINE=VRPUGLINES..1

  sbi VGAVPORT,VGAVBIT   ; {2pre}  start vertical retrace pulse

VRPULINE:

  call HPULSE            ; {4pre+(1pre+90-55spare)}

  ; {55of60}  horizontal retrace left blank, V = 1, H = 0, color = blank
  ldi T0,15              ; {1}  wait its time
  call WAIT              ; {4+(15*3+3)=52}
  nop                    ; {2}
  nop

  ; {480}  "draw"/wait 160 blank pixels, V = 1, H = 0, color = blank
  ldi T0,157             ; {1}  wait its time
  call WAIT              ; {4+(157*3+3)=478}
  nop                    ; {1}

  ; {30}  horizontal retrace right blank, V = 1, H = 0, color = blank
  ldi T0,2               ; {1}  wait rest of its time, 30-1-1-2-7=19
  call WAIT              ; {4+(2*3+3)=13}
  nop                    ; {2}
  nop

  dec DRLINE             ; {1}  next DRLINE
  breq VRTBSTATE         ; {1|2+11pre}  or even next state

  nop                    ; {5} compensate difference next state vs next line
  nop
  nop
  nop
  nop

  rjmp VRPULINE          ; {2+(4+1)=7pre}


; --- vertical retrace top blank - state machine setup and draw lines

; timing 4+2+(4+1)=11pre

VRTBSTATE:

  sbic SIZPIN,SIZBIT     ; {2|1pre}  1(default) = draw 25 rows
  ldi DRLINE,VRTBTLINES  ; {1pre}  for DRLINE=VRTBTLINES..1
  sbis SIZPIN,SIZBIT     ; {2|1pre}  0(jumpered) = draw 30 rows
  ldi DRLINE,VRTBGLINES  ; {1pre}  for DRLINE=VRTBGLINES..1

  cbi VGAVPORT,VGAVBIT   ; {2pre}  end vertical retrace pulse

VRTBLINE:

  call HPULSE            ; {4pre+(1pre+90-55spare)}

  ; {55of60}  horizontal retrace left blank, V = 0, H = 0, color = blank
  ldi T0,15              ; {1}  wait its time
  call WAIT              ; {4+(15*3+3)=52}
  nop                    ; {2}
  nop

  ; {480}  "draw"/wait 160 blank pixels, V = 0, H = 0, color = blank
  ldi T0,157             ; {1}  wait its time
  call WAIT              ; {4+(157*3+3)=478}
  nop                    ; {1}

  ; {30}  horizontal retrace right blank, V = 0, H = 0, color = blank
  ldi T0,2               ; {1}  wait rest of its time, 30-1-1-4-7=17
  call WAIT              ; {4+(2*3+3)=13}

  dec DRLINE             ; {1}  next DRLINE
  breq CHSTATE           ; {1|2+13pre}  or even next state

  nop                    ; {7} compensate difference next state vs next line
  nop
  nop
  nop
  nop
  nop
  nop

  rjmp VRTBLINE          ; {2+(4+1)=7pre}


; ------ drawing, visible output

; --- set up registers and ports for DACs and mode switch

; registers for colour patterns, are the LUT, used for  out VGADPORT,value-reg
  ; use non-immediate fast/specials vars, no immediate load and permanent use
  ; avra fails to make an .equ (or .def) to an register name defined by .def

  .def BLANK   = R2 ; V0  constant for colour blanking (black) after line

  .def LUTFORE = R3 ; V1
  .def LUTBACK = R4 ; V2  2 for foreground and background

  .def TEMP    = R5 ; V3  temp for LUTFORE-TEMP-<target> rotate/swap/exchange

  .def LUTALT3 = R6 ; V4  3 alternate foreground colours per row
  .def LUTALT2 = R7 ; V5
  .def LUTALT1 = R8 ; V6

; how many rows we draw, can be less than framebuffer size
  ; VGA text mode
  .equ CHTROWS = 25
  ; VGA graphics mode
  .equ CHGROWS = 30


CHINIT:

  ; do not set LUT colour registers, they will be set while drawing every row
  ;   but set register for blanking after pixels drawn
  ldi T0,VGABLACK    ; black is all DAC bits set to 0, no signal
  mov BLANK,T0       ;   use this for retrace colour blanking


  ; VGA DACs - get data register ready
  out VGADPORT,BLANK

  ; VGA DACs - get DDR register ready
  in T0,VGADDDR
  ori T0,VGAWHITE    ; DDR = 1 = out, white is all DAC bits set to 1
  out VGADDDR,T0


  ; line 2 mode switch (2nd blank / double scan) - get data register ready
  sbi LN2PORT,LN2BIT ; Port/Pull-Up = 1 (= on)

  ; line 2 mode switch (2nd blank / double scan) - get DDR register ready
  cbi LN2DDR,LN2BIT  ; DDR = 0 (= input)

  ret


; --- draw character based visible output line

; timing (1+1)+(4+1+1)+(4+1)=13pre

CHSTATE:

  ldi XL,low(FBUF)   ; {1pre}  get frame buffer first row begin
  ldi XH,high(FBUF)  ; {1pre}

  sbic SIZPIN,SIZBIT ; {2|1pre}  1(default) = draw 25 rows
  ldi DRROW,CHTROWS  ; {1pre}  for DRROW = CHTROWS..1
  sbis SIZPIN,SIZBIT ; {2|1pre}  0(jumpered) = draw 30 rows
  ldi DRROW,CHGROWS  ; {1pre}  for DRROW = CHGROWS..1

  ldi DRSEGM,0x00    ; {1pre}  for DRSEGM = 0x00,0x10..0x70
  ldi DRLINE,0x00    ; {1pre}  for DRLINE = 0x00|0xFF

CHLINE:

  call HPULSE     ; {4pre+(1pre+90-55spare)}

  ; {55of60}  horizontal retrace left blank, V = 0, H = 0, color = blank
  sbrs DRLINE,0   ; {2|1}  1 = second line, test if double scan or 2nd blank
    rjmp CHFIRST  ;   {2}  0 = first line, always draw
  sbis LN2PIN,LN2BIT  ; {2|1}  1(default) = 2nd line blank, blank line
    rjmp CHDOUBLE     ;   {2}  0(jumpered) = double scan, drawn line


  ; {remaining51of60}
  adiw XH:XL,FBROWLEN ; {2}  simulate XH:XL increments, as if we had drawn

  ldi T0,13       ; {1}  wait rest of its time, 55-(2+2)-2=49
  call WAIT       ; {4+(13*3+3)=46}
  nop             ; {2}
  nop

  ; {480}  "draw"/wait 160 blank pixels, V = 0, H = 0, color = blank
  ldi T0,157      ; {1}  wait its time
  call WAIT       ; {4+(157*3+3)=478}
  nop             ; {1}

  ; {30}  horizontal retrace right blank, V = 0, H = 0, color = blank
  out VGADPORT,BLANK  ; {1}  switch off whatever was last background colour
                      ;        this is timing critical, so before jump
  jmp CHEND       ; {3}  go to common CHLINE end handler for rest


  ; {remaining52of60}
  CHFIRST:
  nop             ; {2}  compensate for only one skip used, (2+1+2)-(1+2)=2
  nop

  ; {remaining50of60}
  CHDOUBLE:
  ldi T0,8        ; {1}  wait rest of its time, 55-(2+1+2)-10-8=32
  call WAIT       ; {4+(8*3+3)=31}


  ; {10}  load LUT colour registers for this row
  ;   XH:XL frame buffer address has already been set up by above
  ld LUTBACK,X+   ; {2}  first 5 bytes in row in frame buffer are colours
  ld LUTFORE,X+   ; {2}
  ld LUTALT1,X+   ; {2}
  ld LUTALT2,X+   ; {2}
  ld LUTALT3,X+   ; {2}


  ; {8}  get first character token and unthread it
  ;   use standard font unthreading technique, see there for generic case
  ;     but compressed, without 4 clocks of intermixed pixel OUT statements
  ;   XH:XL frame buffer address has already been set up by above
  ;   DRSEGM segment offset has already been set up by above
  ld ZL,X+        ; {2}  next (=first) character: get token
  mov ZH,ZL       ; {1}  next (=first) character: extract top part
  andi ZH,0x3F    ; {1}    for ATmega32: 0x3F limits address to 64(*2) chars
  andi ZL,0x80    ; {1}  next (=first) character: extract bottom part
  or ZL,DRSEGM    ; {1}    add in segment offset for this line in row
  ijmp            ; {2}  next (=first) character: go and draw it
                  ;        continuation at next char, repeating above


  ; {40*4*3=480}  draw 40chars*4pixels =160 pixels, V = 0, H = 0, color = pixel

  ; for each X+ fetched char token: index and exec one segment in font space
  ;   drawing done by  out VGADPORT,colour, one for each pixel, every 3 clocks
  ;   {so per segment time 4 pixels a 3 clocks = 12 clocks/segment}

  ; per character 8 such segments, one for each line in row
  ;   so per segment size 16 instructions * 2 bytes = 128instr/char
  ;     on ATmega32 32k Flash this allows 32k/2/128=128chars, minus used space
  ;   note that with max 6 drawable lines this could reduce to 6*10(+4==64instr
  ;     on ATmega32 this would allow 32k/2/64/=256chars, minus used space


  ; *** this stuff is in the font file, included near bottom of this file ***

  ; each char in font is an 8*16=128instr block, use up lowest 7 address bits
  ;   we need fast 8bit indexing (*128) in unthreading code, 7 shifts too slow
  ; split tokens into 1+6 bits (7 and 5..0) and use these for 2 address halfs
  ;   high address: bits 5..0 become address bits 13..8 (15+14 nonexistant!)
  ;   low address: bit 7 becomes address bit 7, rest set to 0
  ;   after that add in line segment offset (0x00 0x10 .. 0x70) in bits 6..3
  ; result is for linear char codes (made by font generator)
  ;   that we have interleaved tokens (00 80 01 81 02 82 ... 3F BF)
  ; to precompensate for this by convert char code to token
  ;   ASCII code bits: .654 3210, high address bits: ..65 4321
  ;     low address bits: 0... ...., token for indexing: 0.65 4321
  ;   this conversion is done above in FBCHAR

  ; this code 8 times per character, once per line segment:
  ;   4 clocks write pixels with  out VGADPORT,xxxx (xxxx = LUTFORE or LUTBACK)
  ;   between these 4*2 clocks for unthreading and jumping to next character

  ; out VGADPORT,xxxx ; {1}  current character: draw pixel 1
  ; ld ZL,X+          ; {2}  next character: get character token

  ; out VGADPORT,xxxx ; {1}  current character: draw pixel 2
  ; mov ZH,ZL         ; {1}  next character: copy/extract high part of index
  ; andi ZH,0x3F      ; {1}    for ATmega32: 0x3F limit address to 64(*2) chars

  ; out VGADPORT,xxxx ; {1}  current character: draw pixel 3
  ; andi ZL,0x80      ; {1}  next character: extract low part for (64*)2 chars
  ; ori ZL,0xo0       ; {1}    add in line segment offset for this line in row
                      ;          o = 0x00 0x10 .. 0x70

  ; out VGADPORT,xxxx ; {1}  current character: draw pixel 4
  ; ijmp              ; {2}  next character: go and execute it
                      ;        continuation at that chars pixel 1, as above
                      ;        this results in an endless loop

  ; .db "012345678901" ; fill/overjump unused space, 16-10 = 6words|12bytes

  ; *** end of font file stuff ***


  ; {30}  horizontal retrace right blank, V = 0, H = 0, color = blank

  ; above continues endless until an  abort pseudo-char  token in frame buffer
  ;   that pseudo-char exits the unthreading instead of drawing, breaks loop
  ;     it must be done this way, because no time for 40 char loop counting
  ;   the pseudo-char also runs one  out VGADPORT,BLANK  to end line drawing


  ; *** this stuff is after font include, as at end of this file ***

  ; post-ASCII special pseudo-char 0x7F
  ; .org $3F80    ; would be place for ASCII non-drawable DEL 0x7F, * 128

                  ; 8 line segments repeated identically

  ; out VGADPORT,BLANK  ; {1}  switch off whatever was last background colour
                        ;        this is timing critical, so before jump
  ; jmp CHEND     ; {3}  go back to common CHLINE handler for rest
                  ; fill/overjump unused space, 16-3 = 13words|26bytes
  ; .db "01234567890123456789012345"

  ; *** end of after font file stuff ***


CHEND:
                ; {1+3}  in abort pseudo-character, or blank line

  com DRLINE    ; {1}  next/other DRLINE switch, 0x00 <-> 0xFF

  breq CHNEXSEG ; {1|2}  LINE now 0x00, was 0xFF, do next segment
                ;        else second line of this segment
    sbiw XH:XL,FBROWLEN ; {2}  undo XH:XL increment, repeat same row

    ldi T0,2    ; {1}  wait rest of its time, 30-(1+3)-1-1-2-7=15
    call WAIT   ; {4+(2*3+3)=13}
    nop         ; {1}

    rjmp CHLINE ; {2+(4+1)=7pre}
  CHNEXSEG:

  subi DRSEGM,-0x10 ; {1}  next DRSEGM, + 0x10instr|16instr|32bytes per segm
  andi DRSEGM, 0x70 ; {1}  keep DRSEGM inside the valid range, 0x00..0x70

  breq CHNEXROW ; {1|2}  DRSEGM now 0x00, was 0x70, do next row
                ;        else second to seventh segments of this row
    sbiw XH:XL,FBROWLEN ; {2}  undo XH:XL increment, repeat same row

    ldi T0,1    ; {1}  wait rest of its time, 30-(1+3)-1-2-(1+1)-1-2-7=11
    call WAIT   ; {4+(1*3+3)=10}

    rjmp CHLINE ; {2+(4+1)=7pre}
  CHNEXROW:

  dec DRROW     ; {1}  next DRROW
  brne CHDOROW  ; {1|2}  reversed logic, as too far for direct breq VRBBSTATE
    nop         ; {3}  wait rest of its time, 30-(1+3)-1-2-(1+1)-2-1-1-14=6
    nop
    nop
    jmp VRBBSTATE  ; {3+11=14pre}  or even next state
  CHDOROW:

  nop           ; {9}  wait rest of its time, 30-(1+3)-1-2-(1+1)-2-1-2-7=9
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop

  rjmp CHLINE   ; {2+(4+1)=7pre}


; ------ main program

; --- blank line, to separate sections

DEBLANK:

  ldi T0,0x00         ; black
  ldi T1,0            ; set background
  call XYWCOLOUR

  call STRLDI         ; empty string
  .dw 40
  .db "                                        "

  call XYWSTR         ; print it
  inc XYROW           ; next row

  ret


; --- black text on coloured background line, for rainbow

; parameters: T1: background colour

DERAIN:

  ldi T1,0            ; set background
  call XYWCOLOUR

  ldi T0,0x00         ; black
  ldi T1,1            ; set foreground
  call XYWCOLOUR

  call XYWSTR         ; print it
  inc XYROW           ; next row

  ret


; --- separator

DESEPAR:

  ldi T0,0x1F         ; all 4 2x2pixel set, gives black
  call XYWCHAR        ; print it
  inc XYROW           ; next row

  ldi T0,0x1F
  call XYWCHAR
  inc XYROW

  ldi T0,0x1F
  call XYWCHAR
  inc XYROW

  ldi T0,0x1F
  call XYWCHAR
  inc XYROW

  ldi T0,0x1F
  call XYWCHAR
  inc XYROW

  ldi T0,0x1F
  call XYWCHAR
  inc XYROW

  ret


; --- logo put together from 2x2pixel subchar graphics

; print this text using 1/2+7+1/2 * 4x12pixels, of 2x2pixel subchar graphics
;   = space used 1/2+7+1/2 * 2x6 chars = 16 chars width, 6 rows

; .... ........ ........ ........ ....  .... ........ ........ ........ ....
; .... ()()()() ()()()() ()()()() ()()  ()() ()()()() ()()()() ()()()() ....

; .... ........ ........ ........ ....  .... ........ ........ ........ ....
; .... ..()().. ........ ....().. ..()  .... ..()..() ....()() ....().. ....

; .... ()...... ........ ..().... ..()  .... ..()..() ..().... ..()..() ....
; .... ..().... ..().... ()()().. ()()  ().. ..()..() ..().... ..()..() ....

; .... ....().. ()..().. ..().... ..()  .... ..()..() ..()..() ..()()() ....
; .... ....().. ()..().. ..().... ..()  .... ....().. ..()..() ..()..() ....

; .... ()().... ..().... ..().... ....  ().. ....().. ....()() ..()..() ....
; .... ........ ........ ........ ....  .... ........ ........ ........ ....

; .... ()()()() ()()()() ()()()() ()()  ()() ()()()() ()()()() ()()()() ....
; .... ........ ........ ........ ....  .... ........ ........ ........ ....


DELOGO:

  call STRLDI
  .dw 16
  .db 0x10,  0x1C, 0x1C,  0x1C, 0x1C,  0x1C, 0x1C,  0x1C
  .db 0x1C,  0x1C, 0x1C,  0x1C, 0x1C,  0x1C, 0x1C,  0x10
  call XYWSTR         ; print it
  inc XYROW           ; next row

  call STRLDI
  .dw 16
  .db 0x10,  0x18, 0x14,  0x10, 0x10,  0x10, 0x14,  0x18
  .db 0x10,  0x18, 0x18,  0x10, 0x1C,  0x10, 0x14,  0x10
  call XYWSTR
  inc XYROW

  call STRLDI
  .dw 16
  .db 0x10,  0x19, 0x10,  0x18, 0x10,  0x1E, 0x14,  0x1E
  .db 0x14,  0x1A, 0x1A,  0x1A, 0x10,  0x1A, 0x1A,  0x10
  call XYWSTR
  inc XYROW

  call STRLDI
  .dw 16
  .db 0x10,  0x10, 0x15,  0x15, 0x15,  0x1A, 0x10,  0x1A
  .db 0x10,  0x12, 0x16,  0x1A, 0x1A,  0x1A, 0x1B,  0x10
  call XYWSTR
  inc XYROW

  call STRLDI
  .dw 16
  .db 0x10,  0x13, 0x10,  0x12, 0x10,  0x12, 0x10,  0x10
  .db 0x11,  0x10, 0x11,  0x10, 0x13,  0x12, 0x12,  0x10
  call XYWSTR
  inc XYROW

  call STRLDI
  .dw 16
  .db 0x10,  0x13, 0x13,  0x13, 0x13,  0x13, 0x13,  0x13
  .db 0x13,  0x13, 0x13,  0x13, 0x13,  0x13, 0x13,  0x10
  call XYWSTR
  inc XYROW

  ret


; --- multistyle and multicoloured RGB text, for feature demo

; parameters: T1: main foreground colour, rest of colours derived from this

DERGB:

  mov S2,T0           ; save for multiple reuse

  ldi T0,0x00         ; black
  ldi T1,0            ; set background
  call XYWCOLOUR

  mov T0,S2           ; full colour
  ldi T1,1            ; set foreground
  call XYWCOLOUR

  mov T0,S2           ; 2/3 bright
  andi T0,0x2A        ; kill bit0 of each DAC
  ldi T1,2            ; set alternate colour 1
  call XYWCOLOUR

  mov T0,S2           ; 1/3 bright
  andi T0,0x15        ; kill bit2 of each DAC
  ldi T1,3            ; set alternate colour 2
  call XYWCOLOUR

  mov T0,S2           ; pastel
  ori T0,0x2A         ; force bit2 of all DACs
  ldi T1,4            ; set alternate colour 3
  call XYWCOLOUR

  call XYWSTR         ; print it
  inc XYROW           ; next row

  ret


; --- show something in frame buffer

DEMO:

  ldi XYROW,0         ; start at top/left of screen
  ldi XYCOL,0


  ; show all 95 ASCII chars in 3 rows, uses 3+1 rows

  call STRLDI
  .dw 40
  .db "     !", 0x22, "#$%&", 0x27, "()*+,-./0123456789:;<=>?    "
  call XYWSTR
  inc XYROW

  call STRLDI
  .dw 40
  .db "    @ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_    "
  call XYWSTR
  inc XYROW

  call STRLDI
  .dw 40
  .db "    `abcdefghijklmnopqrstuvwxyz{|}~     "
  call XYWSTR
  inc XYROW


  call DEBLANK


  ; show ASCII graphic, with rainbow coloured background, uses 6+1 rows
  ; first 23 charactes of width

  call STRLDI         ; common text pattern
  .dw 23
  .db "-'__`-.__.-'__`-.__.-'_ "

  ldi T0,0x30         ; red
  call DERAIN

  ldi T0,0x3C         ; yellow
  call DERAIN

  ldi T0,0x0C         ; green
  call DERAIN

  ldi T0,0x0F         ; cyan
  call DERAIN

  ldi T0,0x03         ; blue
  call DERAIN

  ldi T0,0x33         ; magenta
  call DERAIN


  ; show separator, with same rainbow, uses same 6+1 rows
  ; 1 character of width

  push XYROW          ; save position
  push XYCOL

  subi XYROW,6        ; back up the 6 rows shared with rainbow
  ldi XYCOL,23        ; indent by 23 characters
  call DESEPAR

  ; show SoftVGA logo, with same rainbow, uses same 6+1 rows
  ; remaining 16 characters of width

  pop XYCOL           ; restore position first time
  pop XYROW
  push XYROW          ; and save again
  push XYCOL

  subi XYROW,6        ; back up the 6 rows shared with rainbow
  ldi XYCOL,24        ; indent by 23 characters
  call DELOGO

  pop XYCOL           ; restore position second time
  pop XYROW


  call DEBLANK


  ; show announce text, uses middle 3 rows, of the 11+3+11

  call STRLDI
  .dw 40
  .db "AVR ATmega32 uC + SoftVGA = 400line 70Hz"
  call XYWSTR
  inc XYROW

  call STRLDI
  .dw 40
  .db "18.432MHz, 3clk/pixl, 4pixl/char, 40char"
  call XYWSTR
  inc XYROW

  call STRLDI
  .dw 40
  .db "full ASCII, 4x8pixel font, 2 lines/pixel"
  call XYWSTR
  inc XYROW


  ; show RGB foregrounds with styling and alternative colours, uses 1+7 rows

  call DEBLANK


  call STRLDI         ; common text pattern
  .dw 40
  .db "nor", INON, "inv", INOFF
  .db "nor", ULON, "ulin"
  .db ULOFF, "nor", ALT1, "2/3"
  .db ALT1, "nor", ALT2, "1/3"
  .db ALT2, "nor", ALT3, "pas"

  ldi T0,0x03         ; blue
  call DERGB

  ldi T0,0x0C         ; green
  call DERGB

  ldi T0,0x0F         ; cyan
  call DERGB

  ldi T0,0x30         ; red
  call DERGB

  ldi T0,0x33         ; magenta
  call DERGB

  ldi T0,0x3C         ; yellow
  call DERGB

  ldi T0,0x3F         ; white
  call DERGB


  ; show website info text, uses 1+2 rows

  call DEBLANK


  call STRLDI
  .dw 40
  .db "design and code open source available at"
  call XYWSTR
  inc XYROW

  call STRLDI
  .dw 40
  .db "http://neil.franklin.ch/Projects/SoftVGA"
  call XYWSTR
  inc XYROW


  ; additional 30 row draw mode, uses 1+4 rows

  call DEBLANK


  ret


; --- handle reset, initialise system, enter main loop

RESET:

  ; get stack ready, for calls and interrupts (later not used)
  ldi ZL,low(MAXRAM)
  ldi ZH,high(MAXRAM)
  out SPL,ZL
  out SPH,ZH

  ; get interrupts ready, despite not used
  sei

  ; set up string handling
  call STRINIT

  ; set up frame buffer
  call FBINIT

  ; set up port(s) for generating sync pulses
  call SYNCINIT

  ; set up registers and ports for DACs and mode switch
  call CHINIT

  ; fill frame buffer
  call DEMO

  ; enter endless loop, no return back to here
  jmp DISPLAY


; ------ fixed calculated address stuff, for chars, executed by ijmp, no labels

; overjump program code space, which results in a few unusable character codes
;   for each unused character code we get 8segment*16word|32byte space

; char codes 0x00..0x08 used by system, for code space
;   this allows for 9*(128word|256byte) = 1152word|2304byte of code
; char codes 0x0B..0x0F used by system, for colour switching
; char codes 0x10..0x1F used by font, for 2^(2*2)=16 2x2pixel/char block graph
; char codes 0x20..0x7E used by font, for the standard 95 printable ASCII
; char code  0x7F       used by system, for frame buffer abort pseudo-char code


; --- "drawing" routines for pseudo-characters "specials"

  ; pre-ASCII specials pseudo-chars from 0x09-0x0F
  .org $0480          ; would be place for ASCII non-drawable 0x09, * 128
  ;   no labels after .org, as font code is only ever reached by computed ijmp
  .equ FIRSTPSEUDO = 0x09
  .equ LASTPSEUDO  = 0x0F

  ; include specials as separate file, so they may be generated automatically
  .include "vga_text_specials.inc"


; --- drawing routines for drawable characters

  ; pre-ASCII block graphics 0x10 to 0x1F
  ;   and ASCII chars 0x20 to 0x7E
  .org $0800          ; would be place for ASCII 0x10, * 128
  ;   no labels after .org, as font code is only ever reached by computed ijmp
  .equ FIRSTCAHR = 0x10
  .equ LASTCHAR  = 0x7E

  ; include font as separate file, so they may be generated automatically
  ;   from file vga_text_font.fon, by generator ./genfont
  .include "vga_text_font.inc"


; --- "drawing" routines for pseudo-character "abort"

  ; post-ASCII special pseudo-char 0x7F
  .org $3F80          ; would be place for ASCII non-drawable DEL 0x7F, * 128
  ;   no labels after .org, as font code is only ever reached by computed ijmp
  .equ ABORTPSEUDO = 0x7F

                      ; 8 line segments repeated identically

  out VGADPORT,BLANK  ; {1}  switch off whatever was last background colour
                      ;        this is timing critical, so before jump
  jmp CHEND           ; {3}  go back to common CHLINE handler for rest
                      ; fill/overjump unused space, 16-3 = 13words|26bytes
  .db "01234567890123456789012345"

  out VGADPORT,BLANK
  jmp CHEND 
  .db "01234567890123456789012345"

  out VGADPORT,BLANK
  jmp CHEND 
  .db "01234567890123456789012345"

  out VGADPORT,BLANK
  jmp CHEND 
  .db "01234567890123456789012345"

  out VGADPORT,BLANK
  jmp CHEND 
  .db "01234567890123456789012345"

  out VGADPORT,BLANK
  jmp CHEND 
  .db "01234567890123456789012345"

  out VGADPORT,BLANK
  jmp CHEND 
  .db "01234567890123456789012345"

  out VGADPORT,BLANK
  jmp CHEND 
  .db "01234567890123456789012345"
